"""
@(#) SGeSAP/LX - 12.10.00 - $Rev: 1699 $ - Last modified 04/11/17

SAP HANA database HA provider (contrib)

(C) Copyright 2016-17 Hewlett-Packard Enterprise Development Company, L.P.

Prerequisites:
- Serviceguard (preferably >=12.00.51) installed
- Serviceguard cluster is setup
- Local node is member of the cluster
- "Persitence packages" need to have autorun disabled (and shouldn't be enabled manually)

Caveats:
- once a package start fails it is disabled (switching) for this node and can't be
  start with just cmrunpkg anymore. A cmmodpkg -e -n <node> <pkg> is needed to
  reenable it. This is, however, .... cmmodpkg taking a timeout in certain situations.
  To see if it has switching disabled, a cmviewcl -v ... is need, also taking a timeout.

"""

# python modules
import sys, getpass,re, time, os
from hdb_ha.client import StorageConnectorClient,StorageConfigurationInstallation,Helper,StorageConfigurationHDB,InstallerTracer


#####################################################################################


class SGcluster:
    """
    Base info for a Serviceguard cluster (name,status,state,coordinator,version,,...)
    e.g.
    [root@sgxsap78h ~]# cmviewcl -f line -l cluster
    name=hdbrhel
    id=564664641
    status=up
    state=stable
    coordinator=sgxsap74h

    As there is only one SG cluster per node this could as well be a singleton(?)

    state: stable, reforming
    status:  up, down, reforming, halting, or unknown  (starting?)
    """

    def __init__(self, tracer = None, sgsbin = None):

        if tracer != None:
            tracer.debug('>> SGcluster.init')

        self.tracer = tracer
        self.sgsbin = sgsbin
        self.update() # fill in name,status,state(,coordinator?)
        self.clupkgs = []
        if self.tracer != None:
            self.tracer.info('Cluster %s is in status/state: %s/%s' % (self.name,self.status,self.state))

        if self.tracer != None:
            self.tracer.debug('<< SGsimplepkg.init')

    def run(self):
        pass # cmruncl

    def halt(self):
        pass # cmhaltcl

    def update(self):
        """
        Call cmviewcl on cluster level, scan output for these attribute and update
        the object information.

        Raises an Exception if cmviewcl fails
        """

        if self.tracer != None:
            self.tracer.debug('>> SGcluster.update')
        (code, output) = Helper._runOsCommand('%scmviewcl -f line -l cluster' % (self.sgsbin), self.tracer)
        if code != 0:
            self.tracer.error('Serviceguard cmviewcl for cluster failed: %s' % (output))
            msg = 'Serviceguard cmviewcl for cluster failed'
            raise Exception(msg)
        for line in output.split():
            mo = re.search('^(name|status|state)=(.+)$', line)
            if mo != None:
                #self.tracer.debug('%s %s' % (mo.group(1), mo.group(2)))
                setattr(self,mo.group(1),mo.group(2))
        if self.tracer != None:
            self.tracer.debug('<< SGcluster.update')
        return True

    def get_cmversion(self):
        # call cmversion and return output
        self.tracer.debug('>> SGcluster.get_cmversion')
        (code, output) = Helper._runOsCommand('%s%s' % (self.sgsbin, 'cmversion'), self.tracer)
        if code != 0:
            msg = "Serviceguard call cmversion failed"
            self.tracer.fatal(msg)
            raise Exception(msg)
        self.tracer.debug('cmversion: %s' % output)
        self.tracer.debug('<< SGcluster.get_cmversion')
        return output


    def check_version(self,min_version):
        # Compare minSGversion with version returned
        # (Alternative: use provider_version from ver)
        # Returned version is expected to have the form:
        # <prefix>.<number>.<number>.<number>, example A.12.00.50
        # Returns True if version is ok, False otherwise

        # TODO:
        # instead of looping just over one array (req_vers) and use an index for 2nd,
        # zip could be used to loop over both at once
        # for a,b in zip (alist, blist)

        self.tracer.debug('>> SGcluster.check_version')
        req_vers = min_version.split('.')
        output = self.get_cmversion()
        out_vers = output.split('.')[1:] # get rid of leter prefix
        i = 0
        for rv in req_vers:
            if int(rv) > int(out_vers[i]):
                self.tracer.error('SG version %s lower than required version %s' % (output.split()[0],SGeSAPsc.minSGversion))
                self.tracer.debug('<< SGcluster.check_version (%i)' % i)
                return False
            if int(rv) < int(out_vers[i]):
                self.tracer.debug('<< SGcluster.check_version (%i)' % i)
                return True
                # else equal: next elements
            i += 1
        self.tracer.debug('<< SGcluster.check_version (%i)' % i)
        return True # equal

    def isInTransientState(self):
        # uses status instead of state
        SGcluTransientStatus = ('starting','halting','reforming')
        if self.status in SGcluTransientStatus:
            return True
        return False

    def waitForStableState(self, retries=12, waitretry=5):
        while self.isInTransientState():
            if not retries:
                return False
            time.sleep(waitretry)
            retries-= 1
            self.update()
        return True

    def retrieveConfiguredPkgs(self):
        """
        Get list of configured packages.
        Array of SGsimplepkg assigned to configuredPkgs
        """
        self.clupkgs = []  # clear previous entries

        if self.tracer != None:
            self.tracer.debug('>> SGcluster.retrieveConfiguredPkgs')

        pkgindex = -1

        (code, output) = Helper._runOsCommand('%scmviewcl -f line -l package' % (self.sgsbin), self.tracer)

        # below is similar to SGsimpepkg init except that is also parses the package prefix
        for line in output.split():
            mo = re.search(r'^package:(.+)\|name=\1$', line)
            if mo != None:
                self.tracer.debug('Pkg: %s' % mo.group(1))
                pkgindex += 1
                self.clupkgs.append (SGsimplepkg(None, self.tracer, self.sgsbin)) # do not update with ctor
                self.clupkgs[pkgindex].name = mo.group(1)
                continue

            mo = re.search('^package:%s\|(type|status|state|autorun|owner)=(.+)$' % self.clupkgs[pkgindex].name, line)
            if mo != None:
                #self.tracer.debug('Pkg %s %s %s' % (self.clupkgs[pkgindex].name, mo.group(1), mo.group(2)))
                setattr(self.clupkgs[pkgindex],mo.group(1),mo.group(2))

        self.tracer.info('Cluster has %d packages configured' % len(self.clupkgs))

        self.tracer.debug('<< SGcluster.retrieveConfiguredPkgs')

    def getSCpkgs(self, prefix=None , postfix=None):
        """
        Get the list of storage controller packages
        Either use prefix/postfix naming convention
        or serach for special module (2nd TODO)
        """

        if self.tracer != None:
            self.tracer.debug('>> SGcluster.getSCpkgs')

        scpkgs = []
        for pkg in self.clupkgs:
            # start with prefix, end with postfix (digits in between are not checked)
            if (pkg.name.startswith(prefix) and pkg.name.endswith(postfix)):
                self.tracer.debug('SC pkg: %s'% pkg.name)
                scpkgs.append(pkg)

        if self.tracer != None:
            self.tracer.debug('<< SGcluster.getSCpkgs')

        return scpkgs

    def updateSCpkgs(self):
        """
        update owner/status/state of SC pkg list
        """
        pass # TODO. implementation pending


class SGnode:
    """
    Base info for a Serviceguard cluster node (name,status,state,...)
    usually for the local node
    e.g.
    [root@sgxsap78h ~]# cmviewcl -f line -n sgxsap78h
    name=sgxsap78h
    status=up
    state=running
    initial_incarnation=6
    id=3
    node_pr_key=19410003

    status: up, down, starting, halting, or unknown
    state: running,unknown (halted?)
    """
    def __init__(self, nodename = None, tracer = None, sgsbin = None):

        if tracer != None:
            tracer.debug('>> SGnode.init')

        if not nodename:
            # use localhost (import os required?)
            nodename = os.uname()[1]
            tracer.debug('local nodename = %s', nodename)

        self.name = nodename
        self.tracer = tracer
        self.sgsbin = sgsbin
        # TODO: check if node is configured, would need to query
        #       (singleton?) SGcluster
        self.update()
        if self.tracer != None:
            self.tracer.info('Node %s is in status/state: %s/%s' % (self.name,self.status,self.state))

        if self.tracer != None:
            self.tracer.debug('<< SGnode.init')


    def run(self):
        pass # cmrunnode

    def halt(self):
        pass # cmhaltnode

    def update(self):
        """
        Call cmviewcl on nodename, scan output for these attributes and update
        the object information.

        Raises an exception if cmviewcl fails
        """

        if self.tracer != None:
            self.tracer.debug('>> SGnode.update')
        (code, output) = Helper._runOsCommand('%scmviewcl -f line -n %s' % (self.sgsbin, self.name), self.tracer)
        if code != 0:
            if self.tracer != None:
                self.tracer.error('Serviceguard cmviewcl for %s failed: %s' % (self.name, output))
            msg = 'Serviceguard cmviewcl for node %s failed' % self.name
            raise Exception(msg)
        for line in output.split():
            mo = re.search('^(name|status|state)=(.+)$', line)
            if mo != None:
                #self.tracer.debug('%s %s' % (mo.group(1), mo.group(2)))
                setattr(self,mo.group(1),mo.group(2))
        if self.tracer != None:
            self.tracer.debug('<< SGnode.update')
        return True

    def isInTransientState(self):
        # uses status instead of state
        SGnodeTransientStatus = ('starting','halting')
        if self.status in SGnodeTransientStatus:
            return True
        return False

    def waitForStableState(self, retries=12, waitretry=5):
        while self.isInTransientState():
            if not retries:
                return False
            time.sleep(waitretry)
            retries-= 1
            self.update()
        return True


class SGsimplepkg():
    """
    SGsimplepkg represents a SG package with instance variables for the
    information retrieved by
    cmviewcl -f line -p <pkg>: name,type,status,state, autorun, owner
    """

    def __init__(self, pkgname = None, tracer = None, sgsbin = None):

        if tracer != None:
            tracer.debug('>> SGsimplepkg.init %s' % (pkgname or "None"))

        self.tracer = tracer
        self.sgsbin = sgsbin

        if pkgname == None:
            return

        self.name = pkgname
        # TODO: check if package is configured (ask SGcluster instance?)
        #       if not throw an exception
        self.update()
        if self.tracer != None:
            self.tracer.info('Package %s (owner: %s) is in status/state: %s/%s' % (self.name,self.owner, self.status,self.state))
            self.tracer.debug('<< SGsimplepkg.init')

    def update(self):
        """
        Call cmviewcl on name, scan output for these attributes and update
        the object information.
        Output of version 12.00.51 also contains multilevel node inforamation
        which has to be igngored

        Raises an exception if cmviewcl fails
        """

        if self.tracer != None:
            self.tracer.debug('>> SGsimplepkg.update')
        (code, output) = Helper._runOsCommand('%scmviewcl -f line -p %s' % (self.sgsbin, self.name), self.tracer)
        if code != 0:
            if self.tracer != None:
                self.tracer.error('Serviceguard cmviewcl for %s failed: %s' % (self.name, output))
            msg = 'Serviceguard cmviewcl for package %s failed' % self.name
            raise Exception(msg)
        for line in output.split():
            mo = re.search('^(name|type|status|state|autorun|owner)=(.+)$', line)
            # TODO: assume that 12.00.51 node autput appears at the end and jump
            #       out of the loop on 1st occurence
            if mo != None:
                #self.tracer.debug('%s %s' % (mo.group(1), mo.group(2)))
                setattr(self,mo.group(1),mo.group(2))
        if self.tracer != None:
            self.tracer.debug('<< SGsimplepkg.update')
        return True

    def isInTransientState(self):
        """
        check if package state is in one of
        starting, start_wait, halting, halt_wait, relocate_wait, failing,   changing(MNP)

        what about: blocked? A: doesn't seem to be a general pkg state, only a per node state

        (halt_aborted,failed,unknown,halted,maintenance,detached,running
        are considered stable states
        + reconfigure, reconfigure_wait
        TODO  maintenance ?)
        (status=starting|up)

        states according to man 5 cmviewcl
        VALUES: changing, starting, start_wait,  running,
                halting,  halt_wait,  halted, halt_aborted,
                failing, fail_wait, failed,
                relocate_wait,  unknown,
                reconfiguring,  reconfigure_wait, maintenance, detached.

        state=maintenance on pkg level implies that it is in same state on a node
        if this isn't the local node, the package can't be started locally, even if status=down
        state=maintenace can be combined with status=up|down|starting|halting, so maintenance
        isn't necessarily a stable state
        """
        SGpkgTransientState = ('starting','start_wait','halting','halt_wait','relocate_wait','failing','changing','reconfiguring', 'reconfigure_wait')
        if self.state in SGpkgTransientState:
            return True
        return False

    def waitForStableState(self, retries=12, waitretry=5):
        """
        Wait for package to get into stable state (meaning that it is coming out of
        transient state). The check is executed retries number of times, with
        waitretry delay between. Default is ~1min waittime (12*5) for package to reach
        stable state
        """
        while self.isInTransientState():
            if not retries:
                return False
            time.sleep(waitretry)
            retries-= 1
            self.update()
        return True

    def isInIgnoredState(self):
        """
        Package state being ignored and will return success (0) on attach/detach
        (detached, halt_aborted)

        halt_aborted state related to partially_down status
        There is no "partially_up" status if a package in maintenance mode
        is started up to a certain module
        """
        if self.state in ('halt_aborted','detached'):
            return True
        return False

    def isUnmanaged(self):
        """
        Considers package in state maintenance and detached to be "unmanaged"
        by the cluster and leaves the package alone.
        BTW, a package being in maintenace mode on one node cannot be run on
        another - and it can be in maintenance mode only on a single node.
        """
        if self.state in ('maintenance','detached'):
            return True
        return False
        # TODO: shortcut: return (self.state in ...) ?

    def isInMaintenance(self):
        """
        Check for package state maintenance
        """
        if self.state in ('maintenance'):
            return True
        return False

    def isUnowned(self):
        if self.owner == 'unowned':
            return True
        return False

    def run(self):
        """
        SG cmrunpkg as OS command
        """
        self.tracer.info('Running package %s' % self.name)
        (c,o) = Helper._runOsCommand('%s%s %s' % (self.sgsbin, 'cmrunpkg -v', self.name), self.tracer)
        if c != 0:
            if self.tracer != None:
                self.tracer.error('Running package %s failed' % self.name)
            # TODO: (c,o) output ?
            return False
        self.tracer.info('Running package %s succeeded' % self.name)
        return True

    def halt(self):
        """
        SG cmhaltpkg as OS command
        """
        self.tracer.info('Halting package %s' % self.name)
        (c,o) = Helper._runOsCommand('%s%s %s' % (self.sgsbin, 'cmhaltpkg -v', self.name), self.tracer)
        if c != 0:
            if self.tracer != None:
                self.tracer.error('Halting package %s failed' % self.name)
            return False
        self.tracer.info('Halting package %s succeeded' % self.name)
        return True



class SGeSAPsc(StorageConnectorClient):

    apiVersion = 2
    minSGversion = "12.00.51" # minimal Serviceguard version required

    # TODO: Feature bitmap
    # Define what to with the package of another persistence running locally (default: don't start own package)
    # FEATURE_INGORE_OTHER        0x001    - ingore a locally running pkg belonging to another persistence
    # FEATURE_STOP_OTHER          0x002    - stop  a locally running pkg belonging to another persistence
    # FEATURE_ENABLE_SWITCHING    0x004    - enable switching on local node if disabled (default: leave alone)
    # Define what to do if cluster and or node aren't up (default: do Nothing)
    # FEATURE_START_CLU_ON_ATTCH  0x010    - start cluster on attach (if configured and down)
    # FEATURE_START_NODE_ON_ATTCH 0x020    - start local node on attach (if configured and down)
    # Define how to determine persistence packages (default: by naming convention)
    # FEATURE_DISCOVER_PKGS       0x040    - discover pkgs instead of using naming convention

    def __init__(self, method = None, configLocation = None, sid = None, mnt_data = None, mnt_log = None, tracelevel = "INFO", partition = 1):
        """
        Initializes the Storage Connector Client.
          If method != None: the HDB environment will be used - all remaining parameters can be left untouched
          Else: the installer environment will be used

        @param method: the method to call, attach or detach
        @type method:  string
        @param configLocation: path to global.ini directory
        @type configLocation: string
        @param sid: the SID of the system
        @type sid: string
        @param mnt_data: basepath to data
        @type mnt_data: string
        @param mnt_log: basepath to log
        @type mnt_log: string
        @param tracelevel: the trace level
        @type tracelevel: string
        @param partition: partition number
        @type partition: int
        """

        self.tracelevel = 5

        try:
            import ServiceClient
            self.tracer = ServiceClient.Tracer("ha_%s" % self.__class__.__name__)
        except:
            self.tracer = InstallerTracer(tracelevel)

        if __name__ == '__main__':
            self.tracer = InstallerTracer("DEBUG")

#        if method != None:
#            # use InstallerTracer instead of ServiceClient.Tracer when method is called thru Connector
#            self.tracer = InstallerTracer(tracelevel)

        self.tracer.debug('>> init')

        # cfg and sid are required to init the cluster related objects which
        # in turn are required to fully init this objects - esp. when the methods
        # are already called in the superclass init. So the code lines below
        # are duplicated from superclass init (as well as the try/except above)
        if method == None:
            import ConfigMgrPy
            self._cfg = StorageConfigurationHDB()
            self.sid = ConfigMgrPy.sapgparam("SAPSYSTEMNAME")
        else:
            self._cfg = StorageConfigurationInstallation(configLocation)
            self.sid = sid
        # end of duplication

        self.tracer.debug('sid: '+self.sid)

        # Serviceguard related info
        # TODO: Throw exeption if cmcluster.conf can't be found
        #       and make sgsbin property of SGcluster
        configfile = open('/etc/cmcluster.conf','r').read()
        ini = configfile.find("SGSBIN")+7
        rest = configfile[ini:]
        search_enter = rest.find('\n')
        self.sgsbin = rest[:search_enter]+'/'
        self.tracer.debug('sgsbin: '+self.sgsbin)
        # get cluster (and node) information. may produce an exception
        self.sgcluster = SGcluster(self.tracer, self.sgsbin)
        self.localnode = os.uname()[1]  # TODO: check if this might return a FQDN also
        self.sgnode = SGnode(self.localnode,self.tracer, self.sgsbin)
        # SGNode(None,...) produces an exception

        self.pkgprefix = 'hdbv' # assume a naming convention (appliance-style)
        self.pkgpostfix = self.sid
        self.tracer.debug('pkgprefix: '+self.pkgprefix)
        self.tracer.debug('pkgpostfix: '+self.pkgpostfix)

        self.sgcluster.getSCpkgs(self.pkgprefix,self.pkgpostfix)

        self.tracer.debug('<< init')

        # leave rest up to superclass init
        super(SGeSAPsc, self).__init__(method, configLocation, sid, mnt_data, mnt_log, tracelevel, partition)
        #super(SGeSAPsc, self).__init__(*args, **kwargs)

        self.tracer.debug("tracer 'ha_%s' initialized" % self.__class__.__name__)


    def term(self):
        """Terminates the storage connector client"""

        self.tracer.debug("terminating %s..." % self.__class__.__name__)


    #########################################


    def about(self):
        self.tracer.debug('>> about')
        ver = {'provider_company':'Hewlett Packard Enterprise Development Company, L.P.',
               'provider_name'   :'SGeSAPsc',
               'provider_version':'12.10.00POC',
               'api_version'     :'2'  }
        self.tracer.info(ver['provider_version'])
        self.tracer.debug('<< about')
        return ver


    @staticmethod
    def sudoers():
        """
        Gives information about the necessary sudo permission to be set. Example:

            return "ALL=NOPASSWD: /bin/mkdir, /bin/chown, /sbin/multipath, /usr/bin/sg_persist, /bin/mount, /bin/umount"

        Permissions should be set as restricted as possible.
        @return: the string for /etc/sudoers
        @rtype: string
        """
        return ""

    def attach(self, storages):
        """
        Attaches storages on this host
        @param storages: storage descriptions {partition, usage_type, path}
        @type storages: a list of dicts: [{"partition" : ..., "usage_type" : ..., "path": ...}, ...]
        @return: information about success
        @rtype: int
        """
        self.tracer.debug('>> attach: '+str(storages))
        # TODO: make sure cluster an node are in a stable state (should related
        #       instances be reinitialized in order to allow SG installation,
        #       and/or adding node to cluster while HANA is running?)
        for part in set([storage['partition'] for storage in storages]):
            # default pkg name
            self.pkgname = self.pkgprefix+'%.2d' % int(storages.pop()['partition'])+self.pkgpostfix
            # try to get a package name from global.ini, e.g.
            #   partition_1_*_sgpkname = <your pkg name>
            # as log and data have to be in same pkg, * has to be used instead of _log_ or _data
            # (otherwise it'll be unsupported)
            connectionData = self._getConnectionDataForLun(part, "*")  # equivalent to "log" + "data"
            try:
                self.pkgname = connectionData["sgpkgname"]
            except:
                pass
            self.tracer.info('ha attach requested for partition %s of package %s' % (str(part), self.pkgname))

            sgpkg = SGsimplepkg(self.pkgname,self.tracer, self.sgsbin)
            if sgpkg.isInTransientState():
                self.tracer.info('Package %s is in transient state %s' % ( sgpkg.name, sgpkg.state ))
                if not sgpkg.waitForStableState():
                    self.tracer.error('Package %s didn\'t reach stable state. Current state: %s' % (sgpkg.name, sgpkg.state))
                    return 1 # TODO: revist return -- use break or exception / rollback if more than one pkg? (But SAP's fc Clients also don't roll back)
            if sgpkg.isInIgnoredState():
                self.tracer.warning('Package %s is detached or partially down and  will not be handled by this storage provider (current status: %s/%s/%s)'
                                    % (sgpkg.name, sgpkg.owner, sgpkg.status, sgpkg.state))
                return 0

            # TODO: enabled switching if disabled through a previous pkg failure ?
            #   e.g.: node:sgxsap78h|switching=disabled
            #   caused through
            #  The package script for hdbv01RH1 failed but suggested restarting. hdbv01RH1 can be restarted on a different node
            #  and resulting in a cmrunpkg:
            #  Unable to run package hdbv01RH1 on node sgxsap78h. Node is not eligible.

            # Package is in a stable state. Check who is the current owner (if any)
            # Do nothing if local node is the current owner. If unowned, start it.
            # If owned by another node, stop it there, then start.
            if sgpkg.owner == self.localnode:
                self.tracer.info('Package %s is already up on local node' % sgpkg.name)
            elif sgpkg.owner == 'unowned':
                if sgpkg.isInMaintenance():
                    self.tracer.warning('Package %s is in maintenance mode. Please start manually before HANA' % sgpkg.name)
                    return 1
                # package needs to be run. make sure not other sc package is up on local node
                self.sgcluster.retrieveConfiguredPkgs() # get current list of configured packages and their states
                for otherpkg in self.sgcluster.getSCpkgs(self.pkgprefix,self.pkgpostfix):
                    # TODO: in longer loop state of later pkgs might be outdated. update first?
                    if otherpkg.owner == self.localnode:
                        self.tracer.info('Found another storage controller package (%s) active in state %s owned by local node' % (otherpkg.name,otherpkg.state))
                    otherpkg.waitForStableState() # wait for a possible state change (transient -> stable)
                    if otherpkg.owner == self.localnode:
                        # another sc pkg is running/stable in this node
                        self.tracer.error('Other storage controller package (%s) is up stable on local node' % otherpkg.name)
                        self.tracer.error('Package %s won\'t be started' % sgpkg.name)
                        return 1
                sgpkg.run()
            else:  # owned by another node. Do not halt if pkg is in maintenance
                if sgpkg.isInMaintenance():
                    self.tracer.warning('Package %s is in maintenance mode on node %s. Package won\'t be halted'
                                        % (sgpkg.name,sgpkg.owner))
                    return 1
                self.tracer.info('Package %s is owned by %s. Package needs to be halted first' % (sgpkg.name,sgpkg.owner))
                if not sgpkg.halt():
                    return 1
                if not sgpkg.run():
                    return 1  # TODO: Try to run regardless if halt Failed ?
        self.tracer.debug('<<ha attach')
        return 0

    def detach(self, storages):
        """
        Detaches storages from this host
        @param storages: storage descriptions {partition, usage_type, path}
        @type storages: a list of dicts: [{"partition" : ..., "usage_type" : ..., "path": ...}, ...]
        @return: information about success
        @rtype: int
        """
        self.tracer.debug('>> detach: '+str(storages))
        # TODO: make sure cluster and node are in a stable state
        for part in set([storage['partition'] for storage in storages]):
            # TODO: quite some dupicate code here - with attach method
            self.pkgname = self.pkgprefix+'%.2d' % int(storages.pop()['partition'])+self.pkgpostfix
            # try to get a package name from global.ini, e.g.
            #   partition_1_*_sgpkname = <your pkg name>
            # as log and data have to be in same pkg, * has to be used instead of _log_ or _data
            # (otherwise it'll be unsupported)
            connectionData = self._getConnectionDataForLun(part, "*")  # equivalent to "log" + "data"
            try:
                self.pkgname = connectionData["sgpkgname"]
            except:
                pass
            self.tracer.info('ha detach requested for partition '+str(part)+' of package '+self.pkgname)

            sgpkg = SGsimplepkg(self.pkgname,self.tracer, self.sgsbin)
            if sgpkg.isInTransientState():
                # TODO: maybe transient state can be shortcutted if owner is another node, because then either
                #       the package starts up (and owner stays) or stops (and is unowned afterwards)
                self.tracer.info('Package %s is in transient state %s' % ( sgpkg.name, sgpkg.state ))
                if not sgpkg.waitForStableState():
                    self.tracer.error('Package %s didn\'t reach stable state. Current state: %s' % (sgpkg.name, sgpkg.state))
                    return 1 # TODO: return 0 / exception ?
            if sgpkg.isInIgnoredState():
                self.tracer.warning('Package %s is detached or partially down and  will not be handled by this storage provider (current status: %s/%s/%s)'
                                    % (sgpkg.name, sgpkg.owner, sgpkg.status, sgpkg.state))
                return 0
            #if sgpkg.isUnmanaged():
            #    self.tracer.warning('Package %s is unmanaged. Current status: %s/%s/%s' % (sgpkg.name, sgpkg.owner, sgpkg.status, sgpkg.state))
            #    return 1 # or break ?

            if sgpkg.owner == self.localnode:
                if sgpkg.isInMaintenance():
                    self.tracer.warning('Package %s is in maintenance mode on %s. Please stop it manually after stopping HANA'
                                        % (sgpkg.name,sgpkg.owner))
                    return 1 # TODO: or return 0 to pretend that detach was successful?
                if not sgpkg.halt():
                    return 1
            elif sgpkg.owner == 'unowned':
                self.tracer.info('Package %s is already down' % sgpkg.name)
            else:
                # up somewhere else
                self.tracer.warning('Package %s is owned by node %s and will not be halted'
                                     % (sgpkg.name,sgpkg.owner))
        self.tracer.debug('<< ha detach')
        return 0

    def stonith(self, hostname):
        """
        Runs SHOOT THE OTHER NODE ON THE HEAD for <hostname>
        @param hostname: the failing host to be STONITHed
        @return: information about success
        @rtype: int
        """
        return 0

def main(testpartition=1):
    """
    Sample main program simulating a HDB call, without actually running it with HDB.
    """
    sgesap = SGeSAPsc()
    sgesap.about()

    # sgesap.check_SGversion() # TODO: result ignored so far

#    try:
    sgcluster = SGcluster(sgesap.tracer, sgesap.sgsbin)
    sgcluster.check_version(SGeSAPsc.minSGversion) # TODO: result ignored so far
    sgcluster.retrieveConfiguredPkgs()
    #sgcluster.getSCpkgs('hdbv' , os.environ["SAPSYSTEMNAME"])
    sgcluster.getSCpkgs(sgesap.pkgprefix,sgesap.pkgpostfix)
    #    sgnode = SGnode('sgxsap78h',sgesap.tracer, sgesap.sgsbin)
        #sgnode = SGnode(None,sgesap.tracer, sgesap.sgsbin) # produces exception(?)
#    except Exception:
#        print "Exception occurred"

    try:
    #    spkg1 = SGsimplepkg("hdbpRH1",sgesap.tracer, sgesap.sgsbin)
        spkg2 = SGsimplepkg("nonexisting",sgesap.tracer, sgesap.sgsbin)
    except Exception:
        print "Exception occurred"

    spkg3 = SGsimplepkg("hdbv01%s" % os.environ["SAPSYSTEMNAME"], sgesap.tracer, sgesap.sgsbin)
    sgesap.tracer.info('Package %s has state %s/%s/%s' % (spkg3.name,spkg3.owner,spkg3.status,spkg3.state))
    if spkg3.isInTransientState():
        sgesap.tracer.info('Package %s is in transient state' % spkg3.name)
    else:
        sgesap.tracer.info('Package %s is in stable state' % spkg3.name)

    if len(sys.argv)-1:
        testpartition = sys.argv[1]
        if not testpartition.isdigit():
            print ('Usage: ' + sys.argv[0] + ' [partition]')
            return (-1, 'Usage: ' + sys.argv[0] + ' [partition]')

    # For testing unowned or owned by another node, call a detach before attach.
    # path isn't used by attach/detach as it is configured in package. However, it could be used to extract the SID
    sgesap.detach([{"partition" : int(testpartition), "usage_type" : "DATA", "path" : "/usr/sap/HA1/global/hdb/data/mnt%.5d" % int(testpartition)},
                   {"partition" : int(testpartition), "usage_type" : "LOG", "path" : "/usr/sap/HA1/global/hdb/log/mnt%.5d" % int(testpartition)}])

    if sgesap.attach([{"partition" : int(testpartition), "usage_type" : "DATA", "path" : "/usr/sap/HA1/global/hdb/data/mnt%.5d" % int(testpartition)},
                          {"partition" : int(testpartition), "usage_type" : "LOG", "path" : "/usr/sap/HA1/global/hdb/log/mnt%.5d" % int(testpartition)}]) != 0:
        exit()

    time.sleep(10)

    sgesap.detach([{"partition" : int(testpartition), "usage_type" : "DATA", "path" : "/usr/sap/HA1/global/hdb/data/mnt%.5d" % int(testpartition)},
                   {"partition" : int(testpartition), "usage_type" : "LOG", "path" : "/usr/sap/HA1/global/hdb/log/mnt%.5d" % int(testpartition)}])
    del sgesap

if __name__ == '__main__':
    main()
