#!/usr/bin/perl
#"@(#) SGeSAP/LX - 12.10.00 - $Rev: 1723 $ - Last modified 05/11/17"
# **********************************************************************
# *                                                                    *
# * SGeSAP SAP Configuration Detection                                 *
# *                                                                    *
# * (c) Copyright 2016-2017 Hewlett Packard Enterprise Co.             *
# * (c) Copyright 2011-2015 Hewlett-Packard Enterprise Corp.           *
# *                                                                    *
# **********************************************************************
##############################################################################
#
# Remarks:
# - support VxVM /dev entry is in form /dev/vx/dsk/<vx dgrp>/<vx vol>
#    -- if the disk is "shared" it is considered CFS.
# - mount point could also be /usr/sap/<SID> instead of a list of /usr/sap/<SID>/<INST>
#   This, however, makes only sense if *all* instances of <SID> will go into one
#   package (not sure if it makes sense at all as the "local" part, e.g. SYS is then
#   only accessible via /sapmnt/<SID> and not via /usr/sap/<SID>/SYS anymore, but it
#   could exist locally and then be mounted over). So far, the script doesn't check
#   for mountpoints /usr/sap/SID or even /usr/sap
# - hotstandby livececache parameters to be added
#
# This script also contains code used for easy deployment
#  Parameter: "easy" + a string containing a comma seperated list of
#  name/value pairs with predefined names sap_system (0,1), sap_instance (0,n)
#  if sap_system is defined, db_instance(0,1)
#  Allowing a list of SAP instance now for getsapinstance for this feature
#  This feature will basically print out the discovered information in
#  package ascii format (grouped by module)
#
# Output of attributes related to generic resource usage for enqor/SCS/ERS
# on Linux is based on the setting of environment variabe SGeSAP_ENQOR_PKGNAME.
# If set, it'll trigger the output of related generic_resource and dependency
# attributes and use the value as package name in the dependency_condition.
#
# Restrictions
# - no cross subnet related IP output format
#
#############################################################################

my $USAGE = <<"END_USAGE";
Usage:
  sapgetconf.pl sap_global
  sapgetconf.pl sapinstance  <SAPSID> [<instance>,...]
  sapgetconf.pl dbinstance  [<DBSID>]
  sapgetconf.pl hdbinstance [<HDBSID>]
  sapgetconf.pl lcinstance  [<LCSID>]
  sapgetconf.pl mdminstance [<MDMSID>]
  sapgetconf.pl apiversion
  sapgetconf.pl nfs [pattern]
END_USAGE
# Internal "easy" option - for deploysappkgs usage only
#  sapgetconf.pl easy [sapsid=<SAPSID[,sapinstance=<SAPinst>]*]
#                       [dbinstance=<DBSID>][lcinstance=<LCSID>][nfs=<1|<filter>>]
#                       [hdbinstance=<HDBSID>]

#############################################################################

# SAP_functions module.
# Three different locations. HP-UX,SLES RHEL
use lib qw(
   /opt/cmcluster/sap/lib
   /opt/cmcluster/lib
   /usr/local/cmcluster/lib
);
use SAP_functions;

use Socket;
use FileHandle;
use Carp;
use File::Find;
use File::Basename;
use POSIX;  #uname

#use Switch; # not use, seems to have quite a performance hit at startup
# when statement is available with Perl V5.10, but HP-UX version might be lower


####################################################
# get the runtime environment
####################################################
setup_script_runtime();

# parameter "timeout" used by handler for mainDispatcher
$timeout=getparam('timeout');
$apiversion=getparam('apiversion');
$servoutsgmgr=getparam('servoutsgmgr');

# Global vars for paths of SG conf and SGeSAP bin and services directory
$SGCONFDIR;  # used to check modules
$SGeSAPBINDIR; # mainly nwrksupp, (cmenqord defined externally in EMS)
$SGeSAPSVCDIR; # directory where services/monitors reside

# Comma separated list of network prefixes for which
# apiversion |= 0x40 should be enabled
$xsubnetnwlist=$ENV{"SAPGETCONF_XSUBNWLIST"} || "15.,16.";


# will be used as a glob of SAP SID directories in /sapmnt
my  $SAPSIDSInSapmnt="/sapmnt/[A-Z][A-Z,0-9][A-Z,0-9]";
my  $SAPSIDSInusrsap="/usr/sap/[A-Z][A-Z,0-9][A-Z,0-9]/SYS";
my  $HDBSIDInhshared="/hana/shared/[A-Z][A-Z,0-9][A-Z,0-9]";

# List of symbolic instance filters (-f <symb.inst.filter>) . So far only "SCS" is supported, rest is inofficial
# Specifications allow reg. expressions
my %InstanceFilters= (
scs => "(SCS|ASCS|ERS|DVEBMGS)",
WDP => "W",
ERS => "ERS",
HDB => "HDB",
  );
# currently, filtertype "scs" returns both ASCS and DVEBMGS if both are configured
# define the default filter (undef = no filter used)
my $InstanceActiveFilter=undef;

# global vars for MaxDB/LiveCache
my $maxdbuc = "c";
my $maxdboldparnames = 0;
my %maxdbpars=();
my %maxdbnewoldmap = (
"KernelVersion",  "KERNELVERSION",
"InstanceType",   "INSTANCE_TYPE",
"MaxDataVolumes", "MAXDATAVOLUMES",
"DataVolumeName", "DATA_VOLUME_NAME_",
"DataVolumeType", "DATA_VOLUME_TYPE_",
"MaxLogVolumes",  "MAXLOGVOLUMES",
"LogVolumeName",  "LOG_VOLUME_NAME_",
"LogVolumeType",  "LOG_VOLUME_TYPE_",
"HotStandbyStorageDLLPath", "HS_STORAGE_DLL",
"OfficialNodeName", "OFFICIAL_NODE",
"HotStandbyNodeName001", "HS_NODE_001",
"HotStandbyNodeName002", "HS_NODE_002"
);

# optional SGeSAP network support utility
# (default is HPUX hardcoded path)
my $nwrksupp = "/opt/cmcluster/sap/bin/nwrksupp";

#
# alerts related parameters
# (allows external overwrite)
#
my $alrtbasedir = $ENV{SAPGETCONF_ALRTBASEDIR};
$alrtbasedir="/var/tmp/.sgesap" unless $alrtbasedir;
my $alrtlockexptime = $ENV{SAPGETCONF_ALRTLOCKEXPTIME};
$alrtlockexptime=3 unless $alrtlockexptime;  # seconds
my $alrtlonglifetime = $ENV{SAPGETCONF_ALRTLONGLIFETIME};
$alrtlonglifetime=300 unless $alrtlonglifetime; # seconds
#
my $alrtlockpostfix=".lock";

my $SG_PACKAGE="";

# end of alerts related parameters

# usage flag for generic resources (so far Linux only)
# initialized as "not used", set to used when a generic resource
# is printed out. used in printing out dependecies.
my $genres_used=0;

# some easy deployment (ED) related variables
my $pkgdescr="SGeSAP";
my %EDmodlist=();

# package priority, used for some kind of packages
my $pkgprio;

#############################################################################
#
# Collect data on SAP instances that belong to $SID. getsapinstance() is
# the core function for this. Basically getsapinstance() is invoked by
# running script:  sapgetconf.pl sapinstance MO7.
#
# Values are stored in various arrays and hashes
# (no output is produced anymore)
#
# Prereq: Access to NFS (/sapmnt/<SID>) has been checked.
#
# Arguments:
#   $SID       Name of a SAP SID
#   $filtertype Instance types to be considered (optional)
#
#############################################################################

sub getsapinstance {

    LogEnterSub( @_ );
    $SID = shift;
    my $filtertype = shift; # maybe undef
    # my @SAPinsts = @_;

    #######################################
    # set up global variables and hashes
    #######################################

    # Initialize %addrs      maps IP          -> subnet,
    # Initialize %subnets    maps subnet      -> IP
    %addrs   = ();
    %subnets = ();

    # Initialize %mpsfs      maps mountpoints -> filesystems
    # Initialize %mpsfst     maps mountpoints -> filesystem types
    # Initialize %vxdgshared maps ...
    %mpsfs      = ();
    %mpsfst     = ();
    %vxdgshared = ();

    # Initialize @profiles, @ersprofiles containing SAP profile names
    @profiles=();
    @ersprofiles=();

    # Initialize %ERSmap     maps ERS instance -> SCS instance
    # Initialize %SCSmap     maps SCS instance -> ERS instance
    %ERSmap = ();
    %SCSmap = ();

    # Iniitialize %SMDAVHmap (SolMan Diag Agent virt. hosts map)
    # makes sure that the SMDA agent for a virt. host is only searched once
    %SMDAVHmap = ();

    # directory containing the SAP profiles for the $SID instance
    $profiledir = "/sapmnt/$SID/profile";

    #######################################
    # Start collecting data
    #######################################

    # this subroutine fills in the %addrs, %subnets hashes
    CreateIPAddressAndSubnetAssocArray();

    # this subroutine fills in the %mpsfs, %mpsfst, %vxdgshared hashes
    # previously "/usr/sap/$SID" has been used as filter, however as there
    # may be now "associated" SMDA's belonging to another (DA) SID only
    # "/usr/sap" is used to catch that DA SID also
    CreateMountpointAssocArray("/usr/sap/");

    # Scan SAP profiles to determine SAP instances on the system
    # subroutine fills in arrays @profiles and @ersprofiles
    # (access to profile dir (over NFS) has been checked before calling this subroutine)
    # For filtertype "scs" this could contain an ASCS instance *and* a DVEBMGS instance.
    # TODO ENHANCEMENT:
    # TODO In this case DVEBMGS should be filtered out also (might also be done
    # TODO on the "receiving side").
    CollectListOfSIDProfiles( $SID, "$profiledir", $filtertype );


    # this subroutine fills in the %ERSmap and %SCSmap hashes
    CreateERSandSCSAssocArray();

    #######################################
    # SGMgr ouput - moved
    #######################################

}   # End getsapinstance()

#############################################################################
#
# Collect data on database instances that belong to $SID. getdbinstance() is
# the core function for this. Basically getdbinstance() is invoked by
# running script:  sapgetconf.pl dbinstance MO7.
#
# Arguments:
#   $SID        Name of a SAP SID (or DB SID)
#
#############################################################################

sub getdbinstance {

    LogEnterSub( @_ );
    my $SID    = shift;

    #######################################
    # set up global variables and hashes
    #######################################

    # Initialize %addrs      maps IP          -> subnet,
    # Initialize %subnets    maps subnet      -> IP
    %addrs   = ();
    %subnets = ();

    # Initialize %mpsfs      maps mountpoints -> filesystems
    # Initialize %mpsfst     maps mountpoints -> filesystem types
    # Initialize %vxdgshared maps ...
    %mpsfs      = ();
    %mpsfst     = ();
    %vxdgshared = ();

    # directory containing the SAP profiles
    my $profiledir = "/sapmnt/$SID/profile";

    # the default SAP profile
    my $DEFAULTPFL = $profiledir . "/DEFAULT.PFL";

    # convert to lower case and set the sidadm user
    $sidadm = lc($SID) . "adm";

    # convert to lower case and set the ora admin user
    $oraadm = "ora" . lc($SID);

    # Some configs might have a different SID for the DB
    my $dbsid = $SID;

    #######################################
    # Start collecting data
    #######################################

    # get API version to figure out which DBs are supported
    # call only if it is still 0
    $apiversion = getapiversion() if $apiversion==0;

    # determine type of database: currently ADA/MaxDB or Oracle or DB2
    # $dbsid returned might be different from $SID for a MCOD config
    my ($dbsid,$dbtype) = getDbtype( $SID , $DEFAULTPFL );

    # in case of unsupported dbtypes dbsid is also invalidated,
    # return immediately to avoid unnecessary processing & output
    return unless $dbsid;

    # check if ADA/MaxDB is of type "LVC" or of type "OLTP"
    my $maxdbtype;
    if ( $dbtype eq "ADA" ) {
      setMaxDBUC($dbsid);
      $maxdbtype = getMaxDBtype($dbsid);
      getMaxDBParams($dbsid);
    }

    # fallback to OLTP (standard DB type)
    $maxdbtype = "OLTP" unless ($maxdbtype);

    #"maxdbtype" also "re-"used for HANA replication mode
    # will contain mode and site_name from global.ini's [system_replication] (if defined)
    if ( $dbtype eq "HDB" ) {
      my $hprofiledir = "/hana/shared/$dbsid/profile"; # HANA
      $maxdbtype = gethdbreplmode($dbsid);
      $DEFAULTPFL= $hprofiledir . "/DEFAULT.PFL"; # profile override
    }
    # return undef if maxdbtype (in this case replication mode) is undef
    # - indicating that global.ini wasn't found (and therefore no HANA is installed)
    return (undef,undef,undef,undef) unless defined $maxdbtype;

    # get the SAP database host virtual hostname
    my $vhost = GetSAPDbhost( $DEFAULTPFL, $dbtype, $dbsid );

    # fill in the %addrs and %subnets hashes
    CreateIPAddressAndSubnetAssocArray();

    # Fill in the mountpoints. Depends on database type:
    #   ORA -> everything starting with /oracle/<dbsid>
    #   ADA -> everything starting with /sapdb/<dbsid>
    # Restriction: Raw devices. They can probably only figured out
    #   by querying the DB itself. In case of "ADA" there may also
    #   be softlinks pointing to the raw devices
    $dbtype eq "ORA" && CreateMountpointAssocArray("/oracle/$dbsid");
    $dbtype eq "ADA" && CreateMountpointAssocArray("/sapdb/$dbsid");
    $dbtype eq "DB2" && CreateMountpointAssocArray("/db2/($dbsid|db2\L$dbsid\E)");
    $dbtype eq "SYB" && CreateMountpointAssocArray("/sybase/$dbsid(/|\$)"); # avoid <DBSID>_REP being reported

    LogReturnSub (($dbsid,$dbtype,$maxdbtype,$vhost));

    return ($dbsid,$dbtype,$maxdbtype,$vhost);

}    # End getdbinstance()



#############################################################################
#
# Create a "opposite mode" HDB replication instance
#
# Arguments:
#   $dbsid       SID of DB database
#   $dbtype
#   $replmode
#   $hdbhosts    List of hosts to which hana is configured
#
#   This function only return something if dbtype="hana" and hdbhosts
#   containts more than one member. It'll then make a copy of the provided
#   values and replicate *sync* by primary and primary by sync in replmode.
#
# Returns:
#   Array containing $dbsid (usually $SID), $dbtype ("HDB" here) and
#   replication mode (primary, sync or none) and hostlist
#
#############################################################################

sub createDBSOvalues {
	LogEnterSub( @_ );

	my $dbsid = shift;
	my $dbtype = shift;

	return unless ($dbtype eq "HDB"); # not "hana"

	my $replinfo = shift;
	my $hdbhosts = shift;

	return unless (scalar split(",", $hdbhosts) > 1);

	my ($mode, $siteid, $sitename) = split(",",$replinfo);
	if ($mode eq "primary") {
		$replinfo = "sync,$siteid,$sitename"; # "sync" doesn't have to match with other site
	}
	elsif ($mode =~ /sync/) {
		$replinfo = "primary,$siteid,$sitename";
	}
	else {
		return; # no mode
	}

	LogReturnSub (($dbsid,$dbtype,$replinfo,$hdbhosts));
	return ($dbsid,$dbtype,$replinfo,$hdbhosts);
}

#############################################################################
#
# Get HANA DB info by scanning /sapmnt/<SID>/profile for *HDB* profiles
#
# Arguments:
#   $SID        SID of HDB database
#
# Returns:
#   Array containing $dbsid (usually $SID), $dbtype ("HDB" here) and
#   replication mode (primary, sync or none), hana-host(s) and vhost(s)
#
# forcedmode is used to simulate the other side of the actually configured
# replication by talking it's identity (primary -> secondary or secondary->primary)
# forcedmode gets used when sapgetconf is called with either <SID>s or <SID>p
# of the "other" side of the replication to generate SG package values.
# this maps into variable forcedmode being set to either "primary" or "sync"
#  for example: a primary site has it's own IP addresses for access. the same
# is true for the secondary (as of HANA 2.0) which allows read-only access of
# the secondary.
#############################################################################

sub gethdbinstance {

    LogEnterSub( @_ );
    my $SID    = shift;
    my $forcedmode = shift;
    my $identity = "secondary";   # set a default for identity
    my $hdbnodes = "";
    my $vhost  = "";
    my $vhost_def = "hdb".$SID; # default assumed vhost hdb<SID>

    #######################################
    # set up global variables and hashes
    #######################################

    # Initialize %addrs      maps IP          -> subnet,
    # Initialize %subnets    maps subnet      -> IP
    %addrs   = ();
    %subnets = ();

    # Initialize %mpsfs      maps mountpoints -> filesystems
    # Initialize %mpsfst     maps mountpoints -> filesystem types
    # Initialize %vxdgshared maps ...
    %mpsfs      = ();
    %mpsfst     = ();
    %vxdgshared = ();

    # directory containing the SAP profiles
    my $profiledir = "/sapmnt/$SID/profile";

    # the default SAP profile
    my $DEFAULTPFL = $profiledir . "/DEFAULT.PFL";

    # convert to lower case and set the sidadm user
    $sidadm = lc($SID) . "adm";

    # Some configs might have a different SID for the DB
    my $dbsid = $SID;

    my $startIpAddress;
    #######################################
    # Start collecting data
    #######################################

    # get API version to figure out which DBs are supported
    # call only if it is still 0
    $apiversion = getapiversion() if $apiversion==0;

    my $dbtype="";
    @profiles=();
    CollectListOfSIDProfiles( $SID, "/hana/shared/$SID/profile", "HDB" ); # need to check profiles for filtering, writes @profiles
    # Fallback to /sapmnt/$SID, if nothing found
    if (scalar @profiles == 0 ) {
      CollectListOfSIDProfiles( $SID, "/sapmnt/$SID/profile", "HDB" ); # need to check profiles for filtering, writes @profiles
    }

    # if array @profiles > 0 then it must contain a HANA instance
    if (scalar @profiles > 0 ) {
          $dbtype="HDB";
    }

    # in case of unsupported dbtypes dbsid is also invalidated,
    # return immediately to avoid unnecessary processing & output
    return unless $dbtype;

    # get list of hana nodes from profiles (WVE: duplicate)
    foreach $prof (@profiles) {
      my $profname = ( split( /\//, $prof ) )[-1]; # last part
      my ( $sid, $inst, $hhost ) = split( /_/, $profname );
      LogVerbose ("\$hhost=$hhost");
      if ($hdbnodes eq "") {
      	$hdbnodes = $hhost;
      } else {
      	$hdbnodes = "$hdbnodes,$hhost";
      }
    }
    # WVE: doublecheck a list of nodes with output from
    #      - systemOverview.py -> System -> Distributed -> No/Yes
    #      - landscapeHostConfiguration.py number of nodes

    # note:  example hdbrepl returned: [primary,2,siteB,]
    my $hdbrepl = gethdbreplmode ($SID);
    $identity = "primary" if ($hdbrepl =~ /^primary/);

    # if forcedmode is set then use that identity
    $identity = $forcedmode if ($forcedmode =~ m#primary|sync#);

    # - if hdb<SID> is defined in /etc/hosts then use this for start of ip
    #   hdb_ip_address range... Note: only valid for primary
    # - if environment SGeSAP_START_IP_ADDRESS is set then override hdb<SID> and
    #   use this  for start of ip hdb_ip_address range... Note: valid for primary
    #   and secondary
    if (defined ($ENV{"START_IP_ADDRESS"})) {
      my ($startPriIpAddress, $startSecIpAddress) = split(/,/, $ENV{"START_IP_ADDRESS"});
      LogVerbose("hdbrepl=[$hdbrepl] startPriIpAddress=[$startPriIpAddress] startSecIpAddress=[$startSecIpAddress]");

      if ($identity =~ /^primary/) {
        $startIpAddress=$startPriIpAddress;
      } else {
        $startIpAddress=$startSecIpAddress;
      }
      LogVerbose("Using startIpAddress=[$startIpAddress]");

      # check if startIpAddress can be resolved
      `$nwrksupp ghbn $startIpAddress 2>/dev/null`;
      $vhost = $startIpAddress unless $?;

    } else {
      # IP address via hdb<SID> in /etc/hosts was only supported for primary
      if ($identity =~ /^primary/) {
        # check if default vhost can be resolved
        `$nwrksupp ghbn $vhost_def 2>/dev/null`;
        $vhost = $vhost_def unless $?;
      }
    }
    LogVerbose("using vhost=[$vhost] as start for hdp_ip_address range");

    # introducing global var ($hdbukenv and) $hdbuk
    $hdbuk = "${dbsid}SAPDBCTRL";
    my $hdbukenv = gethdbuserkey($dbsid, "$hdbuk");
    if ($hdbukenv eq "") {
      # Fallback to DEFAULT
      $hdbuk = "DEFAULT";
      $hdbukenv = gethdbuserkey($dbsid, "$hdbuk");
    }

    if ($hdbukenv ne "") {
      # extract vhost from userkey env, but only if it hasn't been
      # set yet by the nwrksupp check
      if ($vhost eq "") {
        my $vhostuk = (split(':',$hdbukenv))[0];
        LogVerbose ("hdb uk vhost: $vhostuk");
        if ((defined $vhostuk) && ($vhostuk ne "") && ($vhostuk ne $ENV{"HOSTNAME"})) {
          # Assume that this is the relocateable host
          # Further checks might be needed!!
          $vhost = $vhostuk;
        }
      }
    }
#    else {
      # Have formatting routines to skip output of userkey attributes as this was
      # removed from first release module spec.
      $hdbuk = "";
#    }

	# collect priorities all packages in the cluster. Generate
    # new priorties for the hdb primary and secondary package:
	# Primary instance get < 1000
	# Secondary instance get >= 100
    # However, collect them only if also a dependency is needed,
    # which is the case when SGeSAP_HDBREP_PKGNAME is set

    if (defined ($ENV{"SGeSAP_HDBREP_PKGNAME"})) {
      %pkgpriorities=(); # hash
      collectpriorities($hdbrepl);
      $pkgprio=sethdbpriority ("$hdbrepl");
    }

    LogReturnSub ("$dbsid,$dbtype,[$hdbrepl],[$hdbnodes],$vhost");

    return ($dbsid,$dbtype,$hdbrepl,$hdbnodes,$vhost);
}

#############################################################################
#
# Get HANA DB replication info
#
# Reads global.ini from /usr/sap/<SID>/global/custom/config
#
# Scans for
#  [system_replication]
#  ...
#  actual_mode = primary | *sync*
#  ...
#  site_name = <sitename>
#
# Arguments:
#   $SID        SID of HDB database
#
# Returns: list of replication properties. Consists of 4 elements
#   taken from global.ini (could be undef or list of empty strings)
#   1. actual_mode: "primary" or "sync" or whatever comes after "actual_mode="
#   2. sitename
#   3. siteid
#   4. fullsync (false/true or empty, default empty)
#
#############################################################################

sub gethdbreplmode {

    LogEnterSub( @_ );
    my $SID     = shift;
    my $globini = "/usr/sap/${SID}/SYS/global/hdb/custom/config/global.ini";
    my $replmode;
    my $sitename, $siteid;
    my $fullsync;

    unless (-r $globini) {
      warn "$globini not found";
      LogReturnSub ("undef");
      #return ($replmode);
      return undef;
    }

    $replmode = "none";
    $sitename = "";
    $siteid = "";
    $fullsync = "";

    open FILE,'<',$globini || warn "Couldn't open $globini";
    my $sysrepsection=0;
    while (<FILE>) {
      #print $_;
      if ($_ =~ m#\[system_replication\]#) {
        #LogVerbose("sysrepsection=1");
        $sysrepsection = 1;
        #$fullsync = "false"
        next;
      }
      if ($sysrepsection) {
        if ($_ =~ m#^actual_mode\s*=\s*(\w+)\s*#) {
          $replmode = $1;
          LogVerbose("\$replmode=$1");
          # last; # found
        }
        if ($_ =~ m#^site_id\s*=\s*(\w+)\s*#) {
          $siteid = $1;
          LogVerbose("\$siteid=$1");
        }
        if ($_ =~ m#^site_name\s*=\s*([\w-]+)\s*#) {
		  # pattern allows dashes in global.ini site_names
          $sitename = $1;
          LogVerbose("\$sitename=$1");
          #last; # all found (assume mode before site)
        }
        if ($_ =~ m#^enable_full_sync\s*=\s*(\w+)\s*#) {
          $fullsync = $1;  # usually just true or false (as string)
          LogVerbose("\$fullsync=$1");
        }

        if ($_ =~ m#\[\w+\]#) {
          $sysrepsection = 0;
          last; # next section -> not found
        }
      }
    }
    close FILE;

    LogReturnSub ("$replmode,$siteid,$sitename,$fullsync");
    return ("$replmode,$siteid,$sitename,$fullsync");
}
#############################################################################
#
# Get HANA DB userkey environment
#
# Get the hdbuserkey ENV
#
# Arguments:
#   $SID        SID of HDB database
#   $hdbuk      Userkey to check
#
# Calls external programs hdbuserstore (and hdbsql - later ).
#
# Returns:ENV contents or emtpy string
#
#############################################################################
sub gethdbuserkey {

  LogEnterSub( @_ );
  my $SID    = shift;
  my $hdbuk  = shift;
  # my @uklist = ("${SID}SAPDBCTRL","DEFAULT");
  my $sidadm = lc($SID)."adm";
  my $hdbukenv = "";

#  foreach my $uk (@uklist) {
    # print "$uk\n";
    my $cmd=qq[su - $sidadm -c "hdbuserstore LIST $hdbuk " 2>/dev/null |];
    LogVerbose("Execute \$cmd=[$cmd]");
    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
       last if ( $_ =~ "KEY $hdbuk NOT FOUND" );
       if ($_ =~ /^KEY $hdbuk$/) {
         LogVerbose "Key $hdbuk found";
         # $hdbuk = $uk;
         # assume that ENV: comes next
         if (<$fh> =~ /ENV\s*:\s*(\S+)/) {
            $hdbukenv = $1;
            LogVerbose "\$hdbukenv = $hdbukenv";
            last;
         }
       }
    }
    # close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";
    close($fh);
#  }
  LogReturnSub("\$hdbukenv=[$hdbukenv]");
  return ($hdbukenv);
}

#############################################################################
#
# Collect information about the replication. So far Sybase only.
#
# So far it is also assumed that the replication has no own IP address
# (no package_ip reported) and that the service is is already added
# by the rs module.
#
# Arguments:
#   $SID        Name of a SAP SID (or DB SID or <DBSID>_REP)
#
#############################################################################

sub getrepdbinstance {
  LogEnterSub( @_ );
  my $SID    = shift;
	my $repdbsid = $SID;
	my $repdbrole = "none";


	$repdbsid .= "_REP" if ($SID =~ /^[A-Z][A-Z,0-9][A-Z,0-9]$/);

	my $repdbtype = getDbtypeFromMountpoint ($repdbsid);

	LogVerbose("\$repdbsid=[$repdbsid], \$repdbtype=[$repdbtype], \$repdbrole=[$repdbrole]");

	return unless $repdbtype eq "SYB";

	# Figure out "replication role" (none, primary,standby)
	# Could also check for -x RUN_$repdbsid_primary|standby.sh
    # Remark: Checking for hardcoded "primary" and "standby" is obsolete
    #   with Carina release. This release allows arbitray labels.
    my $rs_prefix = "/sybase/".$repdbsid."/".$repdbsid;
#	if (-d $rs_prefix."_primary") {
#	  $repdbrole = "primary";
#	}
#	elsif (-d $rs_prefix."_standby") {
#	  $repdbrole = "standby";
#	}
#    else {
      # glob it. should return 0 or 1 result
      my @repservers = <${rs_prefix}_*>;
      my $numrs = @repservers;

      if ( $numrs == 1 ) {
        LogVerbose ("repserver found: $repservers[0]");
        my $lastpart = ( split( /\//, $repservers[0] ) )[3];
        my ($dc,$dc,$label) =  split ( /_/, $lastpart);
        LogVerbose ("\$repdbsid=[$repdbsid], \$label=[$label]");
        #$repdbrole = $label if length ($label // '');  # requires Perl 5.10
        $repdbrole = $label if length (defined($label) ? $label : '');
      }
      elsif ( $numrs == 0 ) {
        warn "No replication server found for $repdbsid (maybe filesystem isn't mounted";
      }
      else {
        warn "More than one repserver found for $repdbsid - ignoring them";
      }
#    }

	LogVerbose("\$repdbsid=[$repdbsid], \$repdbtype=[$repdbtype], \$repdbrole=[$repdbrole]");

	return if $repdbrole eq "none";

	# IP and FS
	# vhost assumes SYB
	my ($vhost,$vport) = getVdbhostFromInterfaces ($repdbsid."_".$repdbrole,"/sybase/$repdbsid/interfaces");

	CreateMountpointAssocArray("/sybase/$repdbsid");

    LogReturnSub (($repdbsid,$repdbtype,$repdbrole,$vhost,$vport));

	return ($repdbsid,$repdbtype,$repdbrole,$vhost,$vport);

}  # End getrepdbinstance()

#############################################################################
#
# Collect all SAP LiveCache instances on a system. Skip if database is
# MaxDB OLTP.
# getlcinstances() is the core function for this. Basically getlcinstances()
# is invoked by # running script:  sapgetconf.pl lcinstance.
#
# Produces output in SGMgr (ServiceGuard Manager) format
#
#############################################################################

sub getlcinstances {
    LogEnterSub( @_ );

    #######################################
    # set up global variables and hashes
    #######################################

    # Initialize %mpsfs      maps mountpoints -> filesystems
    %mpsfs = ();

    # parent directory for all livecache installations
    my $sapdbdir = "/sapdb";

    #######################################
    # Start collecting the data
    #######################################

    LogVerbose("Collecting db instances in sapdbdir=[$sapdbdir]");

    # Collect MaxDB database instances beneath the /sapdb directory
    opendir( DIR_HANDLE, $sapdbdir )
        or warn "Couldn't opendirectory sapdbdir=[$sapdbdir]\n";
    my @filelist = readdir(DIR_HANDLE);
    LogVerbose("Found files \@filelist=[@filelist]");
    closedir(DIR_HANDLE);

    foreach (@filelist) {
        # skip if filename is "data","programs","client" or "."
        next if (/[.]/);
        next if (/^(data|programs|clients)$/);

        my $dbsid = $_;
        LogVerbose("*****************************************");
        LogVerbose("Collecting data for instance \$dbsid=[$dbsid]");

        ($dbsid,$sapdbhost) = getlcinstance ($dbsid);
        #######################################
        # Create SGMgr ouput
        #######################################

        # output the liveCache information required by SGMgr
        FormatLivecacheForSGMgr( "$dbsid", "$sapdbhost" );
    }   # end foreach (@filelist) loop
}

#############################################################################
#
# Collect information abot a signle SAP LiveCache instance. Skip if database is MaxDB OLTP.
#
# Paramters: $LCSID
#
# Returns $dbsid and $sapdbhost (and hssenabled later)
#
#############################################################################

sub getlcinstance {
    LogEnterSub( @_ );

    $dbsid = shift;

    # Convert $dbsid to lowercase and set the $sidadm user.
    $sidadm = lc($dbsid) . "adm";

    # Skip further processing if $sidadm not found.
    return unless getpwnam($sidadm);

    # set (or leave) MaxDB user key
    setMaxDBUC($dbsid);

    getMaxDBParams($dbsid);

    # Get the MaxDB db type: either database (=DB) or liveCache (=LVC).
    my $maxdbtype = getMaxDBtype($dbsid);

    # skip if MaxDB database type isn't a liveCache
    return unless ( $maxdbtype eq "LVC" );

    # Get the virtual host for the db instance
    my $sapdbhost = getVdbhostFromXuser( "$sidadm" );

    # Get mountpoints of liveCache instance...
    CreateMountpointAssocArray("$sapdbdir/$dbsid");

    # fill in the %addrs and %subnets hashes
    CreateIPAddressAndSubnetAssocArray();

    # figure out if this is a HSS, check if
    # HotStandbyStorageDLLPath is ?librtehsgesap.so
    my $lchssenable=0;
    $lchssenable=1 if (getMaxDBParVal("HotStandbyStorageDLLPath") =~ "librtehsgesap.so");
    LogVerbose("\$lchssenable: $lchssenable");

    LogReturnSub(($dbsid,$sapdbhost));

    return ($dbsid,$sapdbhost);
}

#############################################################################
#
# For all MDM instances on the system list their SIDs. This includes listing
# of the correspong MDM database instances SIDs
# MDM 7.1 or higher
#
#############################################################################
sub getallmdminstances {
    LogEnterSub(@_);

#    my @mdm_ini_files = </usr/sap/[A-Z][A-Z,0-9][A-Z,0-9]/*[0-9][0-9]/config/*.ini>;
    my @mdm_ini_files = </usr/sap/[A-Z][A-Z,0-9][A-Z,0-9]/*[0-9][0-9]>;
    LogVerbose "\@mdm_ini_files=[@mdm_ini_files]";

    foreach my $mdm_ini_file (@mdm_ini_files) {
        LogVerbose "Working on mdm_ini_file=[$mdm_ini_file]";
        my $mdmfilter = ( mds | mdis | mdss );
        my $MDMFILTER = ( MDS | MDIS | MDSS );
#        if ($mdm_ini_file =~ m#/usr/sap/([A-Z][A-Z,0-9][A-Z,0-9])/(MDS|MDSS|MDIS)[0-9][0-9]/config/(mds|mdis|mdss).ini# ) {
        if ($mdm_ini_file =~ m#/usr/sap/([A-Z][A-Z,0-9][A-Z,0-9])/(MDS|MDSS|MDIS)[0-9][0-9]# ) {
                LogVerbose "Found MDM system [$1] for mdm_ini_file=[$mdm_ini_file]";
                $MDMSIDSmap{"$1"}="$1";
            }
        }

}
#############################################################################
#
# Collect data for an MDM instance that belongs to $SID. Function
# getmdminstance() is the core function for this. Basically getmdminstance()
# is invoked by  running script:  sapgetconf.pl mdminstance Mo7
#
# Produces output in SGMgr (ServiceGuard Manager) format
#
#############################################################################

sub getmdminstance {
    LogEnterSub( @_ );
    $SID = shift;

    #######################################
    # parts of an MDM instance are
    # configured the same as an SAP instance
    #
    #######################################
    # getsapinstance($SID);
    # FormatInstancesForSGMgr();
    # collect SAP instance profiles instead of doing
    # getsapinstance and formatting above
    # @profiles is required to be set by subsequent subroutines
    CollectListOfSIDProfiles( $SID, "/sapmnt/$SID/profile" );

    #######################################
    # set up global variables and hashes
    #######################################
    # Initialize hash for MDM repositories
    %REPOSITORYmap = ();

    #######################################
    # Start collecting the data
    #######################################

    LogVerbose("Collecting MDM repositories");
    GetMDMRepositories($SID);

    #######################################
    # Create SGMgr ouput
    #######################################

    # output the MDM information required by SGMgr
    # FormatMDMForSGMgr();

    # output the MDM alert/service_ info required bu SGMgr
    # FormatMDMAlertForSGMgr();
}

#############################################################################
#
# Get NFS configuration to be offered for NFS toolkit
#
# Arguments: $filter - filter attribute, usually a SID
#
#############################################################################
sub getnfsconfig{

# /etc/dfs/sharetab might not exist (esp. on 11.23). Using /etc/xtab
# instead as fallback.

    LogEnterSub( @_ );

	#######################################
    # set up global variables and hashes
    #######################################

    # Initialize %addrs      maps IP          -> subnet,
    # Initialize %subnets    maps subnet      -> IP
    %addrs   = ();
    %subnets = ();

    # Initialize %mpsfs      maps mountpoints -> filesystems
    # Initialize %mpsfst     maps mountpoints -> filesystem types
    # Initialize %vxdgshared maps ...
    %mpsfs      = ();
    %mpsfst     = ();
    %vxdgshared = ();

    my $filterspec = shift;
    LogVerbose "filterspec=[$filterspec]";
    $filterspec = "/" unless defined $filterspec;

    # go directly into /etc/auto.direct (doesn't try to figure out if this
    # is the correct one)
    -r "/etc/auto.direct" && ScanAutodirect("/etc/auto.direct");

    # Scan HPUX related config files
    # $filterspec="(/sapmnt/$filter|/usr/sap/trans)" if (defined $filter);
    my $export_conf="/etc/dfs/sharetab";
    -r $export_conf  && ScanDfssharetab($export_conf, $filterspec);
    $export_conf="/etc/dfs/dfstab";
    -r $export_conf && ScanDfstab($export_conf, $filterspec);
    $export_conf="/etc/xtab";
    -r $export_conf && ScanXtab($export_conf);

    # Scan Linux related config files
    $export_conf="/proc/fs/nfsd/exports";
    -r $export_conf && ScanExportsLinux($export_conf, $filterspec);

    $export_conf="/var/lib/nfs/etab";
    -r $export_conf && ScanExportsLinux($export_conf, $filterspec);

    if ($^O eq "linux")  {
        ScanExportDirLinux($filterspec);
    } else {
        ScanExportDir($filterspec);
    }
    if ($^O eq "linux")  {
        CheckForFsid();
    }

    LogReturnSub();
}

#############################################################################
#
# scan the /etc/dfs/sharetab (system record of shared filesytems)
# available on HP-UX > 11.23
# Each line of the file consists of the following fields:
# pathname resource fstype specific_options description
#
#############################################################################
sub ScanDfssharetab{
    LogEnterSub( @_ );

    my $file    = shift;
    my $filterspec = shift;

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
        ( $pathname, $resource, $fstype, $options, $description ) =
            split( /\s+/, $_ );
        LogVerbose("pathname=[$pathname] fstype=[$fstype] options=[$options]");
        if ($pathname =~ $filterspec) {
          $XFSPathnameOptionsmap{$pathname}="$options" ;
          # create "path to mountpoint to device entry" arays
          CreateMountpointAssocArray($pathname);
        }
    }
    $fh->close;

    LogReturnSub();
}

#############################################################################
#
# scan the /etc/dfs/dfstab (share commands to be executed as system startup)
# available on HP-UX > 11.23
# Consider only entries with "-F nfs"
# Skip commented entries
# Each line of the file consists of the following fields:
# share -F <fstype> -o <options> [other optons and stuff] <exported dir>
# Entries with other formats will be ignored
#
#############################################################################
sub ScanDfstab{
    LogEnterSub( @_ );

    my $file    = shift;
    my $filterspec = shift;

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
        chomp;
        next if (/^#/); # starts with a comment - skip
        next unless (/^share /); # doesn't start with "share" command (no leading blanks allowed)
        ( $sharecmd, $fsopt, $fstype, $optopt, $options, @remparts ) =
            split( /\s+/, $_ );
        if (($fsopt ne "-F") || ($optopt ne "-o")) {
          LogVerbose "\$fsopt=$fsopt or \$optopt=$optopt unexpected";
          next;
        }
        next if ($fstype ne "nfs");
        $pathname = $remparts[$#remparts]; # take last part as path
        LogVerbose("pathname=[$pathname] fstype=[$fstype] options=[$options]");
        next if defined $XFSPathnameOptionsmap{$pathname}; # already in map, skip it
        if ($pathname =~ $filterspec) {
          $XFSPathnameOptionsmap{$pathname}="$options" ;
          # create "path to mountpoint to device entry" arays
          CreateMountpointAssocArray($pathname);
        }
    }
    $fh->close;

    LogReturnSub();
}


#############################################################################
#
# read in /etc/auto.direct into autodir hash
# to get virtual hostnames (and subsequently IPs) of the NFS servic(s)
# Each line of the file consists of the following fields:
#
# Example HPUX
# <localdir> <remote host>:<exported dir at host>
# /sapmnt/SP8             -proto=udp hpvm-0-45:/export/sapmnt/SP8
#
# Example Linux
#/sapmnt/SX1             -fstype=nfs,nfsvers=3,udp,nosymlink 16.41.100.31:/export/sapmnt/SX1
#
# Introduces AutodirNodemap  (maps exported/remote fs to nodename)
#
#############################################################################
sub ScanAutodirect{
    LogEnterSub( @_ );

    my $file    = shift;

    LogVerbose("parsing file [$file] for NFS server hostnames...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
	    chomp;
        next if (/^\s*#/);  # starts with a comment - skip
        next if (/^$/);     # skip blank lines
        # may contain options
        my @parts = split( /\s+|:/, $_ );
        $localdir = $parts[0]; $remotenode = $parts[$#parts -1]; $remotedir = $parts[$#parts];
        LogVerbose("localdir=[$localdir] remotenode=[$remotenode] remotedir=[$remotedir]");
        $AutodirNodemap{$remotedir} = $remotenode;
    }
    $fh->close;
    LogReturnSub();
}
#############################################################################
#
# scan /etc/xtab - directories to export to NFS clients
# available on HP-UX =< 11.23
# Each line of the file consists of the following fields:
# directory - option[, option]...
#
#############################################################################
sub ScanXtab{
    LogEnterSub( @_ );

    my $file    = shift;
    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
        ( $pathname, $options ) = split( /\s+/, $_ );
        LogVerbose("pathname=[$pathname] options=[$options]");
        $XFSPathnameOptionsmap{$pathname}="$options";

        # create "path to mountpoint to device entry" arays
        CreateMountpointAssocArray($pathname);
    }
    $fh->close;

    LogReturnSub();
}

#############################################################################
#
# Scan /export directory for /usr/sap/trans and /sapmnt/<SID>
#
#############################################################################
sub ScanExportDir{
    LogEnterSub( @_ );

    my $filterspec = shift;
    my $expdir  = "/export";
    my $expusrsaptrans = $expdir."/usr/sap/trans";
    my $expusrsaptransfl = $expdir."/*sap*trans*";  # more than one?
    my $SAPSIDSInExportSapmnt=$expdir.$SAPSIDSInSapmnt;
    my $SAPSIDSInExportSapmntfl =$expdir."/sapmnt*[A-Z][A-Z,0-9][A-Z,0-9]";


    LogVerbose("parsing $expdir directory ...");

    # check if transport is already in
    unless (defined $XFSPathnameOptionsmap{$expusrsaptrans}) {
      if (-d $expusrsaptrans) {
        if ($expusrsaptrans =~ $filterspec) {
            $pathname = $expusrsaptrans;
            $options = "<options to be set>";
            LogVerbose("pathname=[$pathname] options=[$options]");
            $XFSPathnameOptionsmap{$pathname}="$options";
            CreateMountpointAssocArray($pathname);
        }
      }
      else {
        # try flat
        my @transdir = <$expusrsaptransfl>;
        if ((defined $transdir[0]) && (-d  $transdir[0]) && ($transdir[0] =~ filterspec)) {
            $pathname = $transdir[0];
            $options = "<options to be set>";
            LogVerbose("pathname=[$pathname] options=[$options]");
            $XFSPathnameOptionsmap{$pathname}="$options";
            CreateMountpointAssocArray($pathname);
        }
      }
    }

    # loop through whats found via $SAPSIDSInExportSapmnt
    LogVerbose("Scanning for SIDs in \$SAPSIDSInExportSapmnt=[$SAPSIDSInExportSapmnt]");
    my @expdirs = <${SAPSIDSInExportSapmnt}>;
    LogVerbose("\@expdirs=[@expdirs]");
    for my $expdir (@expdirs) {
      next unless $expdir =~ $filterspec;
      next if (defined $XFSPathnameOptionsmap{$expdir});
      if (-d $expdir) {
          $pathname = $expdir;
          $options = "<options to be set>";
          LogVerbose("pathname=[$pathname] options=[$options]");
          $XFSPathnameOptionsmap{$pathname}="$options";
          CreateMountpointAssocArray($pathname);
      }
    }
    # same code for "flat" directories
    LogVerbose("Scanning for SIDs in \$SAPSIDSInExportSapmnt=[$SAPSIDSInExportSapmntfl]");
    my @expdirs = <${SAPSIDSInExportSapmntfl}>;
    LogVerbose("\@expdirs=[@expdirs]");
    for my $expdir (@expdirs) {
      next unless $expdir =~ $filterspec;
      next if (defined $XFSPathnameOptionsmap{$expdir});
      if (-d $expdir) {
          $pathname = $expdir;
          $options = "<options to be set>";
          LogVerbose("pathname=[$pathname] options=[$options]");
          $XFSPathnameOptionsmap{$pathname}="$options";
          CreateMountpointAssocArray($pathname);
      }
    }

    LogReturnSub();
}

#############################################################################
#
# Scan /export directory for /usr/sap/trans and /sapmnt/<SID>
#
# Note: Scan also for "flat" directories in /export e.g. usr_sap_trans
#############################################################################
sub ScanExportDirLinux{
    LogEnterSub( @_ );

    my $filterspec = shift;
    my $expdir  = "/export";
    my $expusrsaptrans = $expdir."/usr/sap/trans";
    my $expusrsaptransfl = $expdir."/usr*sap*trans";  # more than one?
    my $SAPSIDSInExportSapmnt=$expdir.$SAPSIDSInSapmnt;
    my $SAPSIDSInExportSapmntfl =$expdir."/sapmnt*[A-Z][A-Z,0-9][A-Z,0-9]";
    my $options="rw,no_root_squash,sync,nohide";    # default export options

    LogVerbose("parsing $expdir directory ...");

    # check if transport is already in
    if ( scalar keys %{$XFSPathnameOptionsmapLinux{$expusrsaptrans}} == 0 ) {
      if (-d $expusrsaptrans) {
        if ($expusrsaptrans =~ $filterspec) {
            $pathname = $expusrsaptrans;
            LogVerbose("pathname=[$pathname] options=[$options]");
            $XFSPathnameOptionsmapLinux{$pathname}{'*'}=$options;
            CreateMountpointAssocArray($pathname);
        }
      }
      else {
        # try flat
        my @transdir = <${expusrsaptransfl}>;
        if ((defined $transdir[0]) && (-d  $transdir[0]) && ($transdir[0] =~ filterspec)) {
            $pathname = $transdir[0];
            LogVerbose("pathname=[$pathname] options=[$options]");
            $XFSPathnameOptionsmapLinux{$pathname}{'*'}=$options;
            CreateMountpointAssocArray($pathname);
        }
      }
    }

    # loop through whats found via $SAPSIDSInExportSapmnt
    LogVerbose("Scanning for SIDs in \$SAPSIDSInExportSapmnt=[$SAPSIDSInExportSapmnt]");
    my @expdirs = <${SAPSIDSInExportSapmnt}>;
    LogVerbose("\@expdirs=[@expdirs]");
    for my $expdir (@expdirs) {
      next unless $expdir =~ $filterspec;
      next if ( scalar keys %{$XFSPathnameOptionsmapLinux{$expdir}} > 0 );
      if (-d $expdir) {
          $pathname = $expdir;
          LogVerbose("pathname=[$pathname] options=[$options]");
          $XFSPathnameOptionsmapLinux{$pathname}{'*'}=$options;
          CreateMountpointAssocArray($pathname);
      }
    }
    # same code for "flat" directories
    LogVerbose("Scanning for SIDs in \$SAPSIDSInExportSapmnt=[$SAPSIDSInExportSapmntfl]");
    my @expdirs = <${SAPSIDSInExportSapmntfl}>;
    LogVerbose("\@expdirs=[@expdirs]");
    for my $expdir (@expdirs) {
      next unless $expdir =~ $filterspec;
      next if ( scalar keys %{$XFSPathnameOptionsmapLinux{$expdir}} > 0 );
      if (-d $expdir) {
          $pathname = $expdir;
          LogVerbose("pathname=[$pathname] options=[$options]");
          $XFSPathnameOptionsmapLinux{$pathname}{'*'}=$options;
          CreateMountpointAssocArray($pathname);
      }
    }

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%XFSPathnameOptionsmapLinux ], ["XFSPathnameOptionsmapLinux"] ));

    LogReturnSub();
}

#############################################################################
#
#  Scan /var/lib/nfs/etab or /proc/fs/nfsd/exports for nfs export data
#
# Arguments:
#   $file        the proc filesystem: /proc/fs/nfsd/exports
#   $filterspec  filter on specific files (e.g. /sapmnt/GW1...)
#
# a typical line to match:
# (Note: on Linux a filesystem for export can have multiple entries)
# /export/sapmnt/N73        *(rw,no_root_squash,sync,wdelay,no_subtree_check,fsid=119)
# /export/sapmnt/N73 16.123.*(rw,no_root_squash,sync,wdelay,no_subtree_check,fsid=119)
#
# Comments:
# A device major/minor numbers are embedded in the NFS file handle.
# For a NFS package failing over to another node the  major/minor numbers
# will change on the new node, causing "stale NFS file handle" errors.
# To prevent stale file system handle messages after failing over a NFS
# file system, fsid can be configured avoiding embedded major/minor numbers.
# Other options are:
# - configure persistent minor numbers with lvchange  on all cluster nodes
# - use UUIDs as the mount file handle - requires Linux kernel > 2.6.20 though
#
# man exports:
# fsid=num|root|uuid
#    NFS needs to be able to identify each filesystem that it exports.
#    Normally it will use a UUID for the filesystem (if the filesystem has
#    such a thing)...
#
#    As not all filesystems are stored on devices, and not all filesystems
#    have UUIDs, it is sometimes necessary to explicitly tell NFS how to
#    identify a filesystem. This is done with the fsid= option.....
#############################################################################
sub ScanExportsLinux {
    LogEnterSub( @_ );

    my $file    = shift;
    my $filterspec = shift;

    my $pathname;               # a directory being exported
    my $nfsclient;              # host or nfs clients
    my $options;                # the export options for $pathname
	my $fh = FileHandle->new;   # file handle for /proc/fs/nfsd/exports

	open( $fh, $file ) or warn "Failed: open \$file=[$file] failed: \$?=[$?] \$!=[$!]";
    while (<$fh>) {
        next if (/^#/);         # line starts with a comment - skip
	    LogDebug "parsing [$file] line=[$_]";

        # split into pathname, nfsclient and export options
        if (m#^(\S+)\s+(\S+)\((.*)\)#) {
           $pathname=$1;
           $nfsclient=$2;
           $options=$3;

           # check if $pathname $nfsclient combo was
           # already found - if so skip
           next if (defined $XFSPathnameOptionsmapLinux{$pathname}{$nfsclient});

           if ($pathname =~ $filterspec) {
            LogVerbose "parsed $file: pathname=[$pathname] nfsclient[$nfsclient] options=[$options]";
            $XFSPathnameOptionsmapLinux{$pathname}{$nfsclient}=$options;

            # create "path to mountpoint to device entry" arrays
            CreateMountpointAssocArray($pathname);

          }     # filterspec matched
        }       # found pathname/options export combo
    }           # while
    $fh->close;

    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%XFSPathnameOptionsmapLinux ], ["XFSPathnameOptionsmapLinux"] ));
    LogReturnSub();
}

#############################################################################
#
#  Try to determine "fsid/uuid/persistent minor number" for exported file systems
#
# Arguments:
#   $file        the proc filesystem: /proc/fs/nfsd/exports
#   $filterspec  filter on specific files (e.g. /sapmnt/GW1...)
#
# See if we can determine fsid, or uuid or persistent minor numbers for
# each exported filesystem
#############################################################################
sub CheckForFsid {
    LogEnterSub( @_ );

    my $pathname;               # a directory being exported
    my $nfsclient;              # host or nfs clients
    my $options;                # the export options for $pathname
    my $persistent_major=0;     # found: volume has persistent major nr configured
    my $persistent_minor=0;     # found: volume has persistent minor nr configured

    for $pathname (keys %XFSPathnameOptionsmapLinux ) {

        # check if a persistent minor configured for volume
        # this assumes persistent configured on all nodes...
        if ( defined $mpsfs{$pathname} ){
            my $cmd = "/sbin/lvdisplay $mpsfs{$pathname} 2>&1 |";
            open my $fh, "$cmd" or LogRuntimeError "Open for [$cmd] failed: \$!=[$!]";
            while (<$fh>) {
                LogDebug "lvdisplay line=[$_]";
                $persistent_minor=1   if ( $_ =~ m#Persistent minor# );
                $persistent_major=1   if ( $_ =~ m#Persistent major# );
            }
            close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";
            LogVerbose "persistent major/minor found [$persistent_major][$persistent_minor] for [$mpsfs{$pathname}]";
        }

        # Linux can have multiple export lines for same pathname
        for $nfsclient (keys %{$XFSPathnameOptionsmapLinux{$pathname}}) {

            if ( ($persistent_major==0 ) or ($persistent_minor==0) ){
                # Get options for each nfs client
                $options=$XFSPathnameOptionsmapLinux{$pathname}{$nfsclient};
                # Filter out uuid, if options contain both fsid and uuid (file nfsd/exports only)
                # Replace uuid by fsid, if it only contains uuid
                # Use as is if is only contains fsid
                # Add fsid=FSID_REQUIRED_PLEASE_ADD, if none of above applies
                if ($options !~ /(uuid|fsid)/) {
                    LogVerbose "No fsid/uuid for [$pathname] - FSID_REQUIRED_PLEASE_ADD";
                    $options="fsid=FSID_REQUIRED_PLEASE_ADD," . "$options";
                    $XFSPathnameOptionsmapLinux{$pathname}{$nfsclient}=$options;
                }
                elsif (($options =~ /uuid/) && ($options =~ /fsid/)) {
                    # both. get rid of uuid part (lower case only in uuid of nfsd/exports)
                    $options =~ s/,uuid=[a-f0-9:]+//;
                    $XFSPathnameOptionsmapLinux{$pathname}{$nfsclient}=$options;
                }
                elsif ($options =~ /uuid/) {
                    # uuid only: uuid->fsid
                    $options =~ s/,uuid=/,fsid=/;
                    $XFSPathnameOptionsmapLinux{$pathname}{$nfsclient}=$options;
                }
                # else: fsid only, do nothing
            }

        }
    }
    $dumpdata=1;
    $dumpdata && LogVerbose ( Data::Dumper->Dump( [ \%XFSPathnameOptionsmapLinux ], ["XFSPathnameOptionsmapLinux"] ));
    LogReturnSub();
}
#############################################################################
#
# Figure out SG and SGesAP directory
#
# First check if environment variable SGCONF is defined.
# Assumes that that one has the form <some-prefix>/cmcluster/conf
#
# If that isn't defined it checks for /etc/cmcluster and /usr/local/cmcluster
#
# Sets global vars SGCONFDIR, SGeSAPBINDIR and SGeSAPSVCDIR
# (SGeSAPBINDIR is usually the directory where this script resides)#
#
#############################################################################
sub getSGdirs
{
  LogEnterSub(@_);

  # by using SAP_functions.pm environment variables from cmcluster.conf should be defined
  # at this point (getcmclusterconf)

  $SGCONFDIR=$ENV{"SGCONF"};

  # search for confdir if not found in environment (this code should never get executed
  # in a cluster environment)
  unless (defined($SGCONFDIR)) {
    # Create a searchlist for conf directory (SLES,RHEL,HPUX)
    # Expecting a subdir "modules" there (just in case somebody create a subdir "conf").
    my @searchlist = qw(
                         /opt/cmcluster/conf
                         /usr/local/cmcluster/conf
                         /etc/cmcluster
                        );
    foreach my $confdir (@searchlist) {
      if (-d $confdir."/modules") {
        $SGCONFDIR = $confdir ;
        last;
      }
    }
  }

  # Search for directory containing SGeSAP executables:
  # Seperate /opt/cmcluster/sap/bin on HP-UX, in $SGSBIN on Linux
  # Search for service/monitor dir:
  # On HP-UX /opt/cmcluster/sap/SID. On Linux $SGCONF/monitors/sgesap
  # There's no ENV variable for this (maybe it should be introduced
  # to allow overwrite)
  if ($^O eq "linux")  {
    $SGeSAPBINDIR=$ENV{SGSBIN};
	$SGeSAPSVCDIR="\$SGCONF/monitors/sgesap";
  }
  else {
    # HP-UX
    my $sgesapdir="/opt/cmcluster/sap";
	$SGeSAPBINDIR=$sgesapdir."/bin";
    $SGeSAPSVCDIR=$sgesapdir."/SID";
  }
  LogReturnSub("\$SGCONFDIR=[$SGCONFDIR],\$SGeSAPBINDIR=[$SGeSAPBINDIR],\$SGeSAPSVCDIR=[$SGeSAPSVCDIR]");
}


#############################################################################
#
# This function gets SAP Database Host (= the virtual hostname under which
# the database will run)
#
#   1) first check if a variable SAPDBHOST is defined in DEFAULT.PFL - this is
#      valid  for MaxDB and ORACLE. If variable SAPDBHOST was not found:
#   2) for ADA (=MaxDB) run the "xuser list" command
#   3) for ORA (=Oracle) run the "tnsping $SID" command
#
# Arguments:
#   $DEFAULT.PFL    The default SAP profile
#   $dbtype:        Either ADA,ORA,DB2 or SYB
#   $SID:           The SAP System ID
#
# Returns
#   $sapdbhost (Note: can be empty string if not defined)
#
#############################################################################
sub GetSAPDbhost {
    LogEnterSub( @_ );

    my ( $DEFAULTPFL , $dbtype, $SID ) = @_;

    my $sapdbhost = "";

    # Need to check for NFS access here. DEFAULTPFL usually is on /sapmnt
    # dirname() is imported File::Basename
    # check for a arbitrary directory. if a timeout occurs checkNFSdiracess
    # will return 130

    if (( checkNFSdiraccess(dirname($DEFAULTPFL)."/$$") != 130 ) && ( -f $DEFAULTPFL )) {
        $sapdbhost = getVdbhostFromDefaultProfile( $DEFAULTPFL );
    }

    if ( ( $sapdbhost eq "" ) && ( $dbtype eq "ADA" ) ) {
        $sapdbhost = getVdbhostFromXuser( "$sidadm" );
    }

    if ( ( $sapdbhost eq "" ) && ( $dbtype eq "ORA" ) ) {
        $sapdbhost = getVdbhostFromTnsping( $oraadm, $SID );
    }

    if ( ( $sapdbhost eq "" ) && ( $dbtype eq "HDB" ) ) {
        $sapdbhost = getHDBhostsFromHDBprofilenames( dirname($DEFAULTPFL), $SID );
    }

    # no fallback for DB2 and SYB
    # SYB: could get from <sid>adm env or interface file

    # map IP address into a hostname
    if ( $sapdbhost =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/ ) {
        LogVerbose("converting IP address to name \$sapdbhost=[$sapdbhost]");
        # No need to convert as ultimately the IP address will be reported back
        # and the DB module doesn't support a virtual host name anyway.
    }

    LogReturnSub("\$sapdbhost=[$sapdbhost]");
    return ("$sapdbhost");
}

#############################################################################
#
# Figure out which MaxDB user key to use
# Usually this is "c", but for pure JDB it could be "c_J2EE".
# "c_J2EE" will be set w/o further checks if "c" isn't available
# Implicitly uses and if necessary sets $maxdbuc.
# Checks for string "No xuser entry for key" and assigns other key
# if found.
# Alternative: check for /XUSER Entry  \d+/ and assign "c_J2EE" if not found
#
# Arguments:
#   $SID:           The SAP System ID
# Returns
#    $maxdbuc
# Implicit inputs
#   $sidadm
#   $maxdbuc (output also)
#
#############################################################################

sub setMaxDBUC {
    LogEnterSub( @_ );

    my ($SID) = @_;

    my $cmd=qq[su - $sidadm -c "xuser -U $maxdbuc list " 2>/dev/null |];

    LogVerbose("Execute \$cmd=[$cmd]");

    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
        if ( $_ =~ "No xuser entry for key" ) {
         # LogVerbose ("MaxDB userkey [$maxdbuc] not configured");
         $maxdbuc="c_J2EE";
         # LogVerbose ("Setting it to [$maxdbuc]");
         last;
      }
    }
    close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";

    LogReturnSub("\$maxdbuc=[$maxdbuc]");
    return $maxdbuc;
}

#############################################################################
# This function determines the type of ADA/MaxDB database: It can either be a
# Database (=OLTP) or liveCache (=LVC)
#
# Note: This requires the x_server process to be running, else can't
# get the required information
#
# Arguments:
#   $SID:           The SAP System ID
# Returns
#   $maxdbtype      Either OLTP or LVC
# Implicit inputs
#   $sidadm
#############################################################################
sub getMaxDBtype {
    LogEnterSub( @_ );

    my ($SID) = @_;
    my $maxdbtype = "";

    #NOTE: database (or at least the virtual host used) has to be up for the
    #dbmcli command to succeed. If not, the command fails - and sometimes takes
    #quite a while before failing.
    #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
    my $cmd =
        qq[su - $sidadm -c "dbmcli -U $maxdbuc -d $SID -n $ENV{HOSTNAME} param_directget INSTANCE_TYPE" 2>/dev/null |];

    LogVerbose("Execute \$cmd=[$cmd]");

    my $fh = FileHandle->new;
    open( $fh, $cmd )
            or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while (<$fh>) {
        # String to match:      INSTANCE_TYPE   LVC
        if (/^\s*(INSTANCE_TYPE)\s+(\S+)\s*(.*)$/) { $maxdbtype = $2; last; }
    }
    close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n"
                       . "MaxDB x_server not running?\n"
                       ;

    # second chance if dbmcli failed. Checking for
    # /sapdb/$DBSID/db/sap/lcapps.pkg (lcinit isn't an indicator)
    # (or link /sapdb/$DBSID/db/sap/lccluster)
    # Note: this also requires the filesystem to be mounted.
    if (( $maxdbtype eq "" )
        && (   ( -e "/sapdb/$SID/db/sap/lcapps.pkg" )
            || ( -l "/sapdb/$SID/db/sap/lccluster" ) )
        )
    {
        LogVerbose("LVC indicator files [lcapps.pkg|lccluster] found in filesystem");
        $maxdbtype = "LVC";
    }
    LogReturnSub("\$maxdbtype=[$maxdbtype]");
    return ("$maxdbtype");
}

#############################################################################
#
# Get the virtual DB hostname form the SAP Default profile
# /sapmnt/SID/profile/DEFAULT.PFL
#
# Arguments:
#   $DEFAULTPFL       Name of the Default profile
#
#############################################################################
sub getVdbhostFromDefaultProfile {
    LogEnterSub( @_ );

    my $DEFAULTPFL = shift;
    my $sapdbhost="";

    LogVerbose("Parsing profile=[$DEFAULTPFL] ...");

    my $fh = FileHandle->new($DEFAULTPFL);
    if ( !defined($fh) ) {
        warn "Warning opening $DEFAULTPFL for reading:\$?=[$?] \$!=[$!]";
    }
    else {
        while (<$fh>) {
            # String to match     "SAPDBHOST = virtIP"
            if (/^\s*SAPDBHOST\s*=\s*(\S+).*$/) {
                $sapdbhost = $1;
                last;
            }
        }
        $fh->close;
    }
    LogReturnSub("\$sapdbhost=[$sapdbhost]");
    return $sapdbhost;
}

#############################################################################
#
# Get the virtual DB hostname for a MaxDB database. Run the xuser command
# to get this data
#
# NOTE: the following command implies the default "c" admin user is defined
#
# Arguments:
#   $sidadm        Name of MaxDB/liveCache administrator
#
#############################################################################
sub getVdbhostFromXuser {
    LogEnterSub( @_ );

    my $sidadm = shift;

    #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
    my $cmd = qq[su - $sidadm -c "xuser -U $maxdbuc list " 2>/dev/null |];
    LogVerbose("Executing \$cmd=[$cmd]");

    my $sapdbhost = "";

    my $fh = FileHandle->new;
    open( $fh, $cmd )
        or die "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    while (<$fh>) {
        # String to match       "Nodename    :$sapdbhost"
        if (/^\s*(Nodename)\s*:\s*(\S+)\s*(.*)$/) {
            $sapdbhost = $2;
        }
    }
    close($fh) or die "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";

    LogReturnSub("\$sapdbhost=[$sapdbhost]");
    return $sapdbhost;
}

#############################################################################
#
# Get the virtual DB hostname for an Oracle database. Run the Oracle tnsping
# command to get this data.
#
# Arguments:
#   $oraadm        Name of MaxDB/liveCache administrator
#   $SID           SAP system ID
#
#############################################################################
sub getVdbhostFromTnsping {
    LogEnterSub( @_ );

    my $oraadm = shift;
    my $SID    = shift;

    my $sapdbhost;

    #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
    my $cmd = qq[su - $oraadm -c "tnsping $SID" 2>/dev/null |];
    LogVerbose("Executing \$cmd=[$cmd]");

    my $fh = FileHandle->new;

    open( $fh, $cmd ) or warn "Failed: popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while (<$fh>) {
        # String to match:      (HOST = virtdb)
        if (/\(\s*HOST\s*=\s*(\S+)\s*\).*$/) {
            $sapdbhost = $1;
        }
    }

    # Don't log a warning if the close($fh) fails - probably the listener
    # isn't running or tnsping isn't in patch (FS not mounted).
    # We really are only interested in getting the Host of
    # Oracle connect string anyway

    close($fh) or warn "Failed: pclose for \$cmd=[$cmd] \$?=[$?] \$!=[$!]\n";

    LogReturnSub("\$sapdbhost=[$sapdbhost]");
    return $sapdbhost;
}

#############################################################################
#
# Get the virtual DB hostname for a Sybase ASE database from interfaces file.
# Looks for master entry of the servername.
# Server usually is the <SID>, but could be also <SID>_BS or replication server
# Returns an array containing the hostname and the port.
#
# Arguments:
#   $server        Name of Sybase Server
#   $ifname        Name interfaces file (full path)
#
#############################################################################
sub getVdbhostFromInterfaces {
    LogEnterSub( @_ );

    my $syserver = shift;
    my $ifname   = shift;

	my $fh = FileHandle->new;

	my $sapdbhost;
    my $sapdbport;

	open( $fh, $ifname ) or warn "Failed: open \$ifname=[$ifname] failed: \$?=[$?] \$!=[$!]";
OLS: while (<$fh>) {
	    #print $_;
        if (/^(${syserver})$/) {
		   # loop through service_type entries which have to start with a TAB
		   # and be either master or query. Get nodename from (first?) master entry.
		   # (TODO: what about using query entry instead - or either?)
		   while (<$fh>) {
		     #print $_;
		     # <TAB>|<Space>master tcp ether <host> <port>
			 last OLS unless (/^\s+(master|query)/); # already beyond service entries
			 if ($1 eq "master") {
               ($dc, $dc, $dc, $dc, $sapdbhost, $sapdbport) = split(/\s+/, $_);
		       last OLS;
			 }
		  }
        }
    }
	close($fh);

    LogReturnSub("\$sapdbhost=[$sapdbhost],\$sapdbport=[$sapdbport]");
	return ($sapdbhost,$sapdbport);
}

#############################################################################
#
# Get HANA hostname(s) from HANA profiles.
#
# HANA profile names are build according Netweaver rules:
# <SID>_HDB<INR>_<node>
#
# Arguments:
#   $profiledir    HANA profile directory
#   $SID           HANA SID
#############################################################################
sub getHDBhostsFromHDBprofilenames {
	LogEnterSub( @_ );
	my ($profiledir,$SID) = @_;

	my $hananodes="";

	# get all the HDB profiles and split of hostnames
	CollectListOfSIDProfiles ($SID,$profiledir,"HDB");
	# following code taken from FormInstancesforSGMgr (and adapted)
	# loop thru all available HDB profiles
    foreach $prof (@profiles) {
      # Get the profile filename from the absolute path
      my $profname;
      $profname = ( split( /\//, $prof ) )[5];

      # get instance name and (virtual) hostname out of the profile file name.
      # A SAP profile has the following format <SID>_<INST>_<HOST>
      my ( $sid, $inst, $hhost ) = split( /_/, $profname );
      LogVerbose ("\$hhost=$hhost");
      if ($hananodes eq "") {
      	$hananodes = $hhost;
      }
      else {
      	$hananodes = "$hananodes,$hhost";
      }
    }

    LogReturnSub("\$hananodes=[$hananodes]");
    return("$hananodes");
}

#############################################################################
#
# Get the database type:
# Oracle , MaxDB/ADA, DB2, Sybase are supported
# Informix and other are not supported
#
#   $dbsid  the SID of the database
#
#############################################################################
sub getDbtype{
    LogEnterSub( @_ );

    my $dbsid = shift;
    my $defpfl = shift;

    my $dbtype = getDbtypeFromMountpoint( $dbsid );

    if (!$dbtype) {
      my ($dbsid_from_defpfl,$dbtype_from_defpfl) = getDBFromProfile($dbsid, $defpfl);
      # overwrite dbtype/dbsid - was set in the admin environment
      $dbtype = uc($dbtype_from_defpfl) if ( $dbtype_from_defpfl );
      $dbsid =  $dbsid_from_defpfl if ( $dbsid_from_defpfl );
    }

    if (!$dbtype) {
        # try to get db type from the environment variables of sidadm
        my ($dbsid_from_env,$dbtype_from_env) = getSidadmDbEnv( $sidadm );

        # overwrite dbtype/dbsid - was set in the admin environment
        $dbtype = $dbtype_from_env if ( $dbtype_from_env );
        $dbsid = $dbsid_from_env if ( $dbsid_from_env );
    }

    # checking environment might have returned "db6" (==DB2 LUW for SAP)
    # or "syb"
    # support depends on apiversion
    if ( ($dbtype eq "db6") || ($dbtype eq "DB2") ) {
        if ( $apiversion & 0x01 )  {
          $dbtype="DB2";
        }
        else {
          warn "DB2 unsupported";
          $dbsid="";$dbtype="";
        }
    }
    if ( ($dbtype eq "syb") || ($dbtype eq "SYB") ) {
        if ( $apiversion & 0x10 )  {
          $dbtype="SYB";
        }
        else {
          warn "Sybase unsupported";
          $dbsid="";$dbtype="";
        }
    }

    if ( !$dbtype ) {
         warn "No (supported) dbtype found";
         $dbsid="";
    }
    LogReturnSub("\$dbsid=[$dbsid] \$dbtype=[$dbtype]");
    return($dbsid,$dbtype);
}


#############################################################################
#
# This function writes in all MaxDB/LC database parameters into the
# associative (global) array "maxdbpars". Also figures out if new or old
# style naming is used an sets global varialbe "maxdboldparnames".
#
# Arguments:
#   $SID:           The datbase ID
#
#############################################################################
sub getMaxDBParams {
  LogEnterSub( @_ );

  my $dbSID = shift;
  my $dbadm = lc($dbSID)."adm";
  my $printon=0;

  #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
  #First echo is used to get "OK" on extra line
#  my $cmd=qq[su - $dbadm -c "echo - ; dbmcli -U $maxdbuc -d $dbSID -n localhost param_directgetall" 2>/dev/null |];
# Using hostname instead of localhost increases the chance that network won't be involved and call succeeds when just
# DB directory is mounted (Linux only?)
  my $cmd=qq[su - $dbadm -c "echo - ; dbmcli -U $maxdbuc -d $dbSID -n $ENV{HOSTNAME} param_directgetall" 2>/dev/null |];

  # fill an assoc. array with the parameters
  # starts doing this after it sees the "OK" line (to skip leading output caused by su)
  LogVerbose("getMaxDBParams() executing \$cmd=[$cmd]");
  open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  while (<$fh>) {
    if ($printon) {
      ($parnam, $parval) = split(/\s+/,$_,2);
      chop ($parval); # remove trailing \n
      $maxdbpars{$parnam}=$parval;
    }
    $printon=1 if ($_ =~ "^OK\$");
  }
  close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  # LogVerbose("\@maxdbpars: @maxdbpars");

  if ( $maxdbpars{"KernelVersion"} ne "" ) {
    LogVerbose("MaxDB/LC new style parameter names");
    $maxdboldparnames=0;
  }
  elsif ( $maxdbpars{"KERNELVERSION"} ne "" ) {
    LogVerbose("MaxDB/LC old sytle parameter names") ;
    $maxdboldparnames=1;
  }
  else {
    LogError("Couldn't determine MaxDB/LC parameter naming");
    LogError("Neither KERNELVERSION nor KernelVersion found");
  }
  LogReturnSub();
}

#############################################################################
#
# Get MAXDB database parameter value out of map
#
# Args:
# - parameter name (new style)
# - parameter "index" (optional, e.g. "0005")
#
# Return:
#   parameter value (or empty string)
#
#############################################################################

sub getMaxDBParVal {
    LogEnterSub( @_ );

    my $parname = shift;
    my $paridx  = shift;
    $parname = $maxdbnewoldmap{$parname} if ($maxdbolparnames);
    $parname = $parname.$paridx;

    LogReturnSub("\$maxdbpars{$parname}=$maxdbpars{$parname}");
    return $maxdbpars{$parname};
}


#############################################################################
#
# Get the database type from the default mount points of the databases
#
# Arguments:
#   $SID  the SID of the database
#
# Alterntative ways to start discovery may be (at least to figure out
# if "root" directories are correct):
#  - /etc/oratab for Oracle
#  - /etc/opt/sdb for MaxDB
#  - ?? for DB2
#
#
#############################################################################
sub getDbtypeFromMountpoint {
    LogEnterSub( @_ );

    my $DBSID = shift;

    my $dbtype="";

    if ( -d "/oracle/$DBSID" ) {
        $dbtype = "ORA";
    }
    elsif ( -d "/sapdb/$DBSID" ) {
        $dbtype = "ADA";
    }
    elsif (-d "/db2/$DBSID") {
        $dbtype = "DB2";
    }
    elsif (-d "/sybase/$DBSID") {
        $dbtype = "SYB";
    }
      # Try HANA DB, but only if environment variable is set
      # Check on /hana/shared/$DBSID
      # For deploysappkgs usage only (no "Posix" reporting)
    elsif (($ENV{"SAPGETCONF_HDBASDBINSTANCE"} || 0) && (-d "/hana/shared/$DBSID")){
         $dbtype = "HDB";
    }

    LogReturnSub("\$dbtype=[$dbtype]");
    return $dbtype;
}

#############################################################################
#
# Get the database type and the database SID from the user environemnt
#
# NOTE: This function queries the SAP sidadm and NOT the database admin
#
# NOTE: This only works databases installed by SAP installation tools -
# variable DB_SID and dbms_type get set by these
#
#  WVE: Looks as DB_SID isn't available anymore. Use dbs/hdb/dbname (for HANA)
#         and j2ee/dbname (others, fallback). The latter seems also be there
#         for non-JAVA DBs.
#
# Arguments:
#   $sidadm    the database administrator
#
#############################################################################
sub getSidadmDbEnv {
    LogEnterSub( @_ );

    my $sidadm = shift;

    #NOTE: 2>/dev/null required or else HPUX interactively queries TERM type
    my $cmd = "su - $sidadm -c env 2>/dev/null";
    LogVerbose("Execute cmd=[$cmd]");

    my @sidadmenv = `$cmd` or warn "Couldn't get $sidadm environment";
    LogDebug("\@sidadmenv=[", @sidadmenv, "]");

    my $dbsid = ( split( /=/, ( grep( /DB_SID=/, @sidadmenv ) )[0] ) )[1];

    # can be different, same or empty (compared to SID of sidadm)
    chop($dbsid);

    my $dbtype = ( split( /=/, ( grep( /dbms_type=/, @sidadmenv ) )[0] ) )[1];
    chop($dbtype);

    LogReturnSub("\$dbsid=[$dbsid] \$dbtype=[$dbtype]");
    return($dbsid,$dbtype);
}

sub getDBFromProfile {
  LogEnterSub( @_ );
  my $dbsid = shift;
  my $defpfl = shift;

  open(PFL,$defpfl);
  my @lines = <PFL>;
  close(PFL);

  my $dbtype = ( split( / = /, ( grep( m#j2ee/dbtype#, @lines ) )[0] ) )[1];
  chop($dbtype);

  # if ($dbtype = "hdb") -> use dbs/hdb/dbname ?

  $dbsid = ( split( / = /, ( grep( m#j2ee/dbname#, @lines ) )[0] ) )[1];
  chop($dbsid);

  LogReturnSub("\$dbsid=[$dbsid] \$dbtype=[$dbtype]");
  return($dbsid,$dbtype);
}


#############################################################################
# This functions creates two associative arrays from the output of the
# netstat command:
#
# Does nothing if network support utility is found, otherwise falls back
# to netstat and fills addrs and subnets maps. Fallback IPv4 only.
#
#   addrs{}     maps IP->subnet,
#   subnets{}   maps subnet->IP
#############################################################################
sub CreateIPAddressAndSubnetAssocArray {
    LogEnterSub( @_ );

    if (-x $nwrksupp)
    {
      LogVerbose ("$nwrksupp available - arrays won't be created");
      LogReturnSub();
      return;
    }

    # different code paths for HP-UX and Linux
	if ($^O ne "linux")  {
      # HP-UX: collect 'netstat -in' first, will be needed in loop
      my $cmd = "netstat -in -f inet";
      LogVerbose("Execute cmd=[$cmd]");

      my @interfaces = `$cmd` or warn "cmd=[$cmd], return code [$!]";
      LogDebug("\@interfaces=[", @interfaces, "]");

      # pop off the header string of the netstat command
      shift @interfaces;

      # create two associative arrays: IP->subnet, subnet->IP
      # (IP for the latter not relevant, just existence of subnet)
      # $if and $dc just a placeholder
      my $if, $dc, $subnet, $addr;
      for $_ (@interfaces) {
        ( $if, $dc, $subnet, $addr ) = split( / +/, $_ );
        if ( $addr =~ /^[[:digit:]]/ ) {
            $addrs{$addr}     = $subnet;
            $subnets{$subnet} = 1;
        }
      }
    }
    else  {
      # Linux: collect with "ip show" (v4 only)
      my $cmd=qq[ip -f inet addr show 2>/dev/null |];
      LogVerbose("Execute cmd=[$cmd]");
      open(FH,$cmd) or warn "popen [$cmd] failed";
      while (<FH>) {
        #  print $_;
        $currif = $1 if (/^\d+: (\w+): /);  # next?
        if (m[inet (\d+\.\d+\.\d+\.\d+/\d+) brd]) {
            #print "$currif: $1 \n";
            ($addr,$subnet) = getsubnetfromprefix($1);
            #print "$currif: $addr/$subnet \n";
            $addrs{$addr} = $subnet;
            $subnets{$subnet} = 1;
        }
      }
      close (FH) or warn "pclose [$cmd] failed, return code [$!]";
    }

    # Log the hash data found
    my @printarr;
    @printarr = %addrs;
    LogVerbose("\%addrs=[@printarr]");
    @printarr = %subnets;
    LogVerbose("\%subnets=[@printarr]");

    LogReturnSub();
}

#############################################################################
#
# Helper function translate a IP/Subnet notation of
#  address/prefix-bits into
#  address/subnet
#
# So far only used on Linux to translate notation out of "ip show"
#
# Arguments
#   $addrpref - address in prefix in addr/prefix notation (string)
#
# Returns
#   ($addr,$subnet) - array of address and subnet
#
#############################################################################
sub getsubnetfromprefix {

  my $ipnpref=shift ;
  #print $ipnpref,"\n";

  my ($ipaddr,$prefbits)=split('/',$ipnpref);
  #print $ipaddr," ",$prefbits,"\n";

  my $maskbits = (0xffffffff<<(32-$prefbits)) &  0xffffffff;
  #printf ("maskbits: %x\n", $maskbits);

  my @addrparts = split('\.',$ipaddr);
  #print "@addrparts \n";
  my $addrbits = $addrparts[0]*0x1000000 +  $addrparts[1]*0x10000 +  $addrparts[2]*0x100 +  $addrparts[3];
  #printf ("addrbits: %x\n", $addrbits);

  my $subnbits= $addrbits & $maskbits;
  #printf ("subnbits: %x\n", $subnbits);

  my @subnparts = (0,0,0,0);
  $subnparts[0]=($subnbits&0xff000000)>>24;
  $subnparts[1]=($subnbits&0x00ff0000)>>16;
  $subnparts[2]=($subnbits&0x0000ff00)>>8;
  $subnparts[3]=$subnbits&0x000000ff;
  #print "@subnparts \n";
  return ($ipaddr,"$subnparts[0].$subnparts[1].$subnparts[2].$subnparts[3]");
}

#############################################################################
# This functions creates two associative arrays from the output of the
# mount command:
#   mpsfs{}     maps mountpoints -> filesystems
#   mpsfst{}    maps mountpoints -> filesystem types
#
#   Example:
#   %mpsfs=[/oracle/MO7 /dev/vgoramo7/lvoramo7]
#   %mpsfst=[/oracle/MO7 vxfs]
#
# Filters mountpoints using the expression provided as function argument.
# get the mount information to match it later to the instance
# directories.
# Uses current active mounts as first datasource (mount -v) and
# then will dig through /etc/fstab
#
# Linux: Output of "mount -v" looks the same. /etc/fstab has the same format.
#        vxvm might be different, but ignored when /etc/vx/volboot isn't there.
#
# Arguments:
#   $mpfilter   mount point filter -> the mount we are trying to get the
#               filesystem and the filesystem type for
#############################################################################
sub CreateMountpointAssocArray {
    LogEnterSub( @_ );

    my $mpfilter = shift;

    $vxvm_installed = 0;
    my $vxvm_used = 0;

    if ( -f </etc/vx/volboot> ) {
        # VXVM installed (same test used in SGMgrPI)
        $vxvm_installed = 1;
        LogVerbose("vxvm installed");
    }


    my $cmd = "mount -v";  # no diff on linux w/ or w/o -v
    LogVerbose("Execute cmd=[$cmd]");

    @mounts = `$cmd` or warn "cmd=[$cmd], return code [$!]";
    LogDebug("\@mounts=[@mounts]");

    #------------------------------------------------------------------
    # Create hashes: Split the mount data. Save only if a "mount point"
    # $mp matches the "mount point filter" $mpfilter
    #   $dc   just a placeholder
    #   $fs   filesystem
    #   $mp   mountpoint
    #   $fst  filesystem type
    #------------------------------------------------------------------

    my $fs, $dc, $mp, $fst;
    for $_ (@mounts) {
        # format: <fs> on <mp> type <fst> ...
        ( $fs, $dc, $mp, $dc, $fst ) = split( / +/, $_ );
        if ( $mp =~ /$mpfilter/ ) {
            $mpsfs{$mp}  = $fs;
            $mpsfst{$mp} = $fst;
            if ( $vxvm_installed && $fs =~ /^\/dev\/vx/ ) {
                $vxvm_used = 1;
                LogVerbose("vmvx: $fs");
            }
        }
    }

    # dig through /etc/fstab (although usually not there anymore
    # if configured to a cluster package, but might have noauto)

    my $file    = "/etc/fstab";

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
        chomp;
        next if (/^#/); # starts with a comment - skip
		next if (/^$/); # skip empty line
        ( $fs, $mp, $fst ) = split( /\s+/, $_ );
        next if (defined $mpsfs{$mp});
        if ( $mp =~ /$mpfilter/ ) {
            $mpsfs{$mp}  = $fs;
            $mpsfst{$mp} = $fst;
            if ( $vxvm_installed && $fs =~ /^\/dev\/vx/ ) {
                $vxvm_used = 1;
                LogVerbose("vmvx: $fs");
            }
        }
    }
    $fh->close;

    # Print the hash data found
    my @printarr;
    @printarr = %mpsfs;
    LogVerbose("\%mpsfs=[@printarr]");
    @printarr = %mpsfst;
    LogVerbose("\%mpsfst=[@printarr]");


    # special check for shared vxvm/cvm volumes
    if ($vxvm_used) {
        # executed a vxdg list to figure out if dg is "shared"
        my $cmd = "vxdg list";
        LogVerbose("Execute cmd=[$cmd]");
        @vxdglist = `$cmd` or warn "cmd=[$cmd], return code [$!]";
        LogVerbose("\@vxdglist=[@vxdglist]");

        shift @vxdglist;

        for $_ (@vxdglist) {
            @vxdg = split(/ +/);
            LogVerbose("NAME: $vxdg[0] - STATE: $vxdg[1]");
            $vxdgshared{ $vxdg[0] } = ( $vxdg[1] =~ /shared/ );
            @printarr = %vxdgshared;
            LogVerbose("\%vxdgshared=[@printarr]");
        }
    }
    LogReturnSub();
}

#############################################################################
# Collect all SAP profiles found in the $profiledir="/sapmnt/$SID/profile" .
# A SAP profile has the following format <SID>_<INSTNAME><INSTNR>_<HOST>
#############################################################################

sub CollectListOfSIDProfiles {
    LogEnterSub( @_ );

    my $SID        = shift;
    my $profiledir = shift;
    my $filtertype = shift; # maybe undefined, key for InstanceFilters (already checked if allowed)

    LogVerbose("Collecting profiles in profiledir=[$profiledir]");

    # Do (maybe another) NFS check
    # Trick: Build up some (probably) non-exisiting filename to get around NFS caching issues when the cache
    #   still has information on $profiledir, but NFS isn't available anymore. In such a case, a check on $profiledir
    #   will succeed, but the subsequent opendir will hang.
    # (assumes that a check to $profiledir has been done before to ensure that it exists)
    return if checkNFSdiraccess($profiledir."/$$") == 130; # doesn't exist, but may timeout

    # scan for SAP profiles found in a SAP/SID  profile directory
    opendir( DIR_HANDLE, $profiledir )
        or warn "Opening directory profiledir [$profiledir] failed\n";
    my @filelist = readdir(DIR_HANDLE);
    closedir(DIR_HANDLE);

    foreach my $file (@filelist) {
        # skip any profiles containing a "." in the filename
        # this eliminates hidden files as well as profiles
        # containing a version number

        my $filtered_out=0;

        if ($file =~ /[.]/) {
            LogDebug("skip profile_file=[$file]");
            next;
        }

        # if a profile name was found that belongs to the SID, save it
        # in array @profiles. NOTE: saving it with absolute pathname because
        # required later. Make sure a profile name contains exacty 2 underscores
        # (underscore are not allowed in hostnames!)
        if ($file =~ /^${SID}_/ &&  (($file =~ tr/_//) == 2)) {
            $filtered_out=0;
            if (defined $filtertype)
            {
              #LogVerbose ("Using filtertype $filtertype");
              $filtered_out=1;
              if ($file =~ /^${SID}_${InstanceFilters{$filtertype}}[0-9][0-9]_/) {
                LogVerbose ("filtered in ($filtertype): $file");
                $filtered_out=0;
                #  last;
              }
            }
            next if $filtered_out;
            LogVerbose("push onto array \@profiles profile_file=[$profiledir/$file]");
            push( @profiles, "$profiledir/$file" );

            # if this profile also is a ERS profile type , save it in array @ersprofiles
            # NOTE: saving it with absolute pathname because required later
            if ($file =~ /^${SID}_ERS[0-9][0-9]_/) {
                LogVerbose("push onto array \@ersprofiles profile_file=[$profiledir/$file]");
                push( @ersprofiles, "$profiledir/$file" );
            }
        }
    }
}

#############################################################################
# "precollect" ERS<->[A]SCS relationship and put them in maps
# needed later for sap_replicated_instance for ERS as well as buildung
# resourcename for [A]SCS.
#############################################################################
sub CreateERSandSCSAssocArray {
    LogEnterSub( @_ );

    # Loop thru the ers profile array. For each ERS instance get the corresponding
    # ASCS/SCS instance number (=SCSID) and host(=SCSHOST)
    foreach $prof (@ersprofiles) {
      my $scshost = "";
      my $scsinr  = "";
      my $ersinst, $ersvhost;

      # seperate profile name into components
      ( $ersinst, $ersvhost ) = ( split( /_/, $prof ) )[ 1, 2 ];

      # Get the values for SCSID and SCSHOST from the ERS profile
      ( $scsinr, $scshost ) =
            GetSCSIDAndSCSHOST( "$prof", $scsinr, $scshost );

      LogDebug("scsinr=[",  $scsinr,  "]");
      LogDebug("scshost=[", $scshost, "]");

      my @match_scs_profile = grep ( /$scsinr\_$scshost$/, @profiles );
        # corresponding SCS profile may not be found if filtering is switched on
      LogDebug("match_scs_profile=[$match_scs_profile[0]]");
      $scsinst = ( split( /_/, $match_scs_profile[0] ) )[1];
      LogDebug("scsinst=[", $scsinst, "]");
      LogDebug("ersinst=[", $ersinst, "]");

      $ERSmap{$ersinst} = $scsinst;
      $SCSmap{$scsinst} = $ersinst;
    }
}

#############################################################################
# This function scans a SAP profile file (= MO7_ERS33_virt125 or DEFAULT.PFL)
# and returns the values for SCSID and SCSHOST (only tries to return a value
# if not set). Therfore variables found in MO7_ERS33_virt125 override any
# variables found in DEFAULT.PFL
#
# Arguments:
#   $file   - SAP profile file to read
#   $scsinr - SAP instance number of SCS instance
#   $scshost- virtual hostname of SCS instance
#
# Returns if previously not defined:
#   $scsinr - SAP instance number of SCS instance
#   $scshost- virtual hostname of SCS instance
#############################################################################
sub GetSCSIDAndSCSHOST {
    LogEnterSub( @_ );

    my $file    = shift;
    my $scsinr  = shift;
    my $scshost = shift;

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) { warn "Opening $file for reading failed"; }

    while (<$fh>) {
        chomp;
        if ( $scsinr eq "" and /^\s*(SCSID)\s*=\s*(\S+)\s*(.*)$/ ) {
            $scsinr = $2;
        }
        if ( $scshost eq "" and /^\s*(SCSHOST)\s*=\s*(\S+)\s*(.*)$/ ) {
            $scshost = $2;
        }
    }
    $fh->close;
    LogVerbose("Subroutine GetSCSIDAndSCSHOST() returning scsinr=[$scsinr] scshost=[$scshost]");
    return ( "$scsinr", "$scshost" );
}

#############################################################################
# This function scans a MDM ini file (= mds.ini, mdss.ini or mdis.ini)
# and adds the MDM repositories and/or the MDS admin user, MDS host to the
# appropriate hash
#
# Arguments:
#   -
#
# Returns
#   -
#
#############################################################################
sub GetMDMRepositories {
    LogEnterSub( @_ );

    foreach $prof (@profiles) {
        # seperate profile name into components
        my ( $sid, $inst, $vhost ) = split( /_/, basename ($prof) );

        LogVerbose("/usr/sap/${SID}/${inst}/config/");

        if ($prof =~ /${SID}_(MDS[0-9][0-9])_(\S+)/) {
            $mdm_ini_file="/usr/sap/${SID}/$1/config/mds.ini";
            $mdshost=$2;
            LogVerbose("mdshost=[$mdshost]");
            LogVerbose("Scanning [$mdm_ini_file]");
            FindMDSIniRespositoires($mdm_ini_file);
        }

        if ($prof =~ /${SID}_(MDSS[0-9][0-9])_(\S+)/) {
            $mdm_ini_file="/usr/sap/${SID}/$1/config/mdss.ini";
            LogVerbose("Scanning [$mdm_ini_file]");
            FindMDSUserAndMDSHost($mdm_ini_file);
        }
    }
}

#############################################################################
#
# This function scans the MDM mds.ini file for MDM repository definitions
# and adds them to the MDM assoc array %REPOSITORYmap;
#
# A repository is defined as:
#   [MDM Server/Databases/TEST_REPMO7ORCL_8_3_4_3]
#   ...
#   Login=abc
#   [MDM Server/Databases/XXXXMO7ORCL_8_3_4_3]
#   ...
#   Login=xyz
#
# Arguments:
#   $file   -   $mdm_ini_file="/usr/sap/${SID}/$1/config/mds.ini";
#
# Returns
#   -
#############################################################################
sub FindMDSIniRespositoires {
    LogEnterSub( @_ );

    my $file    = shift;

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) {
        #---------------------------------------------------------------------------
        # no need to warn as a MDM system might be down and then info just isn't
        # available for SGMgr. Warn if in verbose mode though
        #---------------------------------------------------------------------------
        LogVerbose "ATTENTION: Opening $file for reading failed - possibly MDM MDS filesystem not mounted?";
        return;
    } else {
      while (<$fh>) {
          if ( /^\[MDM Server\/Databases\/(\S+)\]$/ ) {
              $repository = $1;
              LogVerbose("Found repository $repository");
          } elsif ( /^\s*(Login)\s*=\s*(\S+)\s*/) {
            my $replogin=$2;
            LogVerbose "Found repository [$repository] Login=[$2]";
            CreateMDMRepositoryAssocArray($repository, $replogin);
          }
      }
      $fh->close;
    }

    # if verbose print the complete MDM mds repository hash
    my @printarr;
    @printarr = %REPOSITORYmap;
    LogVerbose("\%REPOSITORYmap=[@printarr]");
}

#############################################################################
#
# Find the DB/MDM SID for the MDM MDS repositories and add them to the
# assoc array %MDMSIDSmap for all MDM SIDs
#
# A </usr/sap/[A-Z][A-Z,0-9][A-Z,0-9]/MDS[0-9][0-9]/config/mds.ini> file
# can contain the following strings describing the repository, DB/MDM SID,
# database relationship
#   [MDM Server/Databases/TEST_REPMO7ORCL_8_3_4_3]
#   [MDM Server/Databases/XXXXMO7ORCL_8_3_4_3]
# Currently only Oracle is supported
#
# Arguments:
#   $file   -   $mdm_ini_file="/usr/sap/${SID}/$1/config/mds.ini";
#
# Returns
#   -
#
#############################################################################
sub FindMDSIniSIDS {
    LogEnterSub( @_ );

    my $file    = shift;

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) {
      LogVerbose "ATTENTION: Opening $file for reading failed - possibly MDM MDS filesystem not mounted?";
    }
    else {
      while (<$fh>) {
          if ( /^\[MDM Server\/Databases\/(\S+)([A-Z][A-Z,0-9][A-Z,0-9])ORCL_/ ) {
              $repository = $1;
              $dbsid=$2;
              $db=oracle;
              LogVerbose("Found repository[$repository] dbsid=[$dbsid] db=[$db]");
              $MDMSIDSmap{$dbsid}="$dbsid";

          }
          if ( /^\[MDM Server\/Databases\/(\S+)([A-Z][A-Z,0-9][A-Z,0-9])MXDB_/ ) {
              $repository = $1;
              $dbsid=$2;
              $db=maxdb;
              LogVerbose("Found repository[$repository] dbsid=[$dbsid] db=[$db]");
              $MDMSIDSmap{$dbsid}="$dbsid";
          }
      }
      $fh->close;
    }

}

#############################################################################
#
# This function takes a mds.ini repository definitions and adds it to
# MDM assoc array %REPOSITORYmap as a full repository specification;
#
# Arguments:
#   The raw $repository_string from the mds.ini file:
#   $repository_string  = TEST_REPMO7ORCL_8_3_4_3 -> [repository][SID][Database]
#
# Returns
#   -
#############################################################################
sub CreateMDMRepositoryAssocArray {
    LogEnterSub( @_ );

    my $repository_string=shift;
    my $repository_login=shift;
    my $repository;
    my $dbsid;
    my $database;

    if ( $repository_string =~ /(\S+)([A-Z][A-Z,0-9][A-Z,0-9])(ORCL_|MXDB_|IDB2_)/ ) {
     $repository = $1;
     $dbsid=$2;
     $database=$3;
     LogVerbose("Found repository[$repository] dbsid=[$dbsid] database=[$database]");
    }

    my $db="";
    if ( $database =~ /ORCL_/) { $db="o";}
    if ( $database =~ /MXDB_/) { $db="m";}
    if ( $database =~ /IDB2_/) { $db="d";}

    # add a full repository specification to hash
    if ( $db ne "" ) {
        $REPOSITORYmap{$repository}= "$repository:$dbsid:$db:$repository_login:";
    }
}


#############################################################################
# Scan an MDM ini file for definition of an MDS admin user or MDS host
#
# Arguments:
#   $file   =  mdm_ini_file="/usr/sap/${SID}/MDSS/config/mdss.ini"
#
# Sets:
#   $mdshost
#   $mdsuser
#
# Returns:
#   -
#############################################################################
sub  FindMDSUserAndMDSHost {
    LogEnterSub( @_ );

    my $file    = shift;

    LogVerbose("parsing file [$file] ...");

    my $fh = FileHandle->new($file);
    if ( !defined($fh) ) {

    #---------------------------------------------------------------------------
    # no need to warn as a MDM system might be down and then info just isn't
    # available for SGMgr
    #---------------------------------------------------------------------------
      LogVerbose "ATTENTION: Opening $file for reading failed - possibly MDM MDSS filesystem not mounted?";
      return;
    }

    while (<$fh>) {
#        if ( /^MDM Server Name=([0-9]+@)?(\w+)([:][0-9]+)?$/ ) {
#            $mdshost = $2;
#            LogVerbose("Found mdshost=[$mdshost]");
#        }
        if ( /^User=(\S+)$/ ) {
            $mdsuser = $1;
            LogVerbose("Found mdsuser=[$mdsuser]");
        }
    }
    $fh->close;
}

#############################################################################
#
# Output of SAP/HANA SID for SGMgr
#
# Arguments:
#    $sapsid - SAP/HANA SID
#
#
#############################################################################

sub FormatSIDForSGMgr {
    LogEnterSub( @_ );

    my $sapsid = shift;
    my $leftside = "sgesap/sap_global/sap_system"; # standard prefix

    if  (-d (</usr/sap/${sapsid}/HDB[0-9][0-9]>)[0]) {
      # format as HDB
      $leftside = "sgesap/hdb_global/hdb_system";
    }

    printSGMgr($leftside."=".$sapsid);

    LogReturnSub();
}

#############################################################################
#
# Output for IP(s) and subnet(s)
#
# Arguments:
#    $vhost - virtual hostname
#
# Requires:
#     either external helper program ("nwrksupp") or - if this isn't
#     available - addrs and subnets hashes (->CreateIPAddressAndSubnetAssocArray)
#
#############################################################################

sub FormatIPForSGMgr {
    LogEnterSub( @_ );

    my $vhost = shift;

    if ( !$vhost ) {
        LogReturnSub();
        return;
    }

    # generate IP address out of virtual hostname
    # Look for IP helper program, which will give IPv4 as well as IPv6
    # addresses and subnets matching with vhost
    # Otherwise faill back to using previously created associative
    # array.code below could then
    # Be aware: There can be more than one IP address for a hostname,
    ## esp. one for IPv4 and one for IPv6.

    if (-x $nwrksupp)
    {
      my $cmd = "$nwrksupp ghbns $vhost";
      LogVerbose("Execute cmd=[$cmd]");
      my @ipsubn = `$cmd` or warn "cmd=[$cmd], return code [$!]";
      LogVerbose("ipsubn=[@ipsubn]");
      my $ipprot, $addr, $subnet, $interf;
      my %laddr = ();
      for $_ (@ipsubn) {
        ( $ipprot, $addr, $subnet, $interf ) = split( /;/, $_ );
        # duplicate pairs possible - try to filter them out
        if ($laddr{$addr} != $subnet)
        {
           $laddr{$addr}=$subnet;
           printSGMgr("ip_subnet:", $subnet, "|ip_subnet=", $subnet, "");
           printSGMgr("ip_subnet:", $subnet, "|ip_address=", $addr, "");
        }
      }
    }
    else
    {
      # If addresses isn't defined return instead of continue executing
      @addresses = gethostbyname($vhost);
      if ( ! @addresses ) {
        warn
        "Can't get IP address: gethostbyname(vhost=[$vhost]): \$?=[$?] \$!=[$!]";
        return;
      }
      LogVerbose("addresses=[@addresses]");

      # convert binary IP address into ascii format. Multiple IP addresses
      # are contained in addresses array starting at index 4
      @addresses = map { inet_ntoa($_) } @addresses[ 4 .. $#addresses ];
      if ( ! @addresses ) {
        warn "Failed map { inet_ntoa($_) }: \$?=[$?] \$!=[$!]";
      return;
      }

      if ( $addresses[0] ne "" ) {
        $subnet = $addrs{ $addresses[0] };
        LogVerbose("vhost=[$vhost] addresses=[@addresses] addresses[0]=[$addresses[0]] subnet=[$subnet]");

        if ( $subnet eq "" ) {
          # split own address in the four parts
          @addrparts = split( /\./, $addresses[0] );

          for $sn ( keys %subnets ) {
            @snparts = split( /\./, $sn );
            # Do a "next if ..." below
            $x = int $addrparts[0] & int $snparts[0];    # print $x,"\n";
            if ( $x != int $snparts[0] ) { next; }
            $x = int $addrparts[1] & int $snparts[1];    # print $x,"\n";
            if ( $x != int $snparts[1] ) { next; }
            $x = int $addrparts[2] & int $snparts[2];    # print $x,"\n";
            if ( $x != int $snparts[2] ) { next; }
            $x = int $addrparts[3] & int $snparts[3];    # print $x,"\n";
            if ( $x != int $snparts[3] ) { next; }

            $subnet = $sn;
            last;
          }
        }
        printSGMgr("ip_subnet:", $subnet, "|ip_subnet=", $subnet, "");
        printSGMgr("ip_subnet:", $subnet, "|ip_address=", $addresses[0], "");
      }
      else {
        LogError("Can't resolve hostname ", $vhost, " to IP address");
      }
    }
    LogReturnSub();
}

#############################################################################
#
# Output for SAP instances (incl. IP,...)
#
# Input (optional): List of instance names to be selected for output
#
# Loops through profiles to get list of instances
#
#############################################################################

sub FormatInstancesForSGMgr {
    LogEnterSub( @_ );
    my @sapinsts=split(',',shift);

    LogVerbose "[ @sapinsts ]";

    # maps for later services output
    %sapinsttypelist=(); # map of instances types (w/o instance number) for which properties are printed out
    $sapfoundinsttylist=(); # map of found instance types while going through the profiles

    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");

    # loop thru all available SAP profiles
    foreach $prof (@profiles) {

      #---------------------------------------------------------------------------
      #   SGMgr output $inst:      ...D03
      #   SGMgr output $vhost:     ...reloc123
      #---------------------------------------------------------------------------

      # Get the profile filename from the absolute path
      my $profname;
      $profname = ( split( /\//, $prof ) )[4];

      # get instance name and (virtual) hostname out of the profile file name.
      # A SAP profile has the following format <SID>_<INST>_<HOST>
      my ( $sid, $inst, $vhost ) = split( /_/, $profname );

      # enforce SAP instance naming rules
      next
          unless ( $inst
           =~ /^(DVEBMGS|D|ASCS|SCS|ERS|REP|J|JD|JC|MDS|MDIS|MDSS|W|G)[0-9][0-9]$/
          );
      $sapfoundinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      # skip if not in list of selected instances (if such a list is defined)
      if (scalar @sapinsts != 0)
      {
           # restricted to list of instances. Check if instance is in list.
           next unless grep {$_ eq $inst} @sapinsts;
      }

      $sapinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      printSGMgr("sgesap/stack/sap_instance:", $inst, "|sgesap/stack/sap_instance=", $inst, "");
      printSGMgr("sgesap/stack/sap_instance:", $inst, "|sgesap/stack/sap_virtual_hostname=", $vhost, "");

      #---------------------------------------------------------------------------
      #  If the instance is a ERS instance output the matching SCS instance
      #  SGMgr output $scsinst:      ...SCS03
      #---------------------------------------------------------------------------
      if ( $inst =~ /^ERS/ ) {
          my $scsinst;
          $scsinst = $ERSmap{$inst};
          LogDebug("inst=[", $inst, "], scsinst=[", $scsinst, "]");
          if ( $scsinst ne "" ) {
              printSGMgr("sgesap/stack/sap_instance:", $inst, "|sgesap/stack/sap_replicated_instance=", $scsinst, "");
          }
          else {
              LogError("No [A]SCS configured for ", $inst, "");
          }

          #  Output resource data for ERS instance:
          #  SGMgr output $resname:  resource:/applications/sap/enqor/ers33
          #  ...
          FormatResourceForSGMgr($SID,$inst);

          #
          #  Output service data for ERS instance (duplicate if more than one ERS is configured)
          #
          FormatServiceForSGMgr("$pkgname", "enqr");
      }   # end if ERS instance

      #---------------------------------------------------------------------------
      # If the [A]SCS instance has a corresponding ERS instance configured
      # SGMgr output $resname:  resource:/applications/sap/enqor/ers33
      #
      # need to find matching ERS to build resource name
      # maybe we build an associative array here and postpone output
      # because output should be available when running over ERS.
      #---------------------------------------------------------------------------
      if ( $inst =~ /^ASCS|^SCS/ ) {
          #
          # resource
          #
          FormatResourceForSGMgr($SID,$inst);

          #
          # service(s)
          #
          FormatServiceForSGMgr("$pkgname", "enq");
          FormatServiceForSGMgr("$pkgname", "ms");
      }    #end if [A]SCS instance

      if ( $inst =~ /^D/ ) {
        FormatServiceForSGMgr("$pkgname", "disp");
        if ( ($inst =~ /^DVEBMGS/) && !(grep {$_ =~ /SCS[0-9][0-9]/} @profiles)) {
          # no SCS profile, MS part of DVEBMGS (profile content not checked)
          FormatServiceForSGMgr("$pkgname", "ms");
        }
      }

      if ( $inst =~ /^W/ ) {
        FormatServiceForSGMgr("$pkgname", "webdisp");
      }

      if ( $inst =~ /^G/ ) {
        FormatServiceForSGMgr("$pkgname", "gw");
      }
      if ( $inst =~ /^MDS|^MDIS|^MDSS/ ) {

        # prepare MDM repositories for the MDS instance only
        FormatMDMForSGMgr() if ( $inst =~ /^MDS[0-9][0-9]/ ) ;

        # prepare monitor data for the serice_cmd
        FormatServiceForSGMgr($pkgname, "mdm");
      }
      #---------------------------------------------------------------------------
      # get the address/subnet information
      # SGMgr output $ip_subnet         ...16.41.100.0
      # SGMgr output $ip_address        ...16.41.100.48
      #---------------------------------------------------------------------------

      FormatIPForSGMgr($vhost);

      #---------------------------------------------------------------------------
      # get VG/FS information
      # SGMgr output $vg          vg=vgmo7MDS
      # SGMgr output $instfs      fs_name=/dev/vgmo7MDS/lvmo7MDS
      # SGMgr output $instdir     fs_directory=/usr/sap/MO7/MDS01
      # SGMgr output $instfst     fs_type=vxfs
      # (can only return useful information if instance directory
      # is already mounted)
      #---------------------------------------------------------------------------

      my $instdir = "/usr/sap/$SID/$inst";
      LogVerbose ("\$instdir=[$instdir]");

      # search assoc. arrays
      my $instfs  = $mpsfs{$instdir};
      if ( $instfs ne "" ) {
        FormatSingleFileSystemForSGMgr ($instdir, $instfs,  $mpsfst{$instdir}) ;
      }
	  else {
	    # for simple installations mp might be /usr/sap/SID
		$instdir = "/usr/sap/$SID";
		$instfs  = $mpsfs{$instdir};
        if ( $instfs ne "" ) {
          FormatSingleFileSystemForSGMgr ($instdir, $instfs,  $mpsfst{$instdir}) ;
        }
	  }

      # check if vhost has an "asscociated" SMDA instance
      # needs only to be done for instances != ASCS,SCS;ERS,W
      # and if vhost wasn't already checked
      if ( ($inst !~ /^ASCS|^SCS|^ERS|^W/ ) && !defined $SMDAVHmap{vhost} ) {
        LogVerbose ("Looking up SMDA for $vhost");
        $SMDAVHmap{vhost}=1;
        FormatSMDAForSGMgr ($vhost);
      }
    }    # end foreach profile
}    # End FormatForSGMgr()

#############################################################################
#
# Output for SAP instances (incl. IP,...)
#
# Input (optional): List of instance names to be selected for output
#
# Loops through profiles to get list of instances
#
#############################################################################

sub FormatHDBInstancesForSGMgr {
    LogEnterSub( @_ );
    my @sapinsts=split(',',shift);

    my %printedinsts = (); # make sure a HDB instance is just printed once

    LogVerbose "[ @sapinsts ]";


    # loop thru all available SAP profiles
    foreach $prof (@profiles) {

      #---------------------------------------------------------------------------
      #   SGMgr output $inst:      ...D03
      #   SGMgr output $vhost:     ...reloc123
      #---------------------------------------------------------------------------

      # Get the profile filename from the absolute path
      my $profname;
      $profname = ( split( /\//, $prof ) )[-1];

      # get instance name and (virtual) hostname out of the profile file name.
      # A SAP profile has the following format <SID>_<INST>_<HOST>
      my ( $sid, $inst, $vhost ) = split( /_/, $profname );

      # enforce SAP instance naming rules
      next unless ( $inst =~ /^(HDB)[0-9][0-9]$/);
      $sapfoundinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      # skip if not in list of selected instances (if such a list is defined)
      if (scalar @sapinsts != 0)
      {
           # restricted to list of instances. Check if instance is in list.
           next unless grep {$_ eq $inst} @sapinsts;
      }

      $sapinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      # printSGMgr("sgesap/hdb_global/hdb_system:", $sid, "|sgesap/hdbinstance/hdb_instance=", $inst, ""); # <= 06.00.30
      unless( defined $printedinsts{$inst} ) {
        printSGMgr("sgesap/hdb_global/hdb_system:", $sid, "|sgesap/hdbinstance_global/hdb_instance=", $inst, "");   # >= 06.00.40
        $printedinsts{$inst} = 1;
      }
      printSGMgr("sgesap/hdb_global/hdb_system:", $sid, "|sgesap/hdbinstance_global/_hdb_node=", $vhost, "");


    }    # end foreach profile
}    # End FormatForSGMgr()

#############################################################################
#
# Output SMDA info (if any) for SGMgr
# Supports only SMDA with SID starting with "DA"
# and instance names starting with "SMDA"
#
# Arguments:
#    $vhost - virtual hostname
#
# Dervives VG, FS and mountpoint for a SMDA instance for the vhost
# There's no need to print out vhost and IP related info. Is identical to the related instance.
# Instance on vhost  may not have an SMDA.
# Existence of a profile of the form
# /usr/sap/<SID>/SYS/profile/<SID>_SDMA<nn>_<vhost> indicates the
# existence of such a SDMA. "SDMA" and vhost are known, <SID> and <nn> need
# to be pattern matched. There may be more than one, however only one can
# be active.
#
# SMDA can be started up as external instance (5.10 PHSS 40911) but output for this setup
# isn't provided here.
#
#############################################################################

sub FormatSMDAForSGMgr {

    LogEnterSub( @_ );
    my $vh = shift;

    my @smdaprof =
        </usr/sap/DA[A-Z,0-9]/SYS/profile/DA[A-Z,0-9]_SMDA[0-9][0-9]_$vh>;
#        previously, but might hang on stale NFS (but above may also hang, when on NFS !!??):
#        </usr/sap/[A-Z][A-Z,1-9][A-Z,1-9]/SYS/profile/[A-Z][A-Z,1-9][A-Z,1-9]_SMDA[1-9][1-9]_$vh>;

    my $numprof = @smdaprof;

    if ( $numprof == 0 ) {
       LogVerbose ("No SMDA found for $vh");
    }
    elsif ($numprof > 1) {
       LogVerbose ("More than one SMDA for for $vhost - ignoring them");
    }
    else {
       LogVerbose ("SDMA profile found: $smdaprof[0]");
       my $lastpart = ( split( /\//, $smdaprof[0] ) )[6];
       ($sid,$inst,$dc) =  split ( /_/, $lastpart);
       LogVerbose ("\$sid=[$sid], \$inst=[$inst]");
       my $smdaid = "/usr/sap/$sid/$inst";
       LogVerbose ("\$smdaid=[$smdaid]");

       if (-d $smdaid) {
          # no sep. mount -v again, use "widened" mount map
          my $smdafs=$mpsfs{$smdaid};
          if (defined ($smdafs)) {
            LogVerbose ("$smdafs mounted to $smdaid");
            FormatSingleFileSystemForSGMgr( $smdaid, $smdafs, $mpsfst{$smdaid} );
          }
          else {
            warn "Instance [$inst] for host [$vh] found, but no filesystem mounted onto instance directory [$smdaid]";
          }
       }
       else {
            warn "Profile [$smdaprof[0]] found, but no instance directory [$smdaid]";
       }
       printSGMgr("sgesap/sapextinstance/sap_ext_instance:", $inst, "|sgesap/sapextinstance/sap_ext_instance=", $inst, "");
       printSGMgr("sgesap/sapextinstance/sap_ext_instance:", $inst, "|sgesap/sapextinstance/sap_ext_system=", $sid, "");
       printSGMgr("sgesap/sapextinstance/sap_ext_instance:", $inst, "|sgesap/sapextinstance/sap_ext_host=", $vh, "");  #or "localhost"?
       printSGMgr("sgesap/sapextinstance/sap_ext_instance:", $inst, "|sgesap/sapextinstance/sap_ext_treat=yynnn", "");
    }
    LogReturnSub();
} # FormatSMDAForSGMgr

#############################################################################
#
# Output "service" data for SAP instances  for SGMgr
#
#
# Goes throught the instance type list for which properties has been printed out.
# Needs to compare them to the list of found instances, because existence of other instance might influence the services of the current
# e.g. DVEBMGS will have no sapms.mon if a ASCS exists.
# Implicit Inputs:
# sapinsttypelist: List of instance types for which output was printed
# sapfoundinsttypelist: List of instance types found while going through the profiles
#
# Service name needs to be unique. Done by using environment variable $SG_PACKAGE
# or - if this is undefined - the string "${SG_PACKAGE}". In the latter case
# the tool reading the output has to substitute the value.
#
# The first routine isn't used yet.
#
#############################################################################

sub FormatServiceSAPinstForSGMgr {
    LogEnterSub( @_ );

    # print out service setup for instance
    # may depend on other instances (e.g. DVEBMGS will have no sapms.mon if there is a ASCS)
    # print out service per instance or at the end. First may lead to duplicates but the relationship to the instance can be
    # established (anticipated SGMgr usage).
    # this is similar to the IP address: if more than one instance is configured to the same IP address, the address appears after
    # each instance, but is kind of duplicate from the point of view for SGMgr
    # TBD: so far the service name is build with the package name by SGMgr. So far this information isn't avaiable for this tool.
    # Example: (pkg name is MscsVM7)
    # service:MscsVM7ms|name=MscsVM7ms
    # service:MscsVM7ms|failfast=disabled
    # service:MscsVM7ms|halt_timeout=0
    # service:MscsVM7ms|command=<install-dir>/sap/SID/sapms.mon
    # service:MscsVM7ms|restart=none

    # get package name. Used to build unique service names. Get it either from the environment or set
    # it to ${SG_PACKAGE} for later replacement
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");

    # special handling for DVEBMGS first
    if ($sapinsttypelist{"DVEBMGS"} == 1) {
        # dispatch monitor
        FormatServiceForSGMgr($pkgname, "disp");
        # and msg server monitor depending if DVEBMGS and !ASCS and !SCS in found list
        FormatServiceForSGMgr($pkgname, "ms")
           if (($sapfoundinsttypelist{"ASCS"} != 1) && ($sapfoundinsttypelist{"SCS"} != 1));
        # make sure there are no more dispatcher services printed out
        $sapinsttypelist{"DVEBMGS"} = 0;
        $sapinsttypelist{"D"} = 0;
    }

    foreach ( keys %sapinsttypelist) {
      next unless $sapinsttypelist{$_} != 0;
      LogVerbose ("Service for instance type: $_");
      if ($_ eq "D") {
        # dispatch monitor
        FormatServiceForSGMgr($pkgname, "disp");
        # and msg server monitor depending if DVEBMGS and !ASCS in found list
        FormatServiceForSGMgr($pkgname, "ms")
           if (($_ eq "DVEBMGS") && ($sapfoundinsttypelist{"ASCS"} != 1));
      }
      elsif (($_ eq "ASCS") || ($_ eq "SCS")) {
        # enq monitor
          FormatServiceForSGMgr($pkgname, "enq");
          FormatServiceForSGMgr($pkgname, "ms");
          # check if there is another (A)SCS in package, set that one to 0 to avoid output
          $sapinsttypelist{"SCS"}  = 0 if ($_ eq "ASCS");
          $sapinsttypelist{"ASCS"} = 0 if ($_ eq "SCS");
      }
      elsif ($_ eq "ERS") {
        # enqr monitor
        FormatServiceForSGMgr("$pkgname", "enqr");
        # no need to check of other ERS here as we use only one monitor for both and one instance type
      }
      elsif ($_ eq "W") {
        # webdisp monitor
        FormatServiceForSGMgr("$pkgname", "webdisp");
      }
      else {
        # no output for other type of instances
      }
    }
    LogReturnSub();
}

#############################################################################
#
# FormatResourceForSGMgr
#
# Output resource sections for [A]SCS & ERS instances
# So far HP-UX only
#
# Uses SCSmap (hash)
#
# Parameter:
#  - SAP SID
#  - SAP Instance
#
#############################################################################
sub FormatResourceForSGMgr {
    LogEnterSub( @_ );

    my $SID=shift;
    my $inst=shift;
	my $sgmodname="resource:";

	# return immediately if not ERS,SCS,ASCS
	return unless ( $inst =~ /^ERS|^ASCS|^SCS/ );

	if ($^O eq "linux")  {
      # generic resource and related dependency on Linux, but
      # return immediatly if SGeSAP_ENQOR_PKGNAME env var isn't set
	  return unless (defined($ENV{"SGeSAP_ENQOR_PKGNAME"}));
	  FormatGenResourceForSGMgr ($SID,$inst);
	  FormatDependencyForSGMgr ();
      return;
    }

    # use ERS as default
    my $poll="60";
	my $upvalue="SERVING";
	my $ersinst;
	my $ersscsinst;

    $ersinst=$inst if ( $inst =~ /^ERS/ );

    #---------------------------------------------------------------------------
    # If the [A]SCS instance has a corresponding ERS instance configured
    # SGMgr output $resname:  resource:/applications/sap/enqor/ers33
    #
    # need to find matching ERS to build resource name
    # maybe we build an associative array here and postpone output
    # because output should be available when running over ERS.
    #---------------------------------------------------------------------------
    # if not ERS instance, lookup ERS for SCS (might not exist)
    unless (defined ($ersinst)) {
	  $ersscsinst = $SCSmap{$inst};
	  return unless defined $ersscsinst; # neiter ERS nor SCS with ERS
	}

	# is ERS or is SCS with  ERS
	my $resname =
         "/applications/sap/enqor/" . $SID . "ers". substr( ($ersinst||$ersscsinst), -2 );
    if ( defined $ersscsinst) {
	  $upvalue = "IDLE";
	  $poll = "30"
	}

    printSGMgr($sgmodname, $resname, "|name=", $resname, "");
    printSGMgr($sgmodname, $resname, "|start=automatic");
    printSGMgr($sgmodname, $resname, "|polling_interval=",$poll,"");
    printSGMgr($sgmodname, $resname, "|up_value=!= ",$upvalue,"");

    LogReturnSub();
}

#############################################################################
#
# FormatGenResourceForSGMgr
#
# Output generic resource sections for [A]SCS & ERS instances.
# Similar to FormatResourceForSGMgr, but different (and less) attributes
# So far Linux only
#
# Uses SCSmap (hash)
#
# Parameter:
#  - SAP SID
#  - SAP Instance
#
#############################################################################
sub FormatGenResourceForSGMgr {
    LogEnterSub( @_ );

    my $SID=shift;
    my $inst=shift;
	my $sgmodname="generic_resource:";

	# immediate return if not ERS,SCS,ASCS
	return unless ( $inst =~ /^ERS|^ASCS|^SCS/ );

    # use ERS as default
    #my $poll="60";
	my $upvalue="2"; # SERVING
	my $ersinst;
	my $ersscsinst;

	$ersinst=$inst if ($inst =~ /^ERS/);
    # if not ERS instance, lookup ERS for SCS (might not exist)
    unless (defined ($ersinst)) {
	  $ersscsinst = $SCSmap{$inst};
	  return unless defined $ersscsinst; # neiter ERS nor SCS with ERS
	  $upvalue = "1"; # IDLE
	}

    # is ERS or is SCS with ERS
	my $resname =
    #     "sgesap.enqor_" . $SID . "_ERS". substr( ($ersinst||$ersscsinst), -2 );
	     "sgesap.enqor_" . $SID . "_" . ($ersinst||$ersscsinst);

	printSGMgr($sgmodname, $resname, "|name=", $resname, "");
    printSGMgr($sgmodname, $resname, "|evaluation_type=before_package_start");
    printSGMgr($sgmodname, $resname, "|up_value=\"!= ",$upvalue,"\"");

    LogReturnSub();
}

#############################################################################
#
# FormatDependencySGMgr
#
# Output dependency section for [A]SCS & ERS instances.
# So far Linux only. Assumes that it is already called in the correct context
# (Linux, SCS/ASCS/ERS, environment variable set)
#
#
#############################################################################
sub FormatDependencyForSGMgr {
	my $depname="enqor_dep";
	my $leftpart="dependency:".$depname."|";

	printSGMgr($leftpart."name=".$depname);
	printSGMgr($leftpart."location=same_node");
	printSGMgr($leftpart."condition=".$ENV{"SGeSAP_ENQOR_PKGNAME"}." = up");
}
#############################################################################
#
# FormatHDBDependencyForSGMgr
#
# Output dependency section HDB. Skipped of undefined parameter is submitted
#
#############################################################################
sub FormatHDBDependencyForSGMgr {
  LogEnterSub( @_ );

  my $deppkgname=shift;
  return unless (defined ($deppkgname));

  my $leftpart="dependency:".$deppkgname."_rep|";

  printSGMgr($leftpart."name=".$deppkgname."_rep");
  printSGMgr($leftpart."location=same_node");
  printSGMgr($leftpart."condition=".$deppkgname."=down");
  LogReturnSub();
}
#############################################################################
#
# FormatHDBPriorityForSGMgr
#
# Output priority section for HDB primary/secondary instances.
#
#############################################################################
sub FormatHDBPriorityForSGMgr {
  LogEnterSub( @_ );

  printSGMgr("priority=".shift);

  LogReturnSub();
}

#############################################################################
#
# FormatServiceForSGMgr
#
# Output helper for FormatServiceSAPinstForSGMgr
#
# Related default:
# $SG_SERVICE_RESTART_DEFAULT="none"
# $SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT="no" ("disable" required below)
# $SG_SERVICE_HALT_TIMEOUT_DEAULT="0"
#
#############################################################################

sub FormatServiceForSGMgr {

  LogEnterSub( @_ );

  # suppress output if variable isn't set
  return unless $servoutsgmgr;

  my $pkgname=shift;
  my $postfix=shift;
  # my $svccmd=shift;
  my $svccmd="$SGeSAPSVCDIR/sap".$postfix.".mon",
  my $svcname=$pkgname.$postfix;
  my $leftpart="service:".$svcname."|";

  printSGMgr($leftpart."name=".$svcname);
  printSGMgr($leftpart."failfast=disabled");
  printSGMgr($leftpart."halt_timeout=".$SG_SERVICE_HALT_TIMEOUT_DEAULT);
  printSGMgr($leftpart."command=".$svccmd);
  printSGMgr($leftpart."restart=".$SG_SERVICE_RESTART_DEFAULT);

  LogReturnSub();
}

#############################################################################
#
# Output "filesystem" data for SGMgr
#
#############################################################################

sub FormatFileSystemForSGMgr {
    LogEnterSub( @_ );

    foreach ( sort (keys %mpsfs) ) {
        LogDebug("$mpsfs{$_} -- $mpsfst{$_} ");
        FormatSingleFileSystemForSGMgr($_,$mpsfs{$_},$mpsfst{$_});
    }
    LogReturnSub();
}

#############################################################################
#
# Output "single" filesystem and its VG for SGMgr
#
# Derives the VG from the FS path
#
# Arguments:
#   - directory the FS is mounted to
#   - FS
#   - FS type
#
#
#############################################################################

sub FormatSingleFileSystemForSGMgr {
    LogEnterSub( @_ );

    my $dir = shift;
    my $fs  = shift;
    my $fst = shift;
    my $vg = ( split( /\//, $fs ) )[2];

    # if VG is on VxVM instead of LVM, vg is "vx" and
    # instfs has the form /dev/vx/dsk/<vx dgrp>/<vx vol>
    # instead of "vg:", output is "vxvm_dg:<vx_dgrp> | .."
    # to see if the VxVM DG belongs to a cluster
    # vxdg list should be executed and STATE (field 2) should be checked if it
    # contains attribute "shared".
    # Linux: FS have the form /dev/mapper/<vgname>-<lvname>

    if ( ( $vg eq "vx" ) && $vxvm_installed ) {
      $vg = ( split( /\//, $fs ) )[4];
      if ( $vxdgshared{$vg} ) {
         printSGMgr("cvm_dg:$vg|cvm_dg=$vg");
      }
      else {
         printSGMgr("vxvm_dg:$vg|vxvm_dg=$vg");
      }
    }
    # mapper on Linux
    elsif ($vg eq "mapper") {
       my $vglv = ( split( /\//, $fs ) )[3];
       $vglv =~ m#([A-Za-z0-9_.]+)-([A-Za-z0-9_.]+)#;
       printSGMgr("vg:", $1, "|vg=", $1, "");
    }
    # otherwise assume form /dev/<vgname>/<lvname>
    else {
       printSGMgr("vg:", $vg, "|vg=", $vg, "");
    }

    printSGMgr("fs_name:", $fs, "|fs_name=", $fs, "");
    printSGMgr("fs_name:", $fs, "|fs_directory=", $dir, "");
    printSGMgr("fs_name:", $fs, "|fs_type=", $fst, "");

    LogReturnSub();
}

#############################################################################
#
# Output "liveCache" data for SGMgr.
#
# Arguments:
#   $dbsid     SID of the liveCache
#   $sapdbhost Virtual hostname for the liveCache instance
#
#############################################################################

sub FormatLivecacheForSGMgr {
    LogEnterSub( @_ );

    my $dbsid     = shift;
    my $sapdbhost = shift;
    my $hssenable = shift;

    # the SGMgr liveCache prefix  for later use
    my $lcs="sgesap/livecache";

    return unless $dbsid;

    printSGMgr("$lcs/lc_system:$dbsid|$lcs/lc_system=$dbsid");
    printSGMgr("$lcs/lc_system:$dbsid|$lcs/lc_virtual_hostname=$sapdbhost")
        if ($sapdbhost);
    printSGMgr("$lcs/lc_system:$dbsid|$lcs/lc_startmode=online");
    printSGMgr("$lcs/lc_system:$dbsid|$lcs/lc_userkey=c");

    # printSGMgr("sgesap/livecache/lc_system:$dbsid|sgesap/livecache/whichparam?=what?") if ($hssenable);

     # output IP & FS data for SGMgr
    FormatIPForSGMgr($sapdbhost);
    FormatFileSystemForSGMgr();

    # service output for LC
    # Standard formatting routine not used, because LC monitors follows a different schema
    if ($servoutsgmgr) {
      my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
      my $svccmd="/sapdb/$dbsid/db/sap/lccluster monitor";
      # not "/opt/cmcluster/sap/SID/saplc.mon"
      my $svcname=$pkgname."lc";
      my $leftpart="service:".$svcname."|";

      printSGMgr($leftpart."name=".$svcname);
      printSGMgr($leftpart."failfast=disabled");
      printSGMgr($leftpart."halt_timeout=0");
      printSGMgr($leftpart."command=".$svccmd);
      printSGMgr($leftpart."restart=none");
    }

    LogReturnSub();
}

#############################################################################
# output the MDM information required by SGMgr
#############################################################################
sub FormatMDMForSGMgr{
    LogEnterSub( @_ );

    # the SGMgr MDM_spec string for later use
    my $m_s="sgesap/mdm_spec";
    my $h_s="mdm_mdshostspec";
    my $c_s="mdm_credentialspec";
    my $r_s="mdm_repositoryspec";

    # NOTE: no passwords are available for discovery
    if ($mdsuser && $mdshost) {
      printSGMgr("$m_s/${h_s}_host:$mdshost|$m_s/${h_s}_host=$mdshost");
      printSGMgr("$m_s/${c_s}_user:$mdsuser|$m_s/${c_s}_user=$mdsuser");
    }

    for $rep ( keys %REPOSITORYmap ) {
       LogVerbose("\$REPOSITORYmap{$rep}=[$REPOSITORYmap{$rep}]");

       my ( $repname, $dbsid, $dbtype, $dbuser ) = split( /:/, $REPOSITORYmap{$rep} );
       printSGMgr("$m_s/${r_s}_repname:$repname|$m_s/${r_s}_repname=$repname");
       printSGMgr("$m_s/${r_s}_repname:$repname|$m_s/${r_s}_dbsid=$dbsid");
       printSGMgr("$m_s/${r_s}_repname:$repname|$m_s/${r_s}_dbtype=$dbtype");
       printSGMgr("$m_s/${r_s}_repname:$repname|$m_s/${r_s}_dbuser=$dbuser");
    }

    LogReturnSub();
}

#############################################################################
# output the MDM alert/service_ info required by SGMgr  -- not used anymore
#############################################################################
sub FormatMDMAlertForSGMgr{
    LogEnterSub( @_ );

    printSGMgr("service_cmd=$SGeSAPSVCDIR/sapmdm.mon");
    printSGMgr("service_restart=None");
    printSGMgr("service_fail_fast_enabled=no");
    printSGMgr("service_halt_timeout=0");

    LogReturnSub();
}

#############################################################################
# output the MDM alert/service_ info required by SGMgr
#############################################################################
sub FormatMDMSIDSForSGMgr{
    LogEnterSub( @_ );

    for $sids ( keys %MDMSIDSmap ) {
        printSGMgr("sgesap/sap_global/sap_system=$MDMSIDSmap{$sids}");
    }
    LogReturnSub();
}

#############################################################################
#
# Package file systems ...
#
#fs_name                                 /dev/vgtest/lvtest
#fs_directory                            /export/mnfstest
#fs_type                                 vxfs
#fs_mount_opt                            -o rw
#fs_umount_opt                           -s
#fs_fsck_opt                             -s
#
# Calls FormatSingleFileSystemForSGMgr to format VG and FS
#
#############################################################################
sub FormatNfsForSGMgr{
    LogEnterSub( @_ );

    # no printout of defaults anymore
    #printSGMgr("nfs/hanfs_export/FILE_LOCK_MIGRATION=0");
    #printSGMgr("nfs/hanfs_export/MONITOR_INTERVAL=10");
    #printSGMgr("nfs/hanfs_export/MONITOR_LOCKD_RETRY=4");
    #printSGMgr("nfs/hanfs_export/MONITOR_DAEMONS_RETRY=4");
    #printSGMgr("nfs/hanfs_export/PORTMAP_RETRY=4");

    if ($^O eq "linux") {
        FormatNfsForSGMgrLinux();
    } else {
         FormatNfsForSGMgrHPUX();
    }
    LogReturnSub();
}

#############################################################################
#############################################################################
sub FormatNfsForSGMgrLinux{
    LogEnterSub( @_ );

  	my $attrprefix="tkit/nfs/nfs";

    for $pathname (keys %XFSPathnameOptionsmapLinux ) {

        # Linux can have multiple export lines for same pathname
        for $nfsclient (keys %{$XFSPathnameOptionsmapLinux{$pathname}}) {
            # get options for each nfs client
            $opt=$XFSPathnameOptionsmapLinux{$pathname}{$nfsclient};
            printSGMgr("$attrprefix/XFS:\"-o $opt $nfsclient:$pathname\"|$attrprefix/XFS=\"-o $opt $nfsclient:$pathname\"");

            FormatSingleFileSystemForSGMgr ($pathname, $mpsfs{$pathname}, $mpsfst{$pathname}) if defined $mpsfs{$pathname};
            FormatIPForSGMgr($AutodirNodemap{$pathname}) if defined $AutodirNodemap{$pathname};
        }
    }
    LogReturnSub();
}

#############################################################################
#############################################################################
sub FormatNfsForSGMgrHPUX{
    LogEnterSub( @_ );

	my $attrprefix="nfs/hanfs_export";

    for $pathname (keys %XFSPathnameOptionsmap ) {
        # NFS related output (need to clarify if module default values should be provided here)
        #printSGMgr("nfs/hanfs_export/XFS                     \"-o $XFSPathnameOptionsmap{$pathname} $pathname\"");
        # cmviewcl -f line format:
        printSGMgr("$attrprefix/XFS:\"-o $XFSPathnameOptionsmap{$pathname} $pathname\"|$attrprefix/XFS=\"-o $XFSPathnameOptionsmap{$pathname} $pathname\"");
        # Using generic FormatSingleFilesystemForSGMgr (which also includes the VG output)
        # Use "if defined". Otherwise fields will be empty except fs_directory
        # Alternative: use kind of placeholders, so that partial information can be reported
        FormatSingleFileSystemForSGMgr ($pathname, $mpsfs{$pathname}, $mpsfst{$pathname}) if defined $mpsfs{$pathname};
        FormatIPForSGMgr($AutodirNodemap{$pathname}) if defined $AutodirNodemap{$pathname};
    }
    LogReturnSub();
}

#############################################################################
#
# Format output to SGMgr for volume groups
# (used only by FormatNfsForSGMgr so far)
#
#############################################################################

sub FormatVGForSGMgr {
    LogEnterSub( @_ );

    my $fs  = shift;

    my $vg = ( split( /\//, $fs ) )[2];

    # if VG is on VxVM instead of LVM, vg is "vx" and
    # instfs has the form /dev/vx/dsk/<vx dgrp>/<vx vol>
    # instead of "vg:", output is "vxvm_dg:<vx_dgrp> | .."
    # to see if the VxVM DG belongs to a cluster
    # vxdg list should be executed and STATE (field 2) should be checked if it
    # contains attribute "shared".

   if ( ( $vg eq "vx" ) && $vxvm_installed ) {
      # VxVM (or CVM?)
      $vg = ( split( /\//, $fs ) )[4];
      if ( $vxdgshared{$vg} ) {
        printSGMgr("cvm_dg                                   $vg");
      }
      else {
        printSGMgr("vxvm_dg                                  $vg");
      }
    }
    else {
        printSGMgr("vg                                       $vg");
    }
}

#############################################################################
#
# Format DB output for SGMgr
# similar to FormatInstanceForSGMgr this also includes now IP, Filesystem and Service
#
# Arguments:
#    $dbsid      = SID
#    $dbtype     = ORA,ADA(=MaxDB),DB2,SYB
#    $maxdbtype  = OLTP or LVC (LVC= liveCache)
#    $vhost = virtual host for the DB
#   (argument list may be incomplete)
#
# If this routine is used to format HANA DB as a "normal" DB:
#  - maxdbtype contains replication information (comma separated fields)
#  - vhost may be a list of HANA hostnames (in case of scaleout)
#
#############################################################################

sub FormatDBForSGMgr {
    LogEnterSub( @_ );

    $dbsid      = shift;
    $dbtype     = shift;
    $maxdbtype  = shift;
    $vhost      = shift;

    if ( $dbsid && $dbtype ) {
        printSGMgr("sgesap/db_global/db_system=$dbsid");
        $dbtype            eq "ORA"
                && printSGMgr("sgesap/db_global/db_vendor=oracle");
        $dbtype            eq "ADA"
            and $maxdbtype eq "OLTP"
            && printSGMgr("sgesap/db_global/db_vendor=maxdb")
               || printSGMgr("sgesap/maxdb_spec/maxdb_userkey=$maxdbuc");
        $dbtype            eq  "DB2"
 #           and ($apiversion & 0x01)
            && printSGMgr("sgesap/db_global/db_vendor=db2");
        $dbtype            eq  "SYB"
 #           and ($apiversion & 0x10)
            && printSGMgr("sgesap/db_global/db_vendor=sybase");
        $dbtype            eq  "HDB"
 #          replication mode is in 3rd element (put into maxdb_type here)
            && (printSGMgr("sgesap/db_global/db_vendor=hana")
               || printSGMgr("sgesap/hdb_global/_hdb_replmode=$maxdbtype")
               || printSGMgr("sgesap/hdb_global/_hdb_nodes=$vhost"));

        # output IP data for SGMgr (not for HANA)
        $dbtype ne "HDB" && FormatIPForSGMgr($vhost);


        # output FileSystem data required by SGMgr
        FormatFileSystemForSGMgr();

        # output of services.
        my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
        FormatServiceForSGMgr($pkgname, "datab");
    }
    LogReturnSub();
}
#############################################################################
#
# Format HDB output for SGMgr
# similar to FormatInstanceForSGMgr this also includes now IP, Filesystem and Service
#
# Arguments:
#    $dbsid      = SID
#    $dbtype     = HDB
#    $vhost = virtual host for the DB
#   (argument list may be incomplete)
#
#    hdbuk is global var!
#
#############################################################################

sub FormatHDBForSGMgr {
    LogEnterSub( @_ );

    $dbsid      = shift;
    $dbtype     = shift;
    $replmode   = shift;
    $hdbhosts   = shift; # not used so far...
    $vhost      = shift;

    if ($dbsid && $dbtype) {
      # this would be the correct output (in the meantime)
      printSGMgr("sgesap/hdb_global/hdb_system:$dbsid|sgesap/hdb_global/hdb_system=$dbsid");
      #printSGMgr("sgesap/hdb_global/hdb_system=$dbsid");
    }

    if ($hdbuk ne "") {
        printSGMgr("sgesap/hdb_global/hdb_userkey=$hdbuk");
    }

    # Print out module name if primary as is there is no other indicatin
    # module_name:sgesap/hdbprimary|module_name=sgesap/hdbprimary
    #if ($replmode eq "primary") {
    #  printSGMgr ("module_name:sgesap/hdbprimary|module_name=sgesap/hdbprimary");
    #}
    #
    # Dumping approach above which leaves replication kind of uncertain if it
    # is not "primary". Instead report mode from ini in a attribute like form.
    # (However, this attribute isn't part of any module spec)
    printSGMgr("sgesap/hdb_global/_hdb_replmode=$replmode");

    if ($replmode eq "primary") {
      # assume sync/syncmem replication (as the attributes belong
      # to the hdbprimary module, these have to be reported here)
      # latest recommendation by ME (HANA logshipping_timeout recommended to 75 > tolerance+moninterval)
      printSGMgr("sgesap/hdbprimary/hdb_sync_time_tolerance=40");
      #printSGMgr("sgesap/hdbprimary/hdb_sync_logpos_tolerance=0");
    }

    FormatHDBInstancesForSGMgr();

    # output IP data for SGMgr - only for primary
    FormatIPForSGMgr($vhost) if ($replmode eq "primary");

    # output FileSystem data required by SGMgr
    FormatFileSystemForSGMgr();

    # output of services.
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
    FormatServiceForSGMgr($pkgname, "hdbsys");

    if (defined ($ENV{"SGeSAP_HDBREP_PKGNAME"})) {
	  # ouput priority of primary/secondary package
      FormatHDBPriorityForSGMgr($pkgprio);
      # output of primary/secondary package dependencies
      FormatHDBDependencyForSGMgr($ENV{"SGeSAP_HDBREP_PKGNAME"});
    }

    # overwrite default SG attribute script_log_file ($SGRUN/log/$SG_PACKAGE.log)
    printSGMgr("script_log_file=\$SGRUN/log/hdb$dbsid.log");

    LogReturnSub();
}

#############################################################################
#
# Format replicated DB output for SGMgr
#
#
# Arguments:
#    $dbsid      = SID
#    $dbtype     = SYB
#    $dbrole     = primary,standby
#    $vhost      = virtual host
#
# Output for SGeSybase toolkit (tkit/sybase/rs) commented.
# Output sgesap/repdbinstance attributes instead (e.g):
#   sgesap/repdbinstance/repdb_system=SYR
#   sgesap/repdbinstance/repdb_role=primary
#
#
# So far it defaults REP directory to REP-15_5 and OCS directory to OCS_15-0
# instead of taking it from SYBASE.sh
#
#############################################################################

sub FormatREPDBForSGMgr {
    LogEnterSub( @_ );

    my $repdbsid  = shift;
    my $repdbtype = shift;
    my $repdbrole = shift;
	my $vhost     = shift;
    my $vport     = shift;

	my $tkprefix = "tkit/sybase/rs";
	my $installhome  = "/sybase/$repdbsid";
	my $instancename = $repdbsid."_".$repdbrole;


	# output for repdbinstance module
	printSGMgr("sgesap/repdbinstance/repdb_system=".$repdbsid);
	printSGMgr("sgesap/repdbinstance/repdb_role=".$repdbrole);

	# output IP data for SGMgr
    FormatIPForSGMgr($vhost) if defined $vhost;

	# output FileSystem data required by SGMgr
    FormatFileSystemForSGMgr();

	# no output for services

    LogReturnSub();
}

sub FormatSYBTKRSForSGMgr {
    LogEnterSub( @_ );

    my $repdbsid  = shift;
    my $repdbtype = shift;
    my $repdbrole = shift;
	my $vhost     = shift;
    my $vport     = shift;

	my $tkprefix = "tkit/sybase/rs";
	my $installhome  = "/sybase/$repdbsid";
	my $instancename = $repdbsid."_".$repdbrole;

	printSGMgr("$tkprefix/INSTALL_HOME=".$installhome);                 # Should be equal to $SYBASE from SYBASE.sh
	printSGMgr("$tkprefix/RS_INSTALL_PATH=".$installhome."/REP-15_5");  # Default, should come from $installhome/SYBASE.sh
	printSGMgr("$tkprefix/OCS_INSTALL_PATH=".$installhome."/OCS-15_0"); # Default, should come from $installhome/SYBASE.sh

    my $admin_name="syb".lc(substr $repdbsid,0,3);
	printSGMgr("$tkprefix/ADMIN_NAME=".$admin_name);
    my $user_name="DR_admin"; # Default - SAP hardcoded
	printSGMgr("$tkprefix/USER_NAME=".$user_name);
	printSGMgr("$tkprefix/LISTENER_PORT=".$vport);

	printSGMgr("$tkprefix/INSTANCE_NAME=".$instancename);
	printSGMgr("$tkprefix/RUNSCRIPT=".$installhome."/".$instancename."/RUN_".$instancename.".sh");
	printSGMgr("$tkprefix/CONSOLE_LOGFILE=".$installhome."/".$instancename."/RUN_".$instancename.".log");

    # password file (by convention)
    printSGMgr("$tkprefix/PASSWD_FILE=\$SCGCONF/scripts/ext/REPSR2_rs.passwd");

    # other defaults (should lookup related ADF?)
    printSGMgr("$tkprefix/RETRY_TIME=3");
    printSGMgr("$tkprefix/TIME_OUT=10");
    printSGMgr("$tkprefix/MONITOR_INTERVAL=30");

	# monitor: should values have full path?
	# $installhome."/REP-15_5/bin/repserver
	# $installhome."/REP-15_5/ASA12/bin/dbsrv12|dbltm
	my $monprocprefix = $tkprefix."/MONITOR_PROCESS";
	printSGMgr($monprocprefix.":repserver|".$monprocprefix."=repserver");
	printSGMgr($monprocprefix.":dbsrv12|".$monprocprefix."=dbsrv12");
	#printSGMgr($monprocprefix.":dbltm|".$monprocprefix."=dbltm");   # dbltm not there directly after startup

    # Collect SCC values if DRA is configured on it
    # - check $installhome/SCC* (skip if not exist)
    # - look for $sccdir/plugins/DR/agent-plugin.xml (skip if not exist)
    my @sccdir=<$installhome/SCC-*>;
    my $pluginfile="plugins/DR/agent-plugin.xml";
    if ((@sccdir == 1) && (-r "$sccdir[0]/$pluginfile")) {
      $tkprefix = "tkit/sybase/dra";

      # get port from XML
      $vport="8899";
      open (FH,"$sccdir[0]/$pluginfile") || warn "Couldn't open $pluginfile";
      while (<FH>) {
        if (/set-property property="com.sybase.ua.plugins.dr.tdsPort" value="(\d+)"/) {
          $vport=$1;
          last;
        }
      }
      close (FH);

      printSGMgr("$tkprefix/INSTALL_HOME=".$installhome);
	  printSGMgr("$tkprefix/RS_INSTALL_PATH=".$installhome."/REP-15_5");
	  printSGMgr("$tkprefix/OCS_INSTALL_PATH=".$installhome."/OCS-15_0");
      printSGMgr("$tkprefix/ADMIN_NAME=".$admin_name);
      printSGMgr("$tkprefix/USER_NAME=".$user_name);
      printSGMgr("$tkprefix/LISTENER_PORT=".$vport);

      #other defaults (should lookup related ADF?)
      printSGMgr("$tkprefix/RETRY_TIME=3");
      printSGMgr("$tkprefix/TIME_OUT=10");
      printSGMgr("$tkprefix/MONITOR_INTERVAL=30");

      printSGMgr("$tkprefix/RUNSCRIPT=".$sccdir[0]."/bin/scc.sh");
	  printSGMgr("$tkprefix/CONSOLE_LOGFILE=".$sccdir[0]."/log/scc-console.out");

      printSGMgr("$tkprefix/PASSWD_FILE=\$SCGCONF/scripts/ext/REPSR2_dra.passwd");

      printSGMgr("$tkprefix/HEALTHCHK_CMD=sap_version");
      printSGMgr("$tkprefix/HEALTHCHK_RESULT=\"DR Agent\"");
      printSGMgr("$tkprefix/UNIQ_MONITOR_LABEL=".$installhome."/SCC");
      $monprocprefix = $tkprefix."/MONITOR_PROCESS";
	  printSGMgr($monprocprefix.":java|".$monprocprefix."=java");
    }

    # output IP data for SGMgr
    FormatIPForSGMgr($vhost) if defined $vhost;

	# output FileSystem data required by SGMgr
    FormatFileSystemForSGMgr();

	# no output for services

    LogReturnSub();
}
#############################################################################
#
# End of
# Output functions for input by ServiceGuard Manager
#
#############################################################################

#############################################################################
#
# Output functions for cmgetconf/cmmakepkg format.
# This was "POSIX format" - as expected by "-t " option by cmmakepkg previously). Therefore these
# routines have a "Posix" postfix.
#
# Form: FormatxxxForPosix
#
# FormatInstancesForPosix
# FormatDBForPosix
# FormatServiceSAPinstForPosix
# FormatLivecacheForPosix
# FormatIPForPosix
# FormatFileSystemForPosix
# FormatSingleFileSystemForPosix
# FormatServiceForPosix
# FormatModlistForPosix
# FormatLineForPosix
#
# No related output functions for MDM and NFS yet.
#
#############################################################################


#
# Defaults used by output for pkg config format (and some also SGMgr format)
# (it is to be discussed if such kind of defaults should be printed out by sapgetconf or if it should be relied
# on cmmakepkg to pick and add the currently valid defaults and their values)
# Values below are also the default provided by the module
#
$SGeSAP_REM_COMM_DEFAULT="remsh" unless defined $SGeSAP_REM_COMM_DEFAULT;
$SGeSAP_CLEANUP_POLICY_DEFAULT="normal" unless defined $SGeSAP_CLEANUP_POLICY_DEFAULT;
$SGeSAP_RETRY_COUNT_DEFAULT="5" unless defined $SGeSAP_RETRY_COUNT_DEFAULT;
$SG_SERVICE_RESTART_DEFAULT="none" unless defined $SG_SERVICE_RESTART_DEFAULT;
$SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT="no" unless defined $SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT;
$SG_SERVICE_HALT_TIMEOUT_DEAULT="5" unless defined $SG_SERVICE_HALT_TIMEOUT_DEAULT;

#
# datastructure (hash of arrays) to be used for "left hand side" of attribute output and indication to which module it belongs
# (cmgetpkgenv format - this hash isn't used yet)
#
%HoAttr_cmgetpkgenv = (
  #  sg/basic
  "SG_NODE_NAME"       => [ "SG_NODE_NAME", "sg/basic" ],
  "SG_SCRIPT_LOG_FILE" => [ "SG_SCRIPT_LOG_FILE", "sg/basic" ],
  #  sg/resource.
  "SG_RESOURCE_NAME" =>              [ "SG_RESOURCE_NAME", "sg/resource" ],
  "SG_RESOURCE_START" =>             [ "SG_RESOURCE_START", "sg/resource" ],
  "SG_RESOURCE_POLLING_INTERVAL" =>  [ "SG_RESOURCE_POLLING_INTERVAL", "sg/resource" ],
  "SG_RESOURCE_UP_VALUE" =>          [ "SG_RESOURCE_UP_VALUE", "sg/resource" ],
  #  sg/generic_resource
  "SG_GENRES_NAME" =>        [ "SG_GENERIC_RESOURCE_NAME", "sg/generic_resource" ],
  "SG_GENRES_EVAL_TYPE" =>   [ "SG_GENERIC_RESOURCE_EVALUATION_TYPE", "sg/generic_resource" ],
  "SG_RESOURCE_UP_CRIT" =>   [ "SG_GENERIC_RESOURCE_UP_CRITERIA", "sg/generic_resource" ],
  # sg/dependency:
  "SG_DEP_NAME" => [ "SG_DEPENDENCY_NAME", "sg/dependency" ],
  "SG_DEP_LOC"  => [ "SG_DEPENDENCY_LOCATION", "sg/dependency" ],
  "SG_DEP_COND" => [ "SG_DEPENDENCY_CONDITION", "sg/dependency" ],
  # sg/package_ip:
  "SG_IP_SUBNET" =>   [ "SG_IP_SUBNET", "sg/package_ip" ],
  "SG_IP_ADDRESS" =>  [ "SG_IP_ADDRESS", "sg/package_ip" ],
  # sg/priority:
  "SG_PRIORITY" =>   [ "SG_PRIORITY", "sg/priority" ],
  # sg/volume_group:
  "SG_VG" =>       [ "SG_VG", "sg/volume_group" ],
  "SG_VXVM_DG" =>  [ "SG_VXVM_DG", "sg/volume_group" ],
  "SG_CMV_DG" =>   [ "SG_CVM_DG", "sg/volume_group" ],
  # sg/filesystem
  "SG_FS_NAME" =>       [ "SG_FS_NAME", "sg/filesystem" ],
  "SG_FS_DIRECTORY" =>  [ "SG_FS_DIRECTORY", "sg/filesystem" ],
  "SG_FS_TYPE" =>       [ "SG_FS_TYPE", "sg/filesystem" ],
  "SG_FS_SERVER" =>     [ "SG_FS_SERVER", "sg/filesystem" ],
  # sg/service
  "SG_SERVICE_NAME" =>               [ "SG_SERVICE_NAME", "sg/service" ],
  "SG_SERVICE_CMD" =>                [ "SG_SERVICE_CMD", "sg/service" ],
  "SG_SERVICE_RESTART" =>            [ "SG_SERVICE_RESTART", "sg/service" ],
  "SG_SERVICE_FAIL_FAST_ENABLED" =>  [ "SG_SERVICE_FAIL_FAST_ENABLED", "sg/service" ],
  "SG_SERVICE_HALT_TIMEOUT" =>       [ "SG_SERVICE_HALT_TIMEOUT", "sg/service" ],
  # sgesap/sapinstance:
  "SGeSAP_SAP_SYSTEM" =>               [ "SGeSAP_SAP_SYSTEM", "sgesap/sapinstance" ],
  "SGeSAP_SAP_INSTANCE" =>             [ "SGeSAP_SAP_INSTANCE", "sgesap/sapinstance" ],
  "SGeSAP_SAP_VIRTUAL_HOSTNAME" =>     [ "SGeSAP_SAP_VIRTUAL_HOSTNAME", "sgesap/sapinstance" ],
  "SGeSAP_SAP_REPLICATED_INSTANCE" =>  [ "SGeSAP_SAP_REPLICATED_INSTANCE", "sgesap/sapinstance" ],
  "SGeSAP_REM_COMM" =>                 [ "SGeSAP_REM_COMM", "sgesap/sapinstance" ],
  "SGeSAP_CLEANUP_POLICY" =>           [ "SGeSAP_CLEANUP_POLICY", "sgesap/sapinstance" ],
  "SGeSAP_RETRY_COUNT" =>              [ "SGeSAP_RETRY_COUNT", "sgesap/sapinstance" ],
  # sgesap/db
  "SGeSAP_DB_VENDOR" =>      [ "SGeSAP_DB_VENDOR", "sgesap/db" ],
  "SGeSAP_DB_SYSTEM" =>      [ "SGeSAP_DB_SYSTEM", "sgesap/db" ],
  "SGeSAP_MAXDB_USERKEY" =>  [ "SGeSAP_MAXDB_USERKEY", "sgesap/db" ],
  # sgesap/livecache (all indexed, see example)
  "SGeSAP_LC_SYSTEM" =>                 [ "SGeSAP_LC_SYSTEM", "sgesap/livecache" ], # SGeSAP_LC_SYSTEM[0]="LC5"
  "SGeSAP_LC_VIRTUAL_HOSTNAME" =>       [ "SGeSAP_LC_VIRTUAL_HOSTNAME", "sgesap/livecache" ], #SGeSAP_LC_VIRTUAL_HOSTNAME[0]=""
  "SGeSAP_LC_STARTMODE" =>              [ "SGeSAP_LC_STARTMODE", "sgesap/livecache" ], # SGeSAP_LC_STARTMODE[0]="online"
  "SGeSAP_LC_USERKEY" =>                [ "SGeSAP_LC_USERKEY", "sgesap/livecache" ], #SGeSAP_LC_USERKEY[0]="c"
  "SGeSAP_LC_STANDBY_COPY_MECHANISM" => [ "SGeSAP_LC_STANDBY_COPY_MECHANISM", "sgesap/livecache" ], # SGeSAP_LC_STANDBY_COPY_MECHANISM[0]="businesscopy"
  "SGeSAP_LC_STANDBY_RESTART" =>        [ "SGeSAP_LC_STANDBY_RESTART", "sgesap/livecache" ], # SGeSAP_LC_STANDBY_RESTART[0]="no"
  "SGeSAP_LC_STANDBY_AVOID_IDCHECK" =>  [ "SGeSAP_LC_STANDBY_AVOID_IDCHECK", "sgesap/livecache" ], #SGeSAP_LC_STANDBY_AVOID_IDCHECK[0]="yes"


  # end of hash marker, don't add new entries below
  "HoAttrEnd" =>  [ "HoAttrEnd", "---/---" ]
  );

#
# same as above, but this time cmmakepkg format
# (used indirectly in the FormatXXXForPosix)
# %HoAttr_cmmaklepkg = (
%HoAttr = (
  # sg/basic:
  "SG_PACKAGE_NAME" =>         [ "package_name", "sg/basic" ],
  "SG_PACKAGE_DESCRIPTION" =>  [ "package_description", "sg/basic" ],
  "SG_MODULE_NAME" =>          [ "module_name", "sg/basic" ],
  "SG_NODE_NAME" =>            [ "node_name", "sg/basic"],
  "SG_SCRIPT_LOG_FILE" =>      [ "script_log_file", "sg/basic" ],
  #  sg/resource.
  "SG_RESOURCE_NAME" =>              [ "resource_name", "sg/resource" ],
  "SG_RESOURCE_START" =>             [ "resource_start", "sg/resource" ],
  "SG_RESOURCE_POLLING_INTERVAL" =>  [ "resource_polling_interval", "sg/resource" ],
  "SG_RESOURCE_UP_VALUE" =>          [ "resource_up_value", "sg/resource" ],
  #  sg/generic_resource
  "SG_GENRES_PREFIX" =>      [ "^generic_resource_",""],
  "SG_GENRES_NAME" =>        [ "generic_resource_name", "sg/generic_resource" ],
  "SG_GENRES_EVAL_TYPE" =>   [ "generic_resource_evaluation_type", "sg/generic_resource" ],
  "SG_RESOURCE_UP_CRIT" =>   [ "generic_resource_up_criteria", "sg/generic_resource" ],
    # sg/dependency:
  "SG_DEP_NAME" => [ "dependency_name", "sg/dependency" ],
  "SG_DEP_LOC"  => [ "dependency_location", "sg/dependency" ],
  "SG_DEP_COND" => [ "dependency_condition", "sg/dependency" ],
  # sg/package_ip:
  "SG_IP_SUBNET" =>   [ "ip_subnet", "sg/package_ip" ],
  "SG_IP_ADDRESS" =>  [ "ip_address", "sg/package_ip" ],
  # sg/priority:
  "SG_PRIORITY" =>   [ "priority", "sg/priority" ],
  # sg/volume_group:
  "SG_VG" =>       [ "vg", "sg/volume_group" ],
  "SG_VXVM_DG" =>  [ "vxvm_dg", "sg/volume_group" ],
  "SG_CMV_DG" =>   [ "cvm_dg", "sg/volume_group" ],
  # sg/filesystem
  "SG_FS_NAME" =>       [ "fs_name", "sg/filesystem" ],
  "SG_FS_DIRECTORY" =>  [ "fs_directory", "sg/filesystem" ],
  "SG_FS_TYPE" =>       [ "fs_type", "sg/filesystem" ],
  "SG_FS_SERVER" =>     [ "fs_server", "sg/filesystem" ],
  # sg/service
  "SG_SERVICE_NAME" =>               [ "service_name", "sg/service" ],
  "SG_SERVICE_CMD" =>                [ "service_cmd", "sg/service" ],
  "SG_SERVICE_RESTART" =>            [ "service_restart", "sg/service" ],
  "SG_SERVICE_FAIL_FAST_ENABLED" =>  [ "service_fail_fast_enabled", "sg/service" ],
  "SG_SERVICE_HALT_TIMEOUT" =>       [ "service_halt_timeout", "sg/service" ],
  # sg/vmfs (Linux only, no SGMgr output either)
  "SG_VMDK_FILE_NAME"  =>  [ "vmdk_file_name", "sg/vmfs" ],
  "SG_DATASTORE_NAME"  =>  [ "datastore_name", "sg/vmfs" ],
  "SG_SCSI_CONTROLLER" => [ "scsi_controller", "sg/vmfs" ],
  "SG_DISK_TYPE"       => [ "disk_type", "sg/vmfs" ],
  # sgesap/sap_global (indirectly included via sapinstance, has no direct link):
  "SGeSAP_SAP_SYSTEM" =>               [ "sgesap/sap_global/sap_system", "sgesap/sapinstance" ],
  "SGeSAP_REM_COMM" =>                 [ "sgesap/sap_global/rem_comm", "sgesap/sapinstance" ],
  "SGeSAP_CLEANUP_POLICY" =>           [ "sgesap/sap_global/cleanup_policy", "sgesap/sapinstance" ],
  "SGeSAP_RETRY_COUNT" =>              [ "sgesap/sap_global/sgesap_retry_count", "sgesap/sapinstance" ],
  # sgesap/sapinstance:
  "SGeSAP_SAP_INSTANCE" =>             [ "sgesap/stack/sap_instance", "sgesap/sapinstance" ],
  "SGeSAP_SAP_VIRTUAL_HOSTNAME" =>     [ "sgesap/stack/sap_virtual_hostname", "sgesap/sapinstance" ],
  "SGeSAP_SAP_REPLICATED_INSTANCE" =>  [ "sgesap/stack/sap_replicated_instance", "sgesap/sapinstance" ],

  # sgesap/db_global|dbinstance
  "SGeSAP_DB_VENDOR" =>      [ "sgesap/db_global/db_vendor", "sgesap/dbinstance" ],
  "SGeSAP_DB_SYSTEM" =>      [ "sgesap/db_global/db_system", "sgesap/dbinstance" ],
  "SGeSAP_MAXDB_USERKEY" =>  [ "sgesap/maxdb_spec/maxdb_userkey", "sgesap/dbinstance" ],
  "SGeSAP_LISTENER_NAME" =>  [ "sgesap/oracle_spec/listener_name", "sgesap/dbinstance" ],
  "SGeSAP_LISTENER_PASSWORD" =>  [ "sgesap/oracle_spec/listener_password", "sgesap/dbinstance" ],
  # the latter two Oracle attributes aren't used so far by sapgetconf

  # sgesap/hdb_global|hdbinstance|hdbprimary
  #"SGeSAP_HDB_VENDOR" =>      [ "sgesap/hdb_global/hdb_vendor", "sgesap/hdbsystem" ],
  "SGeSAP_HDB_SYSTEM"   =>    [ "sgesap/hdb_global/hdb_system", "sgesap/hdbinstance" ], # actually sgesap/hdb_global.x
  "SGeSAP_HDB_USERKEY"  =>    [ "sgesap/hdb_global/hdb_userkey", "sgesap/hdbinstance" ], # see above
#  "SGeSAP_HDB_INSTANCE" =>    [ "sgesap/hdbinstance/hdb_instance", "sgesap/hdbinstance" ],
  "SGeSAP_HDB_INSTANCE" =>    [ "sgesap/hdbinstance_global/hdb_instance", "sgesap/hdbinstance" ],  # >= 06.00.40
# STO (selftuning timeouts) - all three
  "SGeSAP_HDB_STO_SAMPLE_SIZE"       => [ "sgesap/hdb_global/hdb_sto_sample_size", "sgesap/hdbinstance" ], # actually sgesap/hdb_global.x
  "SGeSAP_HDB_STO_SAFETY_FACTOR"     => [ "sgesap/hdb_global/hdb_sto_safety_factor", "sgesap/hdbinstance" ], # see above
  "SGeSAP_HDB_STO_MINIMUM_THRESHOLD" => [ "sgesap/hdb_global/hdb_sto_minimum_threshold", "sgesap/hdbinstance" ], # see above
  # sgesap/hdbdistinstance "DIST" variant-- same as above
  "SGeSAP_HDB_SYSTEM_DIST"   =>    [ "sgesap/hdb_global/hdb_system", "sgesap/hdbdistinstance" ], # actually sgesap/hdb_global.x
  "SGeSAP_HDB_USERKEY_DIST"  =>    [ "sgesap/hdb_global/hdb_userkey", "sgesap/hdbdistinstance" ], # see above
  "SGeSAP_HDB_INSTANCE_DIST" =>    [ "sgesap/hdbinstance_global/hdb_instance", "sgesap/hdbdistinstance" ],  # >= 06.00.40
# STO (selftuning timeouts) - all three
  "SGeSAP_HDB_STO_SAMPLE_SIZE_DIST"       => [ "sgesap/hdb_global/hdb_sto_sample_size", "sgesap/hdbdistinstance" ], # actually sgesap/hdb_global.x
  "SGeSAP_HDB_STO_SAFETY_FACTOR_DIST"     => [ "sgesap/hdb_global/hdb_sto_safety_factor", "sgesap/hdbdistinstance" ], # see above
  "SGeSAP_HDB_STO_MINIMUM_THRESHOLD_DIST" => [ "sgesap/hdb_global/hdb_sto_minimum_threshold", "sgesap/hdbdistinstance" ], # see above
# sync tolerances
  "SGeSAP_HDB_SYNC_TIME_TOLERANCE"    =>  [ "sgesap/hdbprimary/hdb_sync_time_tolerance", "sgesap/hdbprimary" ],
  "SGeSAP_HDB_SYNC_LOGPOS_TOLERANCE"  =>  [ "sgesap/hdbprimary/hdb_sync_logpos_tolerance","sgesap/hdbprimary" ],
  "SGeSAP_HDB_INIT_INTO_FULLSYNC"     =>  [ "sgesap/hdbprimary/hdb_init_into_fullsync", "sgesap/hdbprimary" ],
# HDB dist instance primaary IP addresses
  "HDBDIST_IP_SUBNET"  => [ "sgesap/hdbdistinstance/hdb_ip_subnet", "sgesap/hdbdistinstance" ],
  "HDBDIST_IP_ADDRESS" => [ "sgesap/hdbdistinstance/hdb_ip_address", "sgesap/hdbdistinstance" ],

  # sgesap/livecache
  "SGeSAP_LC_SYSTEM" =>                 [ "sgesap/livecache/lc_system", "sgesap/livecache" ],
  "SGeSAP_LC_VIRTUAL_HOSTNAME" =>       [ "sgesap/livecache/lc_virtual_hostname", "sgesap/livecache" ],
  "SGeSAP_LC_STARTMODE" =>              [ "sgesap/livecache/lc_startmode", "sgesap/livecache" ],
  "SGeSAP_LC_USERKEY" =>                [ "sgesap/livecache/lc_userkey", "sgesap/livecache" ],
  "SGeSAP_LC_STANDBY_COPY_MECHANISM" => [ "sgesap/livecache/lc_standby_copy_mechanism", "sgesap/livecache" ],
  "SGeSAP_LC_STANDBY_RESTART" =>        [ "sgesap/livecache/lc_standby_restart", "sgesap/livecache" ],
  "SGeSAP_LC_STANDBY_AVOID_IDCHECK" =>  [ "sgesap/livecache/lc_standby_avoid_idcheck", "sgesap/livecache" ],
   # sgesap/repdbinstance
  "SGeSAP_REPDB_SYSTEM" =>      [ "sgesap/repdbinstance/repdb_system", "sgesap/repdbinstance" ],
  "SGeSAP_REPDB_ROLE" =>        [ "sgesap/repdbinstance/repdb_role", "sgesap/repdbinstance" ],

   # nfs/hanfs,nfs/hanfs_export,nfs/hanfs_flm
   # so far only the XFS attribute is used (part of nfs/hanfs_export)
   # (be aware that using hanfs_export automatically defines the service nfs.monitor or nfs_monitor on LX)
   "NFS_XFS" => ["nfs/hanfs_export/XFS", "nfs/hanfs"],
   # on LX: "NFS_XFS" => ["tkit/nfs/nfs/XFS", "tkit/nfs/nfs"],
   "NFS_XFS_LX" => ["tkit/nfs/nfs/XFS", "tkit/nfs/nfs"],

   # Sybase toolkit (rs only, but still quite lengthy stuff)
   "RS_INSTALL_HOME"     =>  ["tkit/sybase/rs/INSTALL_HOME", "tkit/sybase/rs"],
   "RS_RS_INSTALL_PATH"   =>  ["tkit/sybase/rs/RS_INSTALL_PATH", "tkit/sybase/rs"],
   "RS_OCS_INSTALL_PATH"  =>  ["tkit/sybase/rs/OCS_INSTALL_PATH", "tkit/sybase/rs"],
   "RS_CONSOLE_LOGFILE"  =>  ["tkit/sybase/rs/CONSOLE_LOGFILE", "tkit/sybase/rs"],
   "RS_ADMIN_NAME"       =>  ["tkit/sybase/rs/ADMIN_NAME", "tkit/sybase/rs"],
   "RS_USER_NAME"        =>  ["tkit/sybase/rs/USER_NAME", "tkit/sybase/rs"],
   "RS_LISTENER_PORT"    =>  ["tkit/sybase/rs/LISTENER_PORT", "tkit/sybase/rs"],
   "RS_INSTANCE_NAME"    =>  ["tkit/sybase/rs/INSTANCE_NAME", "tkit/sybase/rs"],
   "RS_RUNSCRIPT"        =>  ["tkit/sybase/rs/RUNSCRIPT", "tkit/sybase/rs"],
   "RS_MONITOR_PROCESS"  =>  ["tkit/sybase/rs/MONITOR_PROCESS", "tkit/sybase/rs"],
   # not included: TKIT_DIR, MONITOR_INTERVAL, TIME_OUT, RETRY_TIME,
   #   SHUTDOWN_IN_FILE, MONITOR_IN_FILE, MAINTENANCE_FLAG

   # for writing out a comment (temp. use only)
   "\#\#" => [ "\#", "" ],

  # end of hash marker, don't add new entries below
  "HoAttrEnd" =>  [ "HoAttrEnd", "---/---" ]
  );


# List of attributes (and included modules) can be otained with
#  cmmakepkg -l -f line -m <module> | grep attribute | awk 'FS="|" { print $1" "$2} ' | uniq

# WVE:   FormatHDBDistInstancesForPosix necessary
# parameter: number of nodes (1=scaleup , >1 scaleout)
sub FormatHDBInstancesForPosix {
    LogEnterSub( @_ );
    #my @hdbinsts=@_; # not used
    my $numnodes = shift;
    my $modprefix="SGeSAP_";
    my $instindex=0;
    my %insthash = ();
    my $attrpostfix="";
    $attrpostfix = "_DIST" if ($numnodes>1);

    # usually this is just one instance name. However, on scaleout it may be configured
    # to more than one node. Make sure it appears only once in output

    printSGMgr("# -- HDB instance related variables --");

    # loop thru all available SAP profiles
    foreach $prof (@profiles) {
      # Get the profile filename from the absolute path
      my $profname;
      $profname = ( split( /\//, $prof ) )[-1];

      # get instance name and (virtual) hostname out of the profile file name.
      # A SAP profile has the following format <SID>_<INST>_<HOST>
      my ( $sid, $inst, $vhost ) = split( /_/, $profname );

      # enforce SAP instance naming rules
      next unless ( $inst =~ /^(HDB)[0-9][0-9]$/ );
      # skip if already seen
      if (defined $insthash{$inst}) {
      	$instcounter++; # assumes that the same instance is counted
      	next;
      }

      FormatLineForPosix ($modprefix."HDB_INSTANCE","$inst",$instindex);
      $pkgdescr=$pkgdescr." $inst";
      $EDmodlist{$HoAttr{$modprefix."HDB_INSTANCE".$attrpostfix}[1]}=1;

      $insthash{$inst} = 1;

      $instindex++;
    }    # end foreach profile

} # FormatHDBInstanceForPosix


sub FormatInstancesForPosix {
    LogEnterSub( @_ );
    my @sapinsts=@_;
    my $sapglobout=1;
    my $modprefix="SGeSAP_";
    my $instindex=0;

    # maps for later services output, global (sorry)
    %sapinsttypelist=(); # map of instances types (w/o instance number) for which properties are printed out
    $sapfoundinsttylist=(); # map of found instance types while going through the profiles

    # collect instance directories along the way...
    my  @instdirs = (); #

    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");

    printSGMgr("# -- SAP instance related variables --");

    # loop thru all available SAP profiles
    foreach $prof (@profiles) {

      #---------------------------------------------------------------------------
      #   SGMgr output $inst:      ...D03
      #   SGMgr output $vhost:     ...reloc123
      #---------------------------------------------------------------------------

      # Get the profile filename from the absolute path
      my $profname;
      $profname = ( split( /\//, $prof ) )[4];

      # get instance name and (virtual) hostname out of the profile file name.
      # A SAP profile has the following format <SID>_<INST>_<HOST>
      my ( $sid, $inst, $vhost ) = split( /_/, $profname );

      # enforce SAP instance naming rules
      next
          unless ( $inst
           =~ /^(DVEBMGS|D|ASCS|SCS|ERS|REP|J|JD|JC|MDS|MDIS|MDSS|W|G)[0-9][0-9]$/
          );

      $sapfoundinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      # skip if not in list of selected instances (if such a list is defined)
      if (scalar @_ != 0)
      {
           # restricted to list of instances. Check if instance is in list. Partial specifications allowed
           next unless grep {$inst =~ $_} @sapinsts;
      }

      $sapinsttypelist{substr($inst, 0, length($inst)-2)}=1;

      # First check if we still need to output sap_global attribute
      if ($sapglobout) {
        $sapglobout=0;
        FormatLineForPosix ($modprefix."SAP_SYSTEM","$sid");
        # do not print out (assumed) defaults
        #FormatLineForPosix ($modprefix."REM_COMM","$SGeSAP_REM_COMM_DEFAULT");
        #FormatLineForPosix ($modprefix."CLEANUP_POLICY","$SGeSAP_CLEANUP_POLICY_DEFAULT");
        #FormatLineForPosix ($modprefix."RETRY_COUNT","$SGeSAP_RETRY_COUNT_DEFAULT");

        $EDmodlist{$HoAttr{$modprefix."SAP_SYSTEM"}[1]}=1; #push on modlist (module list)
      }


      FormatLineForPosix ($modprefix."SAP_INSTANCE","$inst",$instindex);
      FormatLineForPosix ($modprefix."SAP_VIRTUAL_HOSTNAME","$vhost",$instindex);
      $pkgdescr=$pkgdescr." $inst";
      $EDmodlist{$HoAttr{$modprefix."SAP_INSTANCE"}[1]}=1;


      $instindex++;

      #---------------------------------------------------------------------------
      #  If the instance is a ERS instance output the matching SCS instance
      #  SGMgr output $scsinst:      ...SCS03
      #---------------------------------------------------------------------------
      if ( $inst =~ /^ERS/ ) {
          my $scsinst;
          $scsinst = $ERSmap{$inst};
          LogDebug("inst=[", $inst, "], scsinst=[", $scsinst, "]");
          if ( $scsinst ne "" ) {
              FormatLineForPosix ($modprefix."SAP_REPLICATED_INSTANCE","$scsinst");
          }
          else {
              LogError("No [A]SCS configured for ", $inst, "");
          }
      }    # end if ERS instance

      FormatResourceForPosix($SID,$inst); # output only for [A]SCS,ERS
      # store virtual hostname is associative array to avoid duplicates & for later outpuit
      $vhosts{$vhost}=1 unless defined $vhosts{$vhost} || $vhosts{$vhost} != 0;

      # search filesytems
      my $instdir = "/usr/sap/$SID/$inst";

      LogVerbose ("\$instdir=[$instdir]");


      # Collect instdirs instead of immediate output. No check here
      # if it is actully mounted.
      #my $instfs  = $mpsfs{$instdir};
      #my $instfst = $mpsfst{$instdir};
      #if ( $instfs ne "" ) {
        #FormatSingleFileSystemForPosix ($instdir, $instfs, $instfst);
      #}
      push @instdirs,$instdir;

      # check if vhost has an "asscociated" SMDA instance
      # needs only to be done for instances != ASCS,SCS;ERS
      # and if vhost wasn't already checked

      if ( ($inst !~ /^ASCS|^SCS|^ERS/ ) && !defined $SMDAVHmap{vhost} ) {
        LogVerbose ("Need to look up SMDA for $vhost");
        $SMDAVHmap{$vhost}=1;
        # FormatSMDAForPosix ($vhost);
      }
    }    # end foreach profile

    # filesystems
	# avoid pushing empty list to FormatFileSystems
	$instdirs[0] = "/usr/sap/$sid" unless @instdirs;
    FormatFileSystemForPosix (@instdirs);
    # Dependency (enqor)
    #TODO $inst not defined any more - only in loop above
    FormatDependencyForPosix($SID,$inst);

    # IPs in one chunck at the end
    FormatIPForPosix (%vhosts) ;
    # services in one chunk at the end
    FormatServiceSAPinstForPosix();
    # SMDA
    FormatSMDAForPosix(); # works on SMDAVHmap

}   # FormatInstancesForPosix

sub FormatDBForPosix {
    LogEnterSub( @_ );
    ($dbsid,$dbtype,$maxdbtype,$vhost) = @_;

    my $modprefix="SGeSAP_";
    my $dbvendprefix=$modprefix."DB_VENDOR";

    if ( $dbsid && $dbtype ) {
      printSGMgr("# -- DB related variables -- ");
      FormatLineForPosix ($modprefix."DB_SYSTEM","$dbsid");
      FormatLineForPosix ("$dbvendprefix","oracle") if ($dbtype eq "ORA");
      FormatLineForPosix ("$dbvendprefix","db2") if ($dbtype eq "DB2");
      FormatLineForPosix ("$dbvendprefix","sybase") if ($dbtype eq "SYB");
      if ($dbtype eq "ADA" and $maxdbtype eq "OLTP") {
        FormatLineForPosix ("$dbvendprefix","maxdb");
        FormatLineForPosix ($modprefix."MAXDB_USERKEY","$maxdbuc");
        $EDmodlist{$HoAttr{$modprefix."MAXDB_USERKEY"}[1]}=1;
      }
      $EDmodlist{$HoAttr{$modprefix."DB_SYSTEM"}[1]}=1;
    }
    else {
      return;
    }

    # output IP data for Posix (if vhost found)
    if ($vhost ne "") {
      $vhosts{$vhost}=1 unless defined $vhosts{$vhost} || $vhosts{$vhost} != 0;
    }
    FormatIPForPosix(%vhosts);

    # output FileSystem data for Posix
    FormatFileSystemForPosix();

    # output of services. As there's only one service to be set, no extra format routine is used
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
    FormatServiceForPosix($pkgname, "datab",$svcindex++);
    LogReturnSub();
}

sub FormatHDBForPosix {
    LogEnterSub( @_ );
    my ($dbsid,$dbtype,$replmode,$hdbhlist,$vhost,$Hh2Cn_ref) = @_;
    my @hdbhosts = split(",",$hdbhlist);
    my $attrpostfix = ""; # for DIST variant, only relevant for module list

    #my @printarr=%$Hh2Cn_ref;
    #LogVerbose("xx: @printarr");


    # site control latch generic resource for SO
    my $scgenres_name = "sitecontroller_genres";
    my $scgenres_start = "before_package_start";
    my $scgenres_cond_prim = ">3";
    my $scgenres_cond = "==3"; # secondary condition as default
    # scale-up with smart quorum (for both type of pkgs)
    my $scgenres_start_su = "during_package_start";
    my $scgenres_cond_su = ">1";

    # simple resource for safesync (+other start condition for scaleout)
    my $safesync_genres_name = "safesync${dbsid}_genres";
    my $safesync_genres_start_so = "before_package_start";

    my $modprefix="SGeSAP_";
    my $dbvendprefix=$modprefix."HDB_VENDOR";

    $attrpostfix = "_DIST" if (scalar @hdbhosts > 1);

    if ( $dbsid && $dbtype ) {
      printSGMgr("# -- HDB related variables -- ");
      FormatLineForPosix ($modprefix."HDB_SYSTEM","$dbsid");

      # STO variables. providing defauls, not discovered.
      FormatLineForPosix ($modprefix."HDB_STO_SAMPLE_SIZE",20);
      FormatLineForPosix ($modprefix."HDB_STO_SAFETY_FACTOR",5);
      FormatLineForPosix ($modprefix."HDB_STO_MINIMUM_THRESHOLD",10);
      $EDmodlist{$HoAttr{$modprefix."HDB_SYSTEM".$attrpostfix}[1]}=1;
    }
    else {
      return;
    }

    # output for replication module and userkey
    if ($hdbuk ne "") {
        FormatLineForPosix ($modprefix."HDB_USERKEY","$hdbuk");
        # related module already checked
    }

    FormatHDBInstancesForPosix(scalar @hdbhosts);


    # add primary module if necessary
    # will result in: module_name sgesap/hdbprimary
    # WVE: scaleout?
    if ($replmode =~ /^primary/) {
      # replmode=<mode>,<sitename>,<siteid>,<fullsync>
      # SYNC attributes. directly add hdbprimary module to list

      printSGMgr("# -- HDB primary related variables -- ");
      # time tolerance only if fullsync != true
      if (( split(',', $replmode ))[3] ne "true") {
        FormatLineForPosix ($modprefix."HDB_SYNC_TIME_TOLERANCE",40);
      }
      else {
        FormatLineForPosix ($modprefix."HDB_INIT_INTO_FULLSYNC","yes");
      }
      #FormatLineForPosix ($modprefix."HDB_SYNC_LOGPOS_TOLERANCE",0);
      $EDmodlist{"sgesap/hdbprimary"}=1;

      $scgenres_cond = $scgenres_cond_prim;
    }

    # setup vhosts hash for primary and secondary. new: with HANA 2.0 the secondary
    # can be used for read-only access of HANA thus it gets it's own IP range
    # WVE: as there's just the convention so far (hdb<SID>) there can be just one
    #      vhost name for the primary
    if ($vhost ne "") {
      $vhosts{$vhost}=1 unless defined $vhosts{$vhost} || $vhosts{$vhost} != 0;
    }
    #
    # dist IP range created HEALTHCHK_RESULT (should also come before node_name)
    FormatHDBdistIPForPosix(scalar @hdbhosts) if  (scalar @hdbhosts > 1);

    printSGMgr("# ---- ");

    # explicit output of found hana nodesnames (array hdbhosts) for SO (more than 1 element)
    # (WVE: extra sub)
    if (scalar @hdbhosts > 1) {
    	foreach my $hdbhost (@hdbhosts) {
        use 5.010;
    		FormatLineForPosix ("SG_NODE_NAME",defined $Hh2Cn_ref->{"$hdbhost"} ? $Hh2Cn_ref->{"$hdbhost"}:"$hdbhost",0);  # counter not set so far
    	}
    }

    # output IP data for Posix (ok at this place as vhosts in only filled when primary)
    if (scalar @hdbhosts == 1) {
    	FormatIPForPosix(%vhosts);
    }
    #else {
    #	FormatHDBdistIPForPosix(scalar @hdbhosts);
    #}

    FormatFileSystemForPosix(); # WVE: could be skipped for HANA

    # output of services. As there's only one service to be set, no extra format routine is used
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
    FormatServiceForPosix($pkgname, "hdbsys",$svcindex++);

    # output dependency data for a HDB instance (scaleout only).
    # WVE SO: instead of checking the number of hosts here, the caller could unset
    #         env.var. before calling this script in the SO case.
    # Note: this function is used by deploysappkgs
    # and requires envrionment variable SGeSAP_HDBREP_PKGNAME to be set.
    # Otherwise no output will be triggered. Same applies for priority
    if ((defined ($ENV{"SGeSAP_HDBREP_PKGNAME"})) && (scalar @hdbhosts <= 1)) {
      FormatHDBDependencyForPosix($ENV{"SGeSAP_HDBREP_PKGNAME"}) ;
      FormatHDBPriorityForPosix ($pkgprio);
    }

    # site controller latch generic resource (for SO)
    # WVE: FormatGenResourceForPosix is specifically programmed for SGeSAP enqor resource
    if (scalar @hdbhosts > 1) {
    	FormatLineForPosix ("SG_GENRES_NAME","$scgenres_name",0);
    	FormatLineForPosix ("SG_GENRES_EVAL_TYPE","$scgenres_start",0);
    	FormatLineForPosix ("SG_RESOURCE_UP_CRIT","$scgenres_cond",0);
    	$EDmodlist{$HoAttr{"SG_GENRES_NAME"}[1]}=1;

      # scaleout using safesync: get the safesync (simple) generic resource printed
      if ($ENV{"SAPGETCONF_USE_SAFESYNC"} > 0) {
        FormatLineForPosix ("SG_GENRES_NAME","$safesync_genres_name",0);
        FormatLineForPosix ("SG_GENRES_EVAL_TYPE","$safesync_genres_start_so",0);
      }
    }
    else {
      if (getcluster_smartquorum()) {
        # Scaleup on cluster with smart quorum, figure out if a package
        # with another SID using sitecontroller_genres already
        # If SAPGETCONF_USE_SMARTQ is set to 2, then related generic resource
        # should be assigned to the current SID, regardless if already assigned
        # to another SID
        my ($scres_sid,$scres_ppkg, $scres_spkg)  = (undef, undef, undef);
		    ($scres_sid,$scres_ppkg, $scres_spkg) = getsid_scresused ($scgenres_name);
		    # print site_controller generic resource if no sid found or sid is identical
		    # or "forced"
        LogVerbose "$scres_sid - $dbsid";
        if (!defined($scres_sid) || ("$scres_sid" eq "$dbsid") || ($ENV{"SAPGETCONF_USE_SMARTQ"} == 2)) {
    	     FormatLineForPosix ("SG_GENRES_NAME","$scgenres_name",0);
    	      FormatLineForPosix ("SG_GENRES_EVAL_TYPE","$scgenres_start_su",0);
    	      FormatLineForPosix ("SG_RESOURCE_UP_CRIT","$scgenres_cond_su",0);
    	      $EDmodlist{$HoAttr{"SG_GENRES_NAME"}[1]}=1;
        }

		    # on forced assignment: if there are already other packages having that resource
		    # assigned already, request removal	from these	(WVE: check if both are defined)
        if (($ENV{"SAPGETCONF_USE_SMARTQ"} == 2) && defined($scres_sid) && ("$scres_sid" ne "$dbsid")) {
          LogVerbose "$scres_sid overwrite to $dbsid";
		      my $grp = 34;
		      foreach my $screspkg ($scres_ppkg,$scres_spkg) {
			      FormatAction("deploysappkgs",$grp,"$screspkg","del","SG_GENRES_NAME","$scgenres_name");
			      FormatAction("deploysappkgs",$grp,"$screspkg","del","SG_GENRES_PREFIX");
			      #FormatAction("deploysappkgs",$grp,"$screspkg","del","SG_GENRES_EVAL_TYPE","$scgenres_start_su");
			      #	FormatAction("deploysappkgs",$grp,"$screspkg","del","SG_RESOURCE_UP_CRIT","$scgenres_cond_su");
		      }
        }
      }
      if ($ENV{"SAPGETCONF_USE_SAFESYNC"} > 0) {
        # scaleup using safesync: get just the simple resource print, leavel EVAL to default
        FormatLineForPosix ("SG_GENRES_NAME","$safesync_genres_name",0);
        $EDmodlist{$HoAttr{"SG_GENRES_NAME"}[1]}=1;
      }
    }
    # overwrite default SG attribute script_log_file ($SGRUN/log/$SG_PACKAGE.log)
    FormatLineForPosix ("SG_SCRIPT_LOG_FILE","\$SGRUN/log/hdb$dbsid.log",0);

    # Add sg/failover to module list if it or sg/*multi_node*
    # isn't on the list yet.
    #(However it might be the case that one one of this modules is indirectly
    # included through others - which isn't checked yet. Most reliable:
    # run a cmmakepkg with the modules collected so far and check if failover or
    # multi_node is in the list of modules)
    # WVE: may be included indirectly through hdbinstance or hdbdistinstance
    if (scalar @hdbhosts > 1) {
    	$EDmodlist{"sg/multi_node"}=1 unless grep {$_ =~ m#sg/(failover|multi_node)$#} (keys %EDmodlist);
    }
    $EDmodlist{"sg/failover"}=1 unless grep {$_ =~ m#sg/(failover|multi_node)$#} (keys %EDmodlist);

    LogReturnSub();
}

sub FormatREPDBForPosix {
# uses sgesap/repdbinstance instead of sybase toolkit
    LogEnterSub( @_ );
    my ($repdbsid,$repdbtype,$repdbrole,$vhost) = @_;

    printSGMgr("# -- DB replication related variables -- ");

	#my $modprefix="RS_";
	my $modprefix="SGeSAP_";

	my $installhome  = "/sybase/$repdbsid";
	my $instancename = $repdbsid."_".$repdbrole;

	#FormatLineForPosix ($modprefix."INSTALL_HOME","$installhome");
	#FormatLineForPosix ($modprefix."RS_INSTALL_PATH",$installhome."/REP-15_5");
	#FormatLineForPosix ($modprefix."OCS_INSTALL_PATH",$installhome."/OCS-15_0");

	#FormatLineForPosix($modprefix."ADMIN_NAME","syb".lc(substr $repdbsid,0,3));
	#FormatLineForPosix($modprefix."USER_NAME","DR_admin"); # Default - get from SCC ?
	#FormatLineForPosix($modprefix."LISTENER_PORT","4904"); # Default - get from interface file or SCC?

	#FormatLineForPosix($modprefix."INSTANCE_NAME",$instancename);
	#FormatLineForPosix($modprefix."RUNSCRIPT",$installhome."/".$instancename."/RUN_".$instancename.".sh");
	#FormatLineForPosix($modprefix."CONSOLE_LOGFILE",$installhome."/".$instancename."/RUN_".$instancename.".log");

	# should monitor values have full path?
	#my $monprocprefix = $modprefix."MONITOR_PROCESS";
	#FormatLineForPosix($monprocprefix, "repserver");
	#FormatLineForPosix($monprocprefix, "dbsrv12");
	#FormatLineForPosix($monprocprefix, "dbltm");

	#$EDmodlist{$HoAttr{$modprefix."INSTALL_HOME"}[1]}=1;

	FormatLineForPosix ("SGeSAP_REPDB_SYSTEM","$repdbsid");
	FormatLineForPosix ("SGeSAP_REPDB_ROLE","$repdbrole");
	$EDmodlist{$HoAttr{"SGeSAP_REPDB_SYSTEM"}[1]}=1;

	$vhosts{$vhost}=1 if defined $vhost; # prepare for FormatIPForPosix
	FormatIPForPosix(%vhosts);
	FormatFileSystemForPosix();

	LogReturnSub();
}

sub FormatServiceSAPinstForPosix {
    LogEnterSub( @_ );

    # print out service setup for instance
    # may depend on other instances (e.g. DVEBMGS will have no sapms.mon if there is a ASCS)
    # Function used to print out instance services at the end of the output  (opposed to SGMgr usage where the service
    # is grouped into the instance output - and SGMgr takes care of the duplicates).

    # get package name. Used to build unique service names. Get it either from the environment or set
    # it to ${SG_PACKAGE} for later replacement
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");

    # special handling for DVEBMGS first
    if ($sapinsttypelist{"DVEBMGS"} == 1) {
        # dispatch monitor
        FormatServiceForPosix($pkgname, "disp", $svcindex++);
        # and msg server monitor depending if DVEBMGS and !ASCS and !SCS in found list
        FormatServiceForPosix($pkgname, "ms", $svcindex++)
           if (($sapfoundinsttypelist{"ASCS"} != 1) && ($sapfoundinsttypelist{"SCS"} != 1));
        # make sure there are no more dispatcher services printed out
        $sapinsttypelist{"DVEBMGS"} = 0;
        $sapinsttypelist{"D"} = 0;
    }

    foreach ( keys %sapinsttypelist) {
      next unless $sapinsttypelist{$_} != 0;
      LogVerbose ("Service for instance type: $_");
      if ($_ eq "D") {
        # dispatch monitor
        FormatServiceForPosix($pkgname, "disp", $svcindex++);
        # and msg server monitor depending if DVEBMGS and !ASCS in found list
        FormatServiceForPosix($pkgname, "ms", $svcindex++)
           if (($_ eq "DVEBMGS") && ($sapfoundinsttypelist{"ASCS"} != 1));
      }
      elsif (($_ eq "ASCS") || ($_ eq "SCS")) {
          # enq monitor
          FormatServiceForPosix($pkgname, "enq", $svcindex++);
          FormatServiceForPosix($pkgname, "ms", $svcindex++);
          # check if there is another (A)SCS in package, set that one to 0 to avoid output
          $sapinsttypelist{"SCS"}  = 0 if ($_ eq "ASCS");
          $sapinsttypelist{"ASCS"} = 0 if ($_ eq "SCS");
      }
      elsif ($_ eq "ERS") {
        # enqr monitor
        FormatServiceForPosix("$pkgname", "enqr", $svcindex++);
        # no need to check of other ERS here as we
      }
      elsif ($_ eq "W") {
        FormatServiceForPosix("$pkgname", "webdisp", $svcindex++);
      }
      elsif ($_ eq "G") {
        FormatServiceForPosix("$pkgname", "gw", $svcindex++);
      }
      else {
        # no output for other type of instances
      }
    }
    LogReturnSub();
}

sub FormatLivecacheForPosix {
    LogEnterSub( @_ );

    my $dbsid     = shift;
    my $sapdbhost = shift;
    my $hssenable = shift;

    return unless $dbsid;

    FormatLineForPosix ("SGeSAP_LC_SYSTEM","$dbsid",0);
    FormatLineForPosix ("SGeSAP_LC_VIRTUAL_HOSTNAME","$sapdbhost",0) if ($sapdbhost);
    # next two hardcoded
    FormatLineForPosix ("SGeSAP_LC_STARTMODE","online",0);
    FormatLineForPosix ("SGeSAP_LC_USERKEY","c",0);
    $EDmodlist{$HoAttr{"SGeSAP_LC_SYSTEM"}[1]}=1;
    # HSS params not supported yet

    $vhosts{$sapdbhost}=1; # prepare for FormatIPForPosix
    FormatIPForPosix();
    FormatFileSystemForPosix();

    # service output for LC
    # Standard formatting routine for service not used, because LC monitors follows a different schema
    my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
    my $svccmd="/sapdb/$dbsid/db/sap/lccluster",
    # not "/opt/cmcluster/sap/SID/saplc.mon"
    my $svcname=$pkgname."lc";
    FormatLineForPosix ("SG_SERVICE_NAME","$svcname",0);
    FormatLineForPosix ("SG_SERVICE_CMD","$svccmd",0);
    FormatLineForPosix ("SG_SERVICE_RESTART","$SG_SERVICE_RESTART_DEFAULT",0);
    FormatLineForPosix ("SG_SERVICE_FAIL_FAST_ENABLED","$SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT",0);
    FormatLineForPosix ("SG_SERVICE_HALT_TIMEOUT","$SG_SERVICE_HALT_TIMEOUT_DEAULT",0);
    $EDmodlist{$HoAttr{"SG_SERVICE_NAME"}[1]}=1;

    LogReturnSub();
}

#############################################################################
#
# sub FormatIPForPosix / FormatHDBdistIPForPosix
#
# Output for ip_address module (FormatIPForPosix) or
# hdbdistinstance module (FormatHDBdistIPForPosix)
#
# Latter routine uses address retrieved from vhost as base address and
# generates the requested number of addresses by incrementing the latest
# byte (IPv4 only so far )
#
# expects global associative array of vhosts
#
#############################################################################

sub FormatIPForPosix {
    LogEnterSub( @_ );

    if ( !%vhosts ) {
        LogReturnSub();
        return;
    }

    # relies on nwrksupp utility - no fallback here
    if (! -x $nwrksupp) {
      warn "$nwrksupp utility not found - can't resolve hostnames";
      LogReturnSub();
      return;
    }

    foreach ( keys %vhosts) {
      next if ($vhosts{$_} == 0); # ignore it
      $vhosts{$_} = 0;
      my $cmd = "$nwrksupp ghbns $_";
      LogVerbose("Execute cmd=[$cmd]");
      my @ipsubn = `$cmd` or warn "cmd=[$cmd], return code [$!]";
      LogVerbose("ipsubn=[@ipsubn]");
      my $ipprot, $addr, $subnet, $interf;
      my %laddr = ();
      for $_ (@ipsubn) {
        ( $ipprot, $addr, $subnet, $interf ) = split( /;/, $_ );
        # duplicate pairs possible - try to filter them out
        # WVE: make sure all IP belonging to a subnet appear underneath
        #      a common ip_subnet attribute
        if ($laddr{$addr} != $subnet)
        {
           $laddr{$addr}=$subnet;
           FormatLineForPosix ("SG_IP_SUBNET","$subnet",$vhostindex);
           FormatLineForPosix ("SG_IP_ADDRESS","$addr",$vhostindex);
           $EDmodlist{$HoAttr{"SG_IP_SUBNET"}[1]}=1;
           $vhostindex++;
        }
      }
    }
    LogReturnSub();
    return;
}

# expects global associative array of vhosts
sub FormatHDBdistIPForPosix {
    LogEnterSub( @_ );
    my $numnodes = shift;

    if ( !%vhosts ) {
        LogReturnSub("<vhost empty>");
        return;
    }

    # relies on nwrksupp utility - no fallback here
    if (! -x $nwrksupp) {
      warn "$nwrksupp utility not found - can't resolve hostnames";
      LogReturnSub("<nwrksupp not found>");
      return;
    }

    my $maxoffset = $numnodes-1;      # set a default

    foreach ( keys %vhosts) {
      next if ($vhosts{$_} == 0); # ignore it
      $vhosts{$_} = 0;
      my $cmd = "$nwrksupp ghbns $_";
      LogVerbose("Execute cmd=[$cmd]");
      my @ipsubn = `$cmd` or warn "cmd=[$cmd], return code [$!]";
      LogVerbose("ipsubn=[@ipsubn]");
      my $ipprot, $addr, $subnet, $interf;
      my %laddr = ();
      for $_ (@ipsubn) {
        ( $ipprot, $addr, $subnet, $interf ) = split( /;/, $_ );
        if ($ipprot eq "IPv4") {
          # duplicate pairs possible - try to filter them out
          if ($laddr{$addr} != $subnet)
          {
             $laddr{$addr}=$subnet;
             #FormatLineForPosix ("HDBDIST_IP_SUBNET","$subnet",$vhostindex); - print before *each* address
             my @addrparts=split(/\./,$addr);
             my $lastpart=@addrparts[-1];
             my $prefix = join(".", @addrparts[0..$#addrparts-1]);
             #LogVerbose "[ @addrparts ], $lastpart, $prefix";

             for my $offset (0 .. $maxoffset) {
                my $addr = "$prefix.".($lastpart+$offset);
                LogVerbose("nwrksupp address check for addr=[$addr]");
                `$nwrksupp ghbn $addr 2>/dev/null`;
                if ($? != 0) { LogVerbose("out of range: can't assign HDBDIST_IP_ADDRESS addr=[$addr]"); next; }

               	FormatLineForPosix ("HDBDIST_IP_SUBNET","$subnet",$vhostindex); # repeat subnet ()similiar to ip_subnet/address)
               	FormatLineForPosix ("HDBDIST_IP_ADDRESS",$addr,$vhostindex);
             }
             #FormatLineForPosix ("HDBDIST_IP_ADDRESS","$addr",$vhostindex);
             # no EDmodlist manipulation here...already included elsewhere
             $vhostindex++;
          }
          # else: IPv6 - WVE (basically same as above, but split at ':'. what about empty fields?)
        }
      }
    }

    LogReturnSub();
    return;
}

# expects a priority
sub FormatHDBPriorityForPosix {
    LogEnterSub( @_ );

    my $priority = shift;
    FormatLineForPosix ("SG_PRIORITY","$priority",0);
    $EDmodlist{$HoAttr{"SG_PRIORITY"}[1]}=1;

    LogReturnSub();
    return;
}


# Format vmfs related attributes
# Parameter: None (used to be: List of VGs)
# Global hash volgrps is used and modified
#
# Checks several prereq's
# - linux
# - vmfs module
# - system is running on VM
# - cluster has vserver or esx configured
#
# Usually vmdk filenames and datastores can't
# be retrieved. There is a SG command
# cmvmdiskquery, but this takes up to 1 minute
# to finish and may be also incomplete.
# However, if SAPGETCONF_CMDISKQUERY_OUTPUT is
# set to an output file of that command it is taken
# to get the attached vmdks and their datastore(s)
#
# Uses "global" vmfsindex
#
sub FormatVMFSForPosix {
    LogEnterSub( @_ );

    #my @vglist = @_;
    my $node_uuid;

    # Return immediately if
    # - vmfs module doesn't exist
    # - current system isn't a vmware system
    # - vcenter_name not set in cluser config
    # - ...
    #print "$SGCONFDIR/modules/sg/vmfs\n";
    unless ( -f "$SGCONFDIR/modules/sg/vmfs" ) {
      LogReturnSub(1);
      return;
    }
    # if Linux, test if this is a VMware system
    if ($^O eq "linux") {
      my ($manufac, $prodname) = getsysinfoProduct();
      LogVerbose("manufac=$manufac,prodname=$prodname");
      if ($prodname !~ /^VMware/) {
        LogReturnSub(2);
        return;
      }
    }
    unless ($node_uuid = getcluster_vcenter()) {
      LogReturnSub(3);
      return;
    }

    # all preq's meet now. get the attributes for each vg,pv,...
	#my @dslist;
	my $dsstring;
	my $uniqds;
	#my @fnlist;
	my $fnstring;
	my $uniqfs;
	my ($datastore_ref,$vmdkfiles_ref) = getDatastoreInfoFromFile ();
	my ($attached_vmdk_ref, $attached_vmdk_ds_ref) = getattachedVMDKs ($vmdkfiles_ref);
	if (scalar (@{$attached_vmdk_ds_ref}) > 1) {
	  $dsstring = join('|', @{$attached_vmdk_ds_ref});
	}
	elsif (scalar (@{$attached_vmdk_ds_ref}) == 1) {
	  $uniqds = $attached_vmdk_ds_ref->[0];
	}
	else {
	  $dsstring = "provide_value_manually_before_cmapplyconf"
	}
	if (scalar (@{$attached_vmdk_ref}) > 1) {
	  $fnstring = join('|', @{$attached_vmdk_ref});
	}
	elsif (scalar (@{$attached_vmdk_ref}) == 1) {
	  $uniqfn = $attached_vmdk_ref->[0];
	}
	else {
	  $fnstring = "provide_value_manually_before_cmapplyconf"
	}

    for my $vg (keys %volgrps) {
	  next if $volgrps{$vg} == 0; # kludge - already handle here
      LogVerbose("$vg");
      my @pvlist = getPVsforVG("$vg");
      LogVerbose("$vg - [@pvlist]");
      my @pvinfo = getdisktypeForPV(@pvlist);
      foreach my $pvi (@pvinfo) {
              LogVerbose ("$pvi->{device} - $pvi->{lunpath} - $pvi->{vendmod}");
			  if ($uniqfn) {
			    FormatLineForPosix ("SG_VMDK_FILE_NAME","\"$uniqfn\"", 0);
			  }
			  else {
                FormatLineForPosix ("SG_VMDK_FILE_NAME",$fnstring."_".$vmfsindex, 0);
			  }
			  if ($uniqds) {
			    FormatLineForPosix ("SG_DATASTORE_NAME","\"$uniqds\"", 0);
			  }
			  else {
                FormatLineForPosix ("SG_DATASTORE_NAME",$dsstring."_".$vmfsindex, 0);
			  }
			  $vmfsindex++;
              FormatLineForPosix ("SG_SCSI_CONTROLLER",join(':',(split(':',$pvi->{lunpath}))[0,2]), 0);
              my $disktype = "RDM"; $disktype = "VMFS" if ($pvi->{vendmod} =~ /^VMware/) ;
              FormatLineForPosix ("SG_DISK_TYPE",$disktype,0);
              $EDmodlist{$HoAttr{"SG_VMDK_FILE_NAME"}[1]}=1;
			  $attrcntr++;
      }
	  $volgrps{$vg} = 0; # mark handled
    }
    LogReturnSub(0);
    return;
}

# format a list of filesytems
# If list isn't submitted as parameter, the mpsfs is taken
sub FormatFileSystemForPosix {
    LogEnterSub( @_ );
    my @fslist = @_;
    @fslist = ( sort keys %mpsfs ) unless (@fslist);

    foreach ( @fslist ) {
        LogDebug("$mpsfs{$_} -- $mpsfst{$_} ");
        # check if FS is known for directory (redundant for mpfs keys)
        if ( $mpsfs{$_} ne "" ) {
        FormatSingleFileSystemForPosix($_,$mpsfs{$_},$mpsfst{$_});
    }
    }
    # FormatSingleFileSystemForPosix fills global volgrps hash with
	# "new" volumegroup names
    # Used to trigger the collection and formatting of vmfs related
    # informtion (WVE: collect in an earlier phase)
    FormatVMFSForPosix();

    LogReturnSub();
}

sub FormatSingleFileSystemForPosix {
    LogEnterSub( @_ );

    my $dir = shift;
    my $fs  = shift;
    my $fst = shift;

    my $vg = ( split( /\//, $fs ) )[2];

    # if VG is on VxVM instead of LVM, vg is "vx" and
    # instfs has the form /dev/vx/dsk/<vx dgrp>/<vx vol>
    # instead of "vg:", output is "vxvm_dg:<vx_dgrp> | .."
    # to see if the VxVM DG belongs to a cluster
    # vxdg list should be executed and STATE (field 2) should be checked if it
    # contains attribute "shared".

    if ( ( $vg eq "vx" ) && $vxvm_installed ) {
      $vg = ( split( /\//, $fs ) )[4];
      if (!defined $volgrps{$vg}) {
         $volgrps{$vg}=1;
         if ( $vxdgshared{$vg} ) {
           FormatLineForPosix ("SG_CMV_DG","$vg", $vgindex);
         }
         else {
           FormatLineForPosix ("SG_VXVM_DG","$vg", $vgindex);
         }
         $EDmodlist{$HoAttr{"SG_VG"}[1]}=1;  # same module, so SG_VG is ok here
         $vgindex++;
      }
    }
    else {
      # undef vg if FS type is NFS
      if ($fst eq "nfs") {
        undef $vg;
      }
      # Assuming LVM. Ceck if vg="mapper" (Linux)
      elsif ($vg eq "mapper") {
        my $vglv = ( split( /\//, $fs ) )[3];
        $vglv =~ m#([A-Za-z0-9_.]+)-([A-Za-z0-9_.]+)#;
        $vg = $1;
      }
      if (defined $vg && !defined $volgrps{$vg}) {
        $volgrps{$vg}=1;
        FormatLineForPosix ("SG_VG","$vg", $vgindex);
        $EDmodlist{$HoAttr{"SG_VG"}[1]}=1;
        $vgindex++;
      }
    }

    # no duplicate check for filesystem
    if ($fst ne "nfs") {
      FormatLineForPosix ("SG_FS_NAME","$fs",$fsindex);
      FormatLineForPosix ("SG_FS_DIRECTORY","$dir",$fsindex);
      FormatLineForPosix ("SG_FS_TYPE","$fst",$fsindex);
    }
    else {
      # split off fs_server port from fs on last ':' (greedy)
      my ($fss,$fsn) = $fs =~ /(.+):(.+)/;
      FormatLineForPosix ("SG_FS_NAME","$fsn",$fsindex);
      FormatLineForPosix ("SG_FS_SERVER","$fss",$fsindex);
      FormatLineForPosix ("SG_FS_DIRECTORY","$dir",$fsindex);
      FormatLineForPosix ("SG_FS_TYPE","$fst",$fsindex);
    }
    $EDmodlist{$HoAttr{"SG_FS_NAME"}[1]}=1;
    $fsindex++;

    LogReturnSub();
}

#############################################################################
#
# FormatResourceForPosix
#
# Output resource sections for [A]SCS & ERS instances
# So far HP-UX only
#
# Parameter:
#  - SAP SID
#  - SAP Instance
#
#############################################################################
sub FormatResourceForPosix {
    LogEnterSub( @_ );

    my $SID=shift;
    my $inst=shift;

	if ($^O eq "linux")  {
      #FormatLineForPosix("##", "generic resource - Test for Linux",0);
      return unless (defined ($ENV{"SGeSAP_ENQOR_PKGNAME"}));
	  FormatGenResourceForPosix($SID,$inst);
      return;
    }

    if ( $inst =~ /^ERS/ ) {
        #  Output resource data for ERS instance:
        #  SGMgr output $resname:  resource:/applications/sap/enqor/ers33
        #  ...
        my $resname =
              "/applications/sap/enqor/" . $SID . "ers". substr( $inst, -2 );

        # This and resource for [A]SCS are printed out with index 0, because they never
        # end up in the same package
        FormatLineForPosix ("SG_RESOURCE_NAME","$resname",0);
        FormatLineForPosix ("SG_RESOURCE_START","automatic",0);
        FormatLineForPosix ("SG_RESOURCE_POLLING_INTERVAL","60",0);
        FormatLineForPosix ("SG_RESOURCE_UP_VALUE","!= SERVING",0);
        $EDmodlist{$HoAttr{"SG_RESOURCE_NAME"}[1]}=1;
    }    # end if ERS instance

    if ( $inst =~ /^ASCS|^SCS/ ) {
        my $ersinst;
        $ersinst = $SCSmap{$inst};
        if ( $ersinst ne "" ) {
            # has ERS
            my $resname =
                  "/applications/sap/enqor/" . $SID . "ers". substr( $ersinst, -2 );
            FormatLineForPosix ("SG_RESOURCE_NAME","$resname",0);
            FormatLineForPosix ("SG_RESOURCE_START","automatic",0);
            FormatLineForPosix ("SG_RESOURCE_POLLING_INTERVAL","30",0);
            FormatLineForPosix ("SG_RESOURCE_UP_VALUE","!= IDLE",0);
            $EDmodlist{$HoAttr{"SG_RESOURCE_NAME"}[1]}=1;
        }
    }  #end if [A]SCS instance

    LogReturnSub();
}

sub FormatGenResourceForPosix {
    LogEnterSub( @_ );

    my $SID=shift;
    my $inst=shift;
	my $resname_pref="sgesap.enqor_" . $SID . "_";
	my $resname;
	my $upcrit;

    if ( $inst =~ /^ERS/ ) {

        $resname = $resname_pref . $inst;
		$upcrit  = "!= 2";
    }

    if ( $inst =~ /^ASCS|^SCS/ ) {
        my $ersinst;
        $ersinst = $SCSmap{$inst};
        if ( $ersinst ne "" ) {
            # has ERS
            $resname = $resname_pref . $ersinst;
			$upcrit  = "!= 1";
        }
    }

	# print out if resname was defined
	if (defined($resname)) {
	   FormatLineForPosix ("SG_GENRES_NAME","$resname",0);
       FormatLineForPosix ("SG_GENRES_EVAL_TYPE","before_package_start",0);
       FormatLineForPosix ("SG_RESOURCE_UP_CRIT","$upcrit",0);
       $EDmodlist{$HoAttr{"SG_GENRES_NAME"}[1]}=1;
       # set flag for later printout of enqor dependency
       $genres_used=1;
	}

    LogReturnSub();
}

sub FormatDependencyForPosix {
    LogEnterSub( @_ );

    my $SID=shift;
    my $inst=shift;

    my $depname="enqor_dep";

	# output only once - if a generic resource has been used
	# return unless (defined ($ENV{"SGeSAP_ENQOR_PKGNAME"}));

	if ($genres_used) {
		FormatLineForPosix ("SG_DEP_NAME","$depname",0);
		FormatLineForPosix ("SG_DEP_COND",$ENV{"SGeSAP_ENQOR_PKGNAME"}. " = UP",0);
		FormatLineForPosix ("SG_DEP_LOC","same_node",0);
		$EDmodlist{$HoAttr{"SG_DEP_NAME"}[1]}=1;
	}
    LogReturnSub();
}


sub FormatHDBDependencyForPosix {
    LogEnterSub( @_ );

     #my $SID=shift;
     #my $inst=shift;
     my $deppkgname=shift;

    # same_node down dependency to "opposite" package.
    # to trigger this run on the cli:
    # export SGeSAP_HDBREP_PKGNAME=hdbsHA1;  sapgetconf.pl easy hdbinstance=HA1

    $depname =  $deppkgname."_rep";

	FormatLineForPosix ("SG_DEP_NAME",$deppkgname."_rep",0);
	FormatLineForPosix ("SG_DEP_COND",$deppkgname." = down",0);
	FormatLineForPosix ("SG_DEP_LOC","same_node",0);
	$EDmodlist{$HoAttr{"SG_DEP_NAME"}[1]}=1;
    LogReturnSub();
}


sub FormatServiceForPosix {
    LogEnterSub( @_ );

    my $pkgname=shift;
    my $postfix=shift;
    my $svcindex=shift,

    my $svccmd="$SGeSAPSVCDIR/sap".$postfix.".mon",
    my $svcname=$pkgname.$postfix;

    FormatLineForPosix ("SG_SERVICE_NAME","$svcname",$svcindex);
    FormatLineForPosix ("SG_SERVICE_CMD","$svccmd",$svcindex);
    FormatLineForPosix ("SG_SERVICE_RESTART","$SG_SERVICE_RESTART_DEFAULT",$svcindex);
    FormatLineForPosix ("SG_SERVICE_FAIL_FAST_ENABLED","$SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT",$svcindex);
    FormatLineForPosix ("SG_SERVICE_HALT_TIMEOUT","$SG_SERVICE_HALT_TIMEOUT_DEAULT",$svcindex);
    $EDmodlist{$HoAttr{"SG_SERVICE_NAME"}[1]}=1;

  LogReturnSub();
}

# special routine for NFS service as FormatServiceForPosix assumes SGeSAP service convention
sub FormatNFSServiceForPosix {
    LogEnterSub( @_ );

    my $pkgname=shift;
    my $postfix=shift;
    my $svcindex=shift;

    my $svccmd="/etc/cmcluster/scripts/nfs/".$postfix.".mon";
	$svccmd="\"\$SGCONF/scripts/tkit/nfs/tkit_module.sh nfs_monitor\"" if ($^O eq "linux");
    my $svcname=$pkgname.$postfix;
    $svcname=$pkgname."nfs_service" if ($^O eq "linux");

    FormatLineForPosix ("SG_SERVICE_NAME","$svcname",$svcindex);
    FormatLineForPosix ("SG_SERVICE_CMD","$svccmd",$svcindex);
    FormatLineForPosix ("SG_SERVICE_RESTART","$SG_SERVICE_RESTART_DEFAULT",$svcindex);
    FormatLineForPosix ("SG_SERVICE_FAIL_FAST_ENABLED","$SG_SERVICE_FAIL_FAST_ENABLED_DEFAULT",$svcindex);
    FormatLineForPosix ("SG_SERVICE_HALT_TIMEOUT","$SG_SERVICE_HALT_TIMEOUT_DEAULT",$svcindex);
    $EDmodlist{$HoAttr{"SG_SERVICE_NAME"}[1]}=1;

  LogReturnSub();
}
sub FormatSMDAForPosix  {
    LogEnterSub( @_ );

    foreach my $vh (keys %SMDAVHmap)  {

      my @smdaprof =
          </usr/sap/DA[A-Z,0-9]/SYS/profile/DA[A-Z,0-9]_SMDA[0-9][0-9]_$vh>; # potentially on NFS - may hang
      my $numprof = @smdaprof;

      if ( $numprof == 0 ) {
        LogVerbose ("No SMDA found for $vh");
        next;
      }
      elsif ($numprof > 1) {
        LogVerbose ("More than one SMDA for $vh - ignoring them");
        next;
      }

      LogVerbose ("SDMA profile found: $smdaprof[0]");
      my $lastpart = ( split( /\//, $smdaprof[0] ) )[6];
      ($sid,$inst,$dc) =  split ( /_/, $lastpart);
      LogVerbose ("\$sid=[$sid], \$inst=[$inst]");
      my $smdaid = "/usr/sap/$sid/$inst";
      LogVerbose ("\$smdaid=[$smdaid]");

      if (-d $smdaid) {
        my $smdafs=$mpsfs{$smdaid};
        if (defined ($smdafs)) {
          LogVerbose ("$smdafs mounted to $smdaid");
          FormatSingleFileSystemForPosix( $smdaid, $smdafs, $mpsfst{$smdaid} );
          printSGMgr("# Formatting SMDA as external instance not implemented yet")
        }
        else {
          warn "Instance [$inst] for host [$vh] found, but no filesystem mounted onto instance directory [$smdaid]";
        }
      }
      else {
        warn "Profile [$smdaprof[0]] found, but no instance directory [$smdaid]";
      }
    }
    LogReturnSub();
}

# format NFS attributes (and related vg and FS)
# will print out also attributes for NFS monitor (not done for SGMgr routine)
# Different key on Linux
sub FormatNfsForPosix {
    LogEnterSub( @_ );

	my $NFS_XFS_key="NFS_XFS";
	$NFS_XFS_key="NFS_XFS_LX" if ($^O eq "linux");
    $space=' ';
    $space='' if ($^O eq "linux");

	printSGMgr("# -- NFS related variables -- ");

    #############################
    if ($^O eq "linux") {
        for $pathname (keys %XFSPathnameOptionsmapLinux ) {
            next if (scalar keys %{$XFSPathnameOptionsmapLinux{$pathname}} == 0);
            # Linux can have multiple export lines for same pathname
            for $nfsclient (keys %{$XFSPathnameOptionsmapLinux{$pathname}}) {
                # get options for each nfs client
                my $opt=$XFSPathnameOptionsmapLinux{$pathname}{$nfsclient};
                FormatLineForPosix ("$NFS_XFS_key","\"-o $opt $nfsclient:$pathname\"",0);
            }
            if (defined $AutodirNodemap{$pathname}) {
		      # collect vhosts
		      my $vhost = $AutodirNodemap{$pathname};
		      $vhosts{$vhost}=1 unless defined $vhosts{$vhost} || $vhosts{$vhost} != 0;
            }
		    $EDmodlist{$HoAttr{"$NFS_XFS_key"}[1]}=1;
        }



    #############################
    } else {

        for $pathname (keys %XFSPathnameOptionsmap ) {
            # Using generic FormatSingleFilesystemForPosix  (which also includes the VG output)
            # FormatSingleFileSystemForPosix ($pathname, $mpsfs{$pathname}, $mpsfst{$pathname}) ;

            # NFS related output (need to clarify if module default values should be provided here)
            FormatLineForPosix ("$NFS_XFS_key","\"-o $XFSPathnameOptionsmap{$pathname} $pathname\"",0);

            if (defined $AutodirNodemap{$pathname}) {
            # collect vhosts
            my $vhost = $AutodirNodemap{$pathname};
            $vhosts{$vhost}=1 unless defined $vhosts{$vhost} || $vhosts{$vhost} != 0;
            }
            $EDmodlist{$HoAttr{"$NFS_XFS_key"}[1]}=1;
        }
    }
    #############################

    FormatFileSystemForPosix();
    FormatIPForPosix();

    if (%XFSPathnameOptionsmap > 0) {
      my $pkgname=($ENV{"SG_PACKAGE"} || "\${SG_PACKAGE}");
      FormatNFSServiceForPosix($pkgname,"nfs_upcc",$svcindex++);
    }

    LogReturnSub();
}

# format module list
sub FormatModlistForPosix {
    LogEnterSub( @_ );

    foreach my $modname (keys %EDmodlist)
    {
      FormatLineForPosix ("SG_MODULE_NAME","$modname",0);
    }
    LogReturnSub();
}

# output of a single line in "Posix format" (now cmgetconf/cmmakepkg)
# uses HoAttr for output to retrieve attribute name
sub FormatLineForPosix {
  my ($left,$right,$index) = @_;

  if (!defined $HoAttr{$left}) {
    warn "No output attribute defined for $left - ignored\n";
    return;
  }

  #$right = "\"".$right."\"" if $right =~ m/\s/; -- no automatic quoting for strings
  if (defined($right)) {
    printf ("%s %s\n","$HoAttr{$left}[0]","$right");
  }
  else {
  	printf ("%s\n","$HoAttr{$left}[0]");
  }
}

# FormatAction
#
# Request an action from the consumer of these lines
# e.g. removal of one or more attributes from a package
#
sub FormatAction {
  my ($target,$grpid,$pkg,$action,$left,$right) = @_;
  printf ("@%s: %d %s %s ", $target, $grpid, $pkg, $action);
  FormatLineForPosix ($left, $right);
}


#############################################################################
#
# End of
# Output functions for POSIX format
#
#############################################################################



#############################################################################
#
# find_apiversion
#
# In /etc/cmcluster/modules/sgesap - scan thru all files which are symbolic
# links and "grep" for the version in the "Module Feature Indicator" string
# OR versions to bitmask
#
#############################################################################

sub find_apiversion {
#    LogEnterSub( @_ );

    # discard unless file is a symbolic link
    return unless -l;

    # just to log link/file names found
    my $F=$File::Find::name;
    LogVerbose "F=[$F]";

    open ($fh, $F) or warn "open $F failed:  \$?=[$?] \$!=[$!]" && return;

    while (<$fh>) {
        chomp ($_);

        # search for "# @(<<hash<) Module Feature Indicator    :  "
        # (module feature indicator is hex!)
        if (   /^# @\(#\) Module Feature Indicator    :\s+([[:xdigit:]]+).*$/ ) {
            LogVerbose("\$_=[$_] Found [$1]");
            # add to bitmask
            $apiversion |= hex($1) ;
            last;  # stop searching if string found
        }
    }
    close ($fh);
#    LogReturnSub("\$apiversion=[$apiversion]");
}
#############################################################################
#
# Return an apiversion based on features supported
#
# Returns feature bitmask
# ... based on "Module Feature Indicator" bitmask
#
# apiversion | 0x01 = DB2 support
# (equivalent to: cmmakepkg -l -f line -m sgesap/db  output contains line
# module:sgesap/db_global.2|attribute:sgesap/db_global/db_vendor|legal_values=db2)
#            | 0x02 = sapwebdisp support (in sapinfra)
#            | 0x04 = parallel startup (in sapinstance)
#            | 0x08 = services output by this script if enabled (default)
#            | 0x10 = Sybase support
# (module:sgesap/db_global.<n>|attribute:sgesap/db_global/db_vendor|legal_values=sybase)
#
#            | 0x40 = restart on cross subnet failover
# )
#############################################################################
sub getapiversion {
    LogEnterSub( @_ );
    $apiversion = 0;

   # Looks into a linked modules

    find ({ wanted => \&find_apiversion,follow => 1 }, "$SGCONFDIR/modules/sgesap");
    LogVerbose "apiversion from modules: $apiversion";
    $apiversion |= 0x08 if ($servoutsgmgr);
    # TODO INVESTIGATE
    # TODO Skip check for xsubnet. Current subroutine will fail on Linux anyway.
    $apiversion |= 0x40 if (xsubnapiversion($xsubnetnwlist));

    LogReturnSub("\$apiversion=[$apiversion]");

    return $apiversion;
}

#############################################################################
#
# Figure out if apiversion bit for "restart after cross subnet failover"
# should be enabled.
# Decision is based on the list of subnet prefixes provided as parameter.
# Elements of this list will be matched to subnets configured on the interfaces.
# If there is a match, true will be returned
#
# Paramter:
#   $xsubnetnwlist - comma separated list of subnet prefixes
#
# Needs:
#   (nwrksupp to get the list of interfaces with their subnets)
#    uses "netstat -in" instead of nwrksupp
#
# Restrictions:
#   - IPv4 only
#   (- lan network prefix only)
#
# Returns
#  0    - not enabled
#  1    - enabled
#
#############################################################################
sub xsubnapiversion {
  LogEnterSub( @_ );
  my @subnetprefixes = split(/ *, */,shift);
  my %subnets = ();
  #my $cmd=qq[/opt/cmluster/sap/bin/nwrksupp xOptionTBDx 2>/dev/null |];
  # use "netstat -in" instead of nwrksupp (latter has no option for this yet)
  my $cmd=qq[netstat -in 2>/dev/null |];
  LogVerbose ("subnet prefixes: [@subnetprefixes]");
  return 0 unless scalar(@subnetprefixes);

  open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  while (<$fh>) {
        my $subnet = (split(/ +/,$_))[2];;
        chomp($subnet);
        # LogVerbose ("subnet: [$subnet]");
        $subnets{$subnet} = 1 if ($subnet =~ m/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/);
           # match only with IPv4 address
   }
   close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";
   @printarr = keys %subnets;
   LogVerbose ("subnets: [@printarr]");

   foreach $subnpref (@subnetprefixes) {
     #LogVerbose "$subnpref";
     return 1 if (grep (/^${subnpref}/,keys(%subnets)));
   }
   LogReturnSub("0");
   return 0;
}

#############################################################################
#
# collect all SGeSAP related alerts and write them to stdout
# Looks .alert files in the alrtbasedir.
# Also does some housekeeping on alert stage and tmp files.
#
# Arguments:
#  $basedir -
#  $pkg - Package for which to report alerts (optional). All package if not set.
#         There's no check if this is really a package in the cluster. So
#         it can be also considered as a prefix of the alerts (base)filename.
# Even if $pkg parameter is set all .stage and .tmp files are handled.
# Only in the final reporting stage, package specific ones are considered
#
#############################################################################
sub getalerts {

  LogEnterSub( @_ );

  my $basedir = shift;
  my $pkg = shift;
  $pkg="" unless defined $pkg;

  %HoP=();  # hash of packages

  # rm old/expired stage files first
  my @stagefiles = <$basedir/*.alert.stage>;

  LogVerbose "alerts basedir: $basedir";
  LogVerbose "lock expiration time (secs): $alrtlockexptime";
  LogVerbose "alert log lifetime (sec): $alrtlonglifetime";
  LogVerbose "\@stagefiles: @stagefiles";

   my $localhost=`hostname`;
   chomp($localhost);

  foreach my $stagefile (@stagefiles) {
    # check if cleanup is necessary,
    # may "orphan" all other related files further down
    if (file_is_older($stagefile,$alrtlonglifetime)) {
      my $prefix="";
      # check if package isn't up locally, otherwise leave stagefile untouched
      $prefix=getalertprefix($stagefile); # for SG_PACKAGE
      getpkginfo($SG_PACKAGE); # initializes HoP and/or gets the information from it

      LogVerbose "localhost: $localhost - pkg status: $pkginfo{status} - pkg owner: $pkginfo{owner}";
      sg_log($SG_PACKAGE,"localhost: $localhost - pkg status: $pkginfo{status} - pkg owner: $pkginfo{owner}");
      if (("$pkginfo{status}" eq "up") && ("$pkginfo{owner}" eq $localhost)) {
        # no op here anymore:
        # stage file has to survive to indicate that alert file isn't updated
        #
        # unhandled case: online de-configuration of service & service failed to remove stagefile.
        #                        needs also cleanup by sapgetconf
      }
      else {
        sg_log($SG_PACKAGE, "Warning: cleanup of alert staging file $stagefile");
        LogVerbose  "Warning: cleanup of alert staging file $stagefile";
        unlink $stagefile;
      }
    }
  }

  # go over all .alert.tmp first, - loop over .alert later

  my @alertfiles = <$basedir/*.alert.tmp>;
  LogVerbose "\@alertfiles (tmp): @alertfiles";
  foreach my $alertfile (@alertfiles) {
    # loop over all alert.tmp
    my $prefix=getalertprefix($alertfile);  # sets SG_PACKAGE & SG_SVC
    # sg_log ($SG_PACKAGE, "Test output for $SG_PACKAGE");
    my $stagefile=$prefix.".alert.stage";
    if ( -f $stagefile ) {
      LogVerbose "1: $alertfile \n";
      if (get_lock($alertfile)) {
        # move .stage.tmp to .stage in any case (even if outdated)
        my $alertfile2=$prefix.".alert";
        while (!get_lock($alertfile2)) { ;}  # nop
        # got lock here
        LogVerbose "Going to move $alertfile to $alertfile2 \n";
        unlink $alertfile2; #|| warn "unlink $alertfile2 failed";
        `mv $alertfile $alertfile2`; #|| warn "move failed";
        # check if Perl rename can be used instead of OS mv
        free_lock($alertfile2);
        free_lock($alertfile);
      }
      # else ignore tmp alert file for that invocation if lock couldn't be acquired
    }
    else { # there seem to be leftover files - cleanup
      sg_log($SG_PACKAGE, "Orphaned files for $prefix. Will be removed.");
      LogVerbose "Orphaned file(s) for $prefix, removing them... \n";
      unlink $prefix.".alert",$prefix.".alert.tmp";
      rmdir $prefix.".alert.lock";
      rmdir $prefix.".alert.tmp.lock";
    }
  } # end of .alert.tmp

  # all .alert
  my @alertfiles = <$basedir/$pkg*.alert>;
  LogVerbose "\@alertfiles: @alertfiles";
  foreach my $alertfile (@alertfiles)
  {
    my $prefix=getalertprefix($alertfile);  # sets SG_PACKAGE
    my $stagefile=$prefix.".alert.stage";

    if ( -f $stagefile )
    {
      # get the lock for the tmp to avoid another sapgetconf instance moving
      # this over into alert.
      LogVerbose "2: $alertfile \n";
      while (!get_lock($alertfile.".tmp")) { ;}  # nop
      if (get_lock($alertfile))
      {
        # Stage timestamp serves also as heartbeat that
        # monitor is active. Monitor won't write a new tmp if stage hasn't changed.
        # Therefore alert file has to be left untouched if stage is outdated
        # - Just rely on .stage's date to determine "outdated state".
        # - Do not deleted outdated .alerts!!, just write SERVICE_ERROR to stdout
        # .alerts may be a bit younger than .stage, because it moved later - thus being younger than alrtlonglifetime.
        # but not taken into account here.
        if (file_is_older($stagefile,$alrtlonglifetime)) {
          # (outdated) stage file exits, indiicating also that alertfile is (probably) outdated
          my $alertnum=1;
          LogVerbose "Updated alerts not available ($alertfile)";
          printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|name=PKG_SGESAP_SERVICE_WARNING:$alertnum");
          printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|alert_parameter=\"Updated alerts not available ($SG_SVC)\"");

          if ( -f "$ENV{SGRUN}/debug_$SG_PACKAGE" ) {
            $alertnum++;
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|name=PKG_SGESAP_SERVICE_WARNING:$alertnum");
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|alert_parameter=\"Package is in debug mode\"");
          }
          if ( -f "$basedir/paused_$SG_PACKAGE" ) {
            $alertnum++;
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|name=PKG_SGESAP_SERVICE_WARNING:$alertnum");
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|alert_parameter=\"Manual SAP administration operation ongoing\"");
          }
          getpkginfo ($SG_PACKAGE);
          if ( "$pkginfo{state}" eq "maintenance" ) {
            $alertnum++;
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|name=PKG_SGESAP_SERVICE_WARNING:$alertnum");
            printSGMgr("package:$SG_PACKAGE|sgesap_alert:PKG_SGESAP_SERVICE_WARNING:$alertnum|alert_parameter=\"Package is in state maintenance\"");
          }
        }
        else {
          report_alrtfile_content($alertfile);
        }
        free_lock($alertfile);
      }
      else {
        sg_log($SG_PACKAGE, "Couldn't get lock for $alertfile");
        LogVerbose "Couldn't get lock for $alertfile \n";
      }
      free_lock($alertfile.".tmp") ;
    }
    else { # check orphaned (again). Not neccessarily redundant, as there may be no alert.tmps.
      sg_log($SG_PACKAGE, "Orphaned files for $prefix. Will be removed.");
      LogVerbose "Orphaned file(s) for $prefix, removing them... \n";
      unlink $prefix.".alert",$prefix.".alert.tmp";
      rmdir $prefix.".alert.lock";
      rmdir $prefix.".alert.tmp.lock";
    }
  }
  LogReturnSub();
}


#############################################################################
#  more alerts related subroutines
#############################################################################

#############################################################################
# print the content of an alert file to stdot
# (no other processing done so far)
# removes the file, if so and return true, false otherwise
#
# Arguments:
#   $filename
#############################################################################
sub report_alrtfile_content
{
   LogEnterSub( @_ );
   my $fname=shift;

   open FILE,'<',$fname;
   while (<FILE>) {
     print $_;
   }
   close FILE;

   LogReturnSub();
}

#############################################################################
# check if file is older than a specified time in seconds
# 0: isn't older (or doesn't exist)
# 1: is older
#
# uses stat()'s mtime field
#
# Arguments:
#   $filename
#   $expiration-time
#############################################################################
sub file_is_older
{
  LogEnterSub( @_ );

  my $fname=shift;
  my $ltime=shift;
  my $mtime=0;

  return 0 unless -f $fname;

  $mtime = (stat ($fname))[9];
  LogDebug("$fname mtime: ",$mtime," / current time: ", time(), "");
  return 1 if (time()-$mtime > $ltime);
  return 0;
}

#############################################################################
# return prefix of a filename
# prefix = path+part of the filename before the first '.'
# string of form /dir1/dir2/.../dirn/fp1.fp2...fpn
#     --> /dir1/dir2/.../dirn/fp1
#
# Implicitly sets SG_PACKAGE and SG_SVC derived from prefix:
#
# Arguments:
#   $filename
#
# Returns:
#   prefix
#############################################################################
sub getalertprefix
{
  my $fname=shift;

  # LogEnterSub( @_ );

  # my $basename=(split(/\//,$fname))[-1]; # last part
  my @nameparts=split(/\//,$fname);
  my $dirname=join('/',@nameparts[0..$#nameparts-1]); # omit last part
  my $prefix=(split(/\./,$nameparts[-1]))[0]; # first part of last part of nameparts
  #my $prefix=(split(/\./,$basename))[0]; # part before "."
  # could also be

  # so var. basename won't be needed here

  (@nameparts, $SG_SVC)=(split(/_/,$prefix)); # part before and after "_"
  # SG_PACKAGE name may also contain '_'. Assuming that the service type doesn't
  # SG_SVC reflects the SGeSAP monitor which has written the alert file
  $SG_SVC=$nameparts[-1];
  $SG_PACKAGE=join('_',@nameparts[0..$#nameparts-1]);
  LogVerbose "SG_PACKAGE: $SG_PACKAGE, SG_SVC: $SG_SVC";

  #LogReturnSub("prefix: $dirname/$prefix");
  return $dirname."/".$prefix
}

#############################################################################
# simulate flocks
#
# necessary to be done this way because for some locks the other locker is a shell script
# trys the age locks and removes them if they are there w/o being touched within a certain period
#
# Arguments:
#   $filename
#
# works with <filename>.lock directory in the base directory
# Returns:
#   true/flase
#############################################################################
sub get_lock
{
  LogEnterSub( @_ );

  my $fname=shift;
  my $lfname=$fname.$alrtlockpostfix;
  my $retries=2*$alrtlockexptime*4;
  # try 2xlockexptime in total and 4x per second
  for ($count=$retries; $count>0; $count--)
  {
    return 1 if mkdir ($lfname);

    #sleep 1 if is_locked ($fname); # too long, use select with a 1/4 sec instead
    select(undef, undef, undef, 0.25) if is_locked ($fname);

  }
  sg_log($SG_PACKAGE, "Warning: Couldn't get lock for $fname");
  LogVerbose "Couldn't get lock for $fname \n";

  return 0;
}

sub free_lock
{
  LogEnterSub( @_ );

  my $fname=shift;
  my $lfname=$fname.$alrtlockpostfix;
  rmdir $lfname or warn "rmdir failed: $!" if -d $lfname;
  LogReturnSub();
  return 1;
}

# Test if locked.
# Might remove directory if file is expired
sub is_locked
{
  LogEnterSub( @_ );

  my $fname=shift;
  my $lfname=$fname.$alrtlockpostfix;

  if (lock_is_expired($lfname,$alrtlockexptime))
  {
    sg_log($SG_PACKAGE, "Warning: Breaking alert lock for $fname");
    LogVerbose "lock $lfname expired, going to break it\n";
    rmdir $lfname;
  }
  return 1 if -d $lfname;

  return 0;
}
#
# declare a lock as expired (old) if it is
# older than the number of seconds provided by the parameter
sub lock_is_expired
{
  LogEnterSub( @_ );

  my $lock=shift;
  my $exptime=shift;
  my $mtime=0;

  return 1 unless -d $lock; # not existing considered expired

  $mtime = (stat ($lock))[9];
  LogDebug("$lock mtime: ",$mtime," / current time: ",time(),"");
  return 1 if (time()-$mtime > $exptime);
  return 0;
}

#
# creating a sapgetconf alerts logfile
#
# parameter:
# - string to print
#
# currently unused
#
sub alertslog
{
  #basedir = shift @_
  $outtxt = shift;

  open  LOG, ">>$alrtbasedir/sapgetconf-alerts.log";
  print LOG $$," ",time(),": ", $outtxt;
  close LOG;
}

##

sub getpkginfo {
    LogEnterSub( @_ );

    my $pkg = shift;
    %pkginfo=(); # hash

    # get *all* package information into a buffer hash of hashes to avoid repeated queries . There's a very little chance
    # that a package state changed  until the next invocation of this sub and that the  buffer will therefore contain
    # outdated information.
    my $cmd=qq[cmviewcl -f line -l package 2>/dev/null |];

    if ( ! exists($HoP{$pkg}) ) {
      LogVerbose("Execute \$cmd=[$cmd]");

      open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
      while (<$fh>) {
        my ($left,$right) =split(/\|/,$_);
        chomp($left); chomp($right);
        my $pkgname=(split(/:/,$left))[1];
        my ($propname,$propval)=split(/=/,$right);
        chomp($pkgname); chomp($propname); chomp($propval);
        $HoP{$pkgname}{$propname}=$propval;
      }
      close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";
    }

    my @printarr;
    %pkginfo=%{ $HoP{$pkg} };
    verbose && @printarr = %pkginfo;
    LogVerbose("\%pkginfo=[@printarr]");

    LogReturnSub();
}

#############################################################################
# collect all priorities in cluster and assign cluster unique priorites to
# HANA replication system.
# Note: lower number == higher priority
# Note: a cluster could have multiple HANA instances running, using multiple
# priorities:
#
# Primary instance get < 1000
# Secondary instance get >= 1000
#############################################################################
sub collectpriorities {
    LogEnterSub( @_ );

	# TODO: not used
    my $replmode = shift;

    my $cmd=qq[cmviewcl -v -f line -s config -l package|];

    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while ($line = <$fh>) {
    # LogVerbose("$line");
        chomp $line;
        if ( $line =~ m#priority=(\d+)$# ) {
#            print("Found cmviewcl priority 1=[$1]");
            LogVerbose("$line");
            $pkgpriorities{$1} = 1;
        }
    }
    close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";

    $dumpdata && print ( Data::Dumper->Dump( [ \%pkgpriorities ], ["pkgpriorities"] ));


    LogReturnSub();
}

#############################################################################
# getcluster_smartquorum
#
# Get the value for the smart quorum (true/false or undefined).
# Report smartq allow/to be set if smart_quorum is set to on
#
# Uses: cmviewcl -f line -v -s config and searches output for
#   quorum_server:<node>|smart_quorum=(on|off)
# On older SG version smart_quorum doesn't exist (equiv. to false)
#
# Override default behaviour using environment variable SAPGETCONF_USE_SMARTQ:
# -1    - assume that cluster smart_quorum is off
# >=1    - assume smartq even if cluster smart_quorum is off
# (1: report smartq even if not enabled in cluster ,
#  2: report even if already assigned to another package -> skip getsid_scresused
# )
#
# Returns:
#  0 - no smart_quorum
#  1 - smart_quorum enabled
#############################################################################
sub getcluster_smartquorum {

    LogEnterSub( @_ );

    my $smartq = 0;  # assume disabled
    my $env_smartq = $ENV{"SAPGETCONF_USE_SMARTQ"} || 0;
    my $cmd=qq[cmviewcl -v -f line -s config|];
    # unfortunately, no matching -l type for quorum

    # first check ENV if there is an override to "enable" smart quorum even
    # if it won't be set in cluster config
    if($env_smartq != 0) {
      LogVerbose ("SAPGETCONF_USE_SMARTQ set ($env_smartq)");
      $smartq = 1 ;
      LogReturnSub($smartq);
      return $smartq;
    }

    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while ($line = <$fh>) {
        chomp $line;
        if ( $line =~ m#quorum_server:(\S+)\|smart_quorum=(on|off)$# ) {
            LogVerbose("$line");
            $smartq = 1 if ($2 eq "on");
            last;
        }
    }
    close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";

    LogReturnSub($smartq);
    return $smartq;
}

#############################################################################
# gethdb_haprovider
#
# check if the SGeSAP HA provider is configured in HANA's global.ini
# The provider is considered to be configured if global.ini contains the
# section [ha_dr_provider_SGeSAPcl] with an entry execution_order
#
# Argument: SID
#
# Returns:
#  0 - not configured
#  1 - configured in global.ini
#
# WVE: Get global.ini path defined externally (see also gethdbreplmode)
#
#############################################################################
sub gethdb_haprovider {

    LogEnterSub( @_ );

    my $SID     = shift;
    my $globini = "/usr/sap/${SID}/SYS/global/hdb/custom/config/global.ini";

    my $haprovider_configured = 0;

    unless (-r $globini) {
      warn "$globini not found";
      LogReturnSub ("$haprovider_configured");
      return $haprovider_configured;
    }

    open FILE,'<',$globini || warn "Couldn't open $globini";
    my $haprovidersection=0;
    while (<FILE>) {
      #print $_;
      if ($_ =~ m#\[ha_dr_provider_SGeSAPcl\]#) {
        LogVerbose("$_");
        $haprovidersection = 1;
        next;
      }
      if ($haprovidersection) {
        if ($_ =~ m#^execution_order\s*=\s*(\w+)\s*#) {
          $haprovider_configured = 1;
          LogVerbose("\$haprovider_configured=1");
          last; # found
        }
      }
    }
    close FILE;

    LogReturnSub($haprovider_configured);
    return $haprovider_configured;
}

#############################################################################
# getcluster_vcenter

# From SG 12.00.40 a vcenter server name or IP address can be defined.
# Alternatively, a esx server can be defined per host.
# This address is used to trigger virtual disk (either on VMFS or RDM)
# attach and detach through the new sg/vmfs module (existence of that
# module is assumed by this routine and should be checked before calling it)
#
# Along with vcenter or esx each server has a uuid (vmware "BIOS" uuid).
#
# Uses: cmviewcl -f line -v -s config -l cluster
# and searches output for
#   vcenter_name=(<name>|<address>)
# or
#   node:<localnode>|esx_name=(<name>|<address>)
#
# uuids pattern match is:
#   node:<localnode>|uuid=<uuid>
#
# On older SG version these attribute don't exist (equiv. to unset)
#
# No external override by a SAPGETCONF_ env variable possbile so far.
#
# Returns:
#  nothing - relevant attributes not found in cluster configs
#  <uuid> - uuid found for this host
#############################################################################
sub getcluster_vcenter {
    LogEnterSub( @_ );

    my $vcenter ;  # assume disabled
    my $esx;
    my $uuid;
    my $hostname=$ENV{"HOSTNAME"};

    my $cmd=qq[cmviewcl -v -f line -s config |];
    # unfortunately, no matching -l type for quorum

    open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
    while ($line = <$fh>) {
    # LogVerbose("$line");
        chomp $line;
        if ( $line =~ m#^vcenter_name=(\S+)# ) {
            LogVerbose("$line");
            $vcenter = $1; next;
        }
        if ( $line =~ m#^node:$hostname\|esx_name=(\S+)# ) {
            LogVerbose("$line");
            $esx = $1; next;
        }
        if ( $line =~ m#^node:$hostname\|uuid=(\S+)# ) {
            LogVerbose("$line");
            $uuid = $1; next;
        }
    }
    close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";

    undef $uuid unless ($vcenter || $esx);
    LogReturnSub($uuid);
    return $uuid;
}

# more helpers for vmfs related attributes
# WVE: move further up
#
# get the "Manufacturer" and "Product Name" out of
# dmidecode "System Information"
# calls: dmidecode -t 1
# returns: list (manufacturer, product name)
#
sub getsysinfoProduct {
  LogEnterSub();
  my ($manufac, $prodname);
  my $cmd2exec = "dmidecode -t 1";

  open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
  while (<IN>) {
    chomp;
    #print "$_\n";
    $manufac = $1 if (/\s+Manufacturer: (.+)$/);
    $prodname = $1 if (/\s+Product Name: (.+)$/);
  }
  close IN;
  LogReturnSub($manufac, $prodname);
  return ($manufac, $prodname);
}

# WVE: simply getting PVs for VG might not be sufficient.
#      there might another level of indirection there
sub getPVsforVG {
  LogEnterSub(@_);
  my $vgname = shift;
  my @pvlist = ();
  my $cmd2exec = "vgs -o vg_name,pv_name";

  open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
  while (<IN>) {
    chomp;
    #print "$_\n";
    push @pvlist,(split(/\s+/, $_))[2] if (/\s+$vgname\s+/);
    # split above assumes that line starts with whitespaces
  }
  close IN;
  LogReturnSub(@pvlist);
  return @pvlist;
}

#
# If SAPGETCONF_CMDISKQUERY is set, the file is used for (kind of
# offline reading) to get vmdk_filenames and datastores.
# Done offline because cmvmdiskquery might take quite a while.
# WVE:
# - filename as param
# - VM uuid as param to check against provided file
#
# Returns: ref to hash of datastores and ref to hash of vmdk files
#
sub getDatastoreInfoFromFile {
    LogEnterSub(@_);
	my %datastores = ();  # datastore -> files -> type,attached
	my %vmdkfiles = ();   # file -> datastore,type,attached
	my $F = $ENV{"SAPGETCONF_CMVMDISKQUERY_OUTPUT"}; # cmvmdiskquery -n <node-uuid>

	LogVerbose("File: $F");

	open ($fh, $F) or warn "open $F failed:  \$?=[$?] \$!=[$!]" && return;

	while (<$fh>) {
	  if (m#\|Datastore:(.+)\|name=\1$#) {
		$datastores{$1} = (); next
	  }
	#  if (m#\|Datastore:(.+)\|File:(.+)\|name=\1$#) {
	# \1 above won't work ??
	  if (m#\|Datastore:(.+)\|File:(.+)\|name=(.+)$#) {
	  # assumes that $2 = $3
	  # print "$2 $3 - $1\n";
		$datastores{$1}->{$2} = ();
		#$vmdkfiles{$2} = (); # init, check if necessary
		$vmdkfiles{$2}->{"datastore"} = $1;
		LogVerbose("$2 - $1");
		next;
	  }
	  if (m#\|Datastore:(.+)\|File:(.+)\|type=(VMFS|RDM)$#) {
		$datastores{$1}->{$2}->{"type"} = $3;
		$vmdkfiles{$2}->{"type"} = $3;
		next;
	  }
	  if (m#\|Datastore:(.+)\|File:(.+)\|attached=(no|yes)$#) {
		$datastores{$1}->{$2}->{"attached"} = $3;
		$vmdkfiles{$2}->{"attached"} = $3;
		next;
	  }
	}
	close ($fh) or warn "close $F failed:  \$?=[$?] \$!=[$!]" ;

	#print STDERR Data::Dumper->Dump( [ \%datastores ], [ "datastores" ] );
    #print STDERR Data::Dumper->Dump( [ \%vmdkfiles ], [ "vmdkfiles" ] );

	LogReturnSub(\%datastores,\%vmdkfiles);
	return (\%datastores,\%vmdkfiles);
}

#
# get the list of vmdks attached to the VM
#
# Input: ref to hash of vmdk (returned from getDatastoreInfoFromFile)
#
# Returns: refs to 2 lists
#          1. List of attached vmdk(s)
#          2. Datastore(s) on which they reside
#
sub getattachedVMDKs () {
  LogEnterSub(@_);
  my $vmdk_ref = shift;
  my @attached_vmdks = ();
  my @attached_vmdks_ds = ();
  my %vmdks_ds = ();

  foreach my $vmdk (keys %{$vmdk_ref}) {
    #LogVerbose ("$vmdk - ".$vmdk_ref->{$vmdk}->{"attached"});
    if ($vmdk_ref->{$vmdk}->{"attached"} eq "yes") {
	  push @attached_vmdks,$vmdk;
	  $vmdks_ds{$vmdk_ref->{$vmdk}->{"datastore"}} = 1;
	}
  }

  @attached_vmdks_ds = keys %vmdks_ds;
  LogVerbose ("attached vmdks: @attached_vmdks");
  LogVerbose ("attached ds: @attached_vmdks_ds");

  LogReturnSub("@attached_vmdks - @attached_vmdks_ds");
  return (\@attached_vmdks, \@attached_vmdks_ds);
}

# Try to figure out if PV is a "Virtual Disk" or a RDM
#
# Input: List of PV
# Output: List of (correspodning) hashes ("lunpath", "vendmod" (vendor&model), "rev" )
#
# H:C:T:L = Host, Channel, Target, Lun
#
# Example 1: 3PAR(3PARdata) VV
# [1:0:0:11]   disk    3PARdata VV               3110  /dev/sdl
#
# Exampel 2: VMware - Virtual Disk
# [1:0:3:0]    disk    VMware   Virtual disk     1.0   /dev/sde
#
# HOWEVER: The scsi ids reported locally in the VM's OS might not be those
#          used as SCSI(x:y) in the VM config (at least on for the y).

sub getdisktypeForPV {
  LogEnterSub(@_);
  my @pvdisks = @_;
  my %diskmap = ();
  my @pvinfo = ();

  my $cmd2exec = "lsscsi";

  open IN, "$cmd2exec |" or die "Could not run $cmd2exec: $!\n";
  while (<IN>) {
    chomp;
    if (m#\[(\S+)\]\s+disk\s+(.+)\s(\d\S*)\s+(/\S+)\s*$#) {
      #print "$_\n";
      $diskmap{$4} = { "device" => $4 , "lunpath" => $1, "vendmod" => $2, "rev" => $3 };
    }
  }
  close IN;
  # print %diskmap;

  my $arridx = 0;
  foreach my $pvd (@pvdisks) {
    #print "$pvd exists\n" if exists $diskmap{"$pvd"};
    $pvinfo[$arridx] = $diskmap{"$pvd"} if exists $diskmap{"$pvd"};
    $arridx++;
  }
  LogReturnSub(@pvinfo);
  return @pvinfo
}

#############################################################################
# getsid_scresused
#
# Get HANA sid and primary+secondary pkg name for a given resource.
# (Used to avoid duplicate definition of site controller resource)
#
# Parameter: Resource name
#
# Uses: cmviewcl -f line -v -s config -l package
# and searches output for matching
#   package:<pkg>|generic_resource:$scresname|name=$scresname
#   package:<pkg>|sgesap/hdb_global/hdb_system:<SID>|sgesap/hdb_global/hdb_system=<SID>
#
# Returns:
# Array containing HANA SID, primary and secondary package
# (dbsid, prim-pkg, scnd-pkg) if exactly one SID (with a package pair) already exists
# undef otherwise (undef,undef,undef) if none or more than 1 SID exists
#
#############################################################################
sub getsid_scresused {

  LogEnterSub( @_ );

  my $scresname = shift;

  my @rtnarray = (undef,undef,undef);

  my %screspkgs=(); # hash. pkg containing scgenres as key
  my %sidpkgs=(); # hash, pkg -> sid
  my %scressid=(); # SIDs having scgenres defined

  my $cmd=qq[cmviewcl -v -f line -l package -s config|];

  open ($fh, $cmd) or warn "popen \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]";
  while ($line = <$fh>) {
  # LogVerbose("$line");
      chomp $line;
      if ( $line =~ m#package:([\S-]+)\|generic_resource:${scresname}\|name=${scresname}$# ) {
#         print("Found cmviewcl scgenres package [$1]");
          LogVerbose("$line");
          $screspkgs{$1}=1;
      }
      if ( $line =~ m#package:([\S-]+)\|sgesap/hdb_global/hdb_system:(\S+)\|sgesap/hdb_global/hdb_system=(\2)$# ) {
#         print("Found cmviewcl hana SID package [$1 - $2]");
          LogVerbose("$line");
          $sidpkgs{$1}=$2;
      }
  }
  close($fh) or warn "pclose \$cmd=[$cmd] failed: \$?=[$?] \$!=[$!]\n";

  # look if there is something in screspkgs already
  # if so figure out SID with pkgs as key
  if (scalar (keys  %screspkgs) > 2) {
    LogVerbose "WARNING: More than 2 packages have $scresname defined: [". keys(%screspkgs) ."]";
  }

  foreach my $pkg (keys %screspkgs) {
    LogVerbose "$pkg -> $sidpkgs{$pkg}";
    $scressid{$sidpkgs{$pkg}}=1 if defined $sidpkgs{$pkg};
  }

  LogVerbose %scressid;

  if (scalar(keys %scressid) == 1) {
    @rtnarray=((keys %scressid)[0],(keys %screspkgs)[0,1]);
  }
  elsif (scalar(keys %scressid) > 1) {
    LogVerbose "More than 1 SID found";
  }

  LogReturnSub(@rtnarray);
  return @rtnarray ;
}

#############################################################################
# get new / unused priority

# Primary instance get < 1000
# Secondary instance get >= 1000
#############################################################################
sub sethdbpriority {
    LogEnterSub( @_ );

    my $replmode = shift;
    my $startpriority=1000;
    my $p=$startpriority;

    if ($replmode =~ /^primary/) {
        while ($p > 11) {
            $p=$p-10;
            last if $pkgpriorities{$p} != 1;
        }
    } else {
        $p=$startpriority-10;
        while ($p < 2090) {
            $p=$p+10;
            last if $pkgpriorities{$p} != 1;
        }
    }
    # save priority so any additional calls
    # will have correct state
    $pkgpriorities{$p} = 1;
    LogReturnSub("$p");
}

#############################################################################
#
# mapHDBhost2CLUnode
#
# Parameter: HDB SID
# Returns: HANA host to cluster node map (assoc. array
#          (for those hosts being configured to a node)
# create a map between HANA host and cluster node
# used later (e.g.) to replace a virtual HANA host with the cluster node name
# when it comes to the output of node_name for package config ("Posix")
#
# uses cmexe to visit all cluste nodes and gets the HANA host there by
# extracting it from the /usr/sap/sapservices contents
# this is the more "naive" approach. more sophisticated:
# 1: check with nodes belong to the local scaleout (e.g. ls /hana/shared/<HSID>/HDB??/*/daemon.ini)
# 2: If these nodes are also clusternodes, no need to check them, otherwise do
# 3: Pick one from the rest of the clusternodes returning another set of HANA nodes
#    (there may be one in cluster having no HANA installed)
# 4: Do the same as 2
# (repeat 3+4? for multi-tier scaleout?)
#
#############################################################################

sub mapHDBhost2CLUnode {
  LogEnterSub( @_ );
  my $SID = shift;
  my @clunodes = @_;
  my $sidadm = lc($SID)."adm";
  my %Hh2Cn = ();

  if (scalar @clunodes ==  0) {
    my $cmd=qq[$ENV{SGSBIN}/cmviewcl -f line -s config -l node 2>/dev/null];
    LogVerbose("Excecute cmd=[$cmd]");
    open IN, "$cmd |" or die "Can't run $cmd: $?";
    while (<IN>) {
      chomp;
      if ( $_ =~ m#^node:(\S+)\|name=(\S+)$# ) {
        push @clunodes,$1;
        next;
      }
    }
    close IN or warn "Couldn't run $cmd: $?";
  }
  LogVerbose("cluster nodes: [@clunodes]");

  foreach my $cnode (@clunodes) {
    my $cmd2 = "cmexec $cnode \'cat /usr/sap/sapservices\'";
    open IN, "$cmd2 |" or die "Can't run $cmd2: $?";
    while (<IN>) {
      chomp;
      if ( $_ =~ m#pf=/usr/sap/${SID}/SYS/profile/${SID}_HDB[0-9][0-9]_(.*) -D# ) {
        $Hh2Cn{$1} = $cnode;
        LogVerbose("Hh2Cn mapping: [$1 -> $cnode]");
        # stop at first entry found. otherwise there would be more than 1 mapping
        last;
      }
    }
    close IN or ( warn "Couldn't run $cmd2: $?" and $Hh2Cn{$cnode} = $cnode );
  }
  my @printarr = %Hh2Cn;
  LogVerbose("Hh2Cn mappings: [@printarr]");
  LogReturnSub(@printarr);
  return (%Hh2Cn);
}

# stub:
# - requires the local node to be in cluster
# - gets all the node in cluster
#
# - starting from localnode (or just first from clusternode-list?) get all the HANA nodes seen by it
#   ... through parsing /hana/shared/<HSID>/HDB??/*/daemon.ini
#   (if that list is empty "<none>" is assigned as HANA host to cluster node -- or empty string?)
# - if parsed node is in cluster node list, add it to mappings
#   else put it into a "unassgned" list (maybe not part of cluster or HANA virt. host)
# - get next from clusternode-list which hasn't a mapping yest and start over
#   if at end of clusternode-list there are unassigned HANA hosts loop through the clusternodes
#   not having an assignment yet and try to get the assigment like mapHDBhost2CLUnode.
#   Reverse the assigment. There stil could be clusternode w/ HANA host (-> doesn't have HANA installed)
#   or a HAAN host w/o a clusternode (installed outside of cluster)
#
# A cluster node is only query if it was seen "up" in node status. Otherwise the cmexec to it
# might hang (just being cautious... keep in mind however that on the other hand a query might
# not hang even if status is != up - but that we don't know in advance)
#
sub mapCLUnode2HDBhost {
  LogEnterSub( @_ );
  my $SID = shift;
  my $sidadm = lc($SID)."adm";
  my %Cn2Hh = ();
  my @clunodes = ();
  my %unassignedclunodes = ();
  my %unassignedclunodes_up = ();
  my %unassignedhanahosts = ();

  # WVE: duplicate - create separate sub
  my $cmd=qq[$ENV{SGSBIN}/cmviewcl -f line -l node 2>/dev/null];
  LogVerbose("Excecute cmd=[$cmd]");
  open IN, "$cmd |" or die "Can't run $cmd: $?";
  while (<IN>) {
    chomp;
    if ( $_ =~ m#^node:(\S+)\|name=(\S+)$# ) {
      my $clunode=$1;
      push @clunodes,$clunode;
      $unassignedclunodes{"$clunode"}=1;
      #next;
      # instead:
      # - no "-s config"
      # next line here (assumes: status next line after name)
      # scan for: ^node:<last node>|status=(\S+)$
      # if not "up" remove it from ...
      if ( <IN> =~ m#^node:($clunode)\|status=(\S+)$# ) {
        if ($2 eq "up") {
          $unassignedclunodes_up{$clunode} = 1;
        }
        else {
          $unassignedclunodes_up{$clunode} = 0;
        }
      }
    }
  }
  close IN or warn "Couldn't run $cmd: $?";
  LogVerbose("cluster nodes: [@clunodes]");

  foreach my $cnode (@clunodes) {
    # if not assigned yset
    if ( $unassignedclunodes{"$cnode"} && $unassignedclunodes_up{"$cnode"} ) {
      # get HANA hosts this system knows about
      my $cmd=qq[$ENV{SGSBIN}/cmexec $cnode 'ls /hana/shared/$SID/HDB??/*/daemon.ini' 2>/dev/null];
      LogVerbose("Excecute cmd=[$cmd]");
      open IN, "$cmd |" or die "Can't run $cmd: $?";
      my $hhost;
      my $hhostnum = 0;
      while (<IN>) {
        $hhost = (split('/',$_))[5];
        $hhostnum++;
        LogVerbose("hhost=$hhost");
        # check if this is a cluster node and not assigned yet
        if ( exists $unassignedclunodes{$hhost} ) {
          if ( $unassignedclunodes{$hhost} ) {
            $Cn2Hh{$hhost}=$hhost;
            $unassignedclunodes{$hhost} = 0;
            next;
          }
          # else: already assignAllUnassignedInst
        }
        else {
          # either a virtual hana name or something not in cluster
          # store it in the unassigned hana hosts
           $unassignedhanahosts{$hhost} = 1;
           # possible optimization (implemented below after while loop):
           # if there is just one hhhost within the while loop
           #   it can be considered the hana host for that cluster node. right?
        }
      } # while IN
      close IN or warn "Couldn't run $cmd: $?";
      # check hhostnum. if 0: no hhost found, if 1: exactly one hhhost found, use it
      # if >1: no decision posible
      if ($hhostnum == 1) {
        $Cn2Hh{$cnode}=$hhost;
        $unassignedhanahosts{$hhost} = 0;
        $unassignedclunodes{$cnode} = 0;
      }
    }  # unassigend cnode
  }  # foreach
  # if there are unassigned cluster nodes run hana2cluster mapping function with those unassigned
  my @printarr = %unassignedclunodes;
  LogVerbose("unassignedclunodes map: [@printarr]");
  my @printarr = %unassignedclunodes_up;
  LogVerbose("unassignedclunodes up map: [@printarr]");
  my @unassignedCluNodesA = grep { $unassignedclunodes{$_} == 1 } keys(%unassignedclunodes);
  my @unassignedHanaHostsA = grep { $unassignedhanahosts{$_} == 1 } keys(%unassignedhanahosts);
  LogVerbose "Unassigned cluster nodes: [@unassignedCluNodesA]";
  LogVerbose "Unassigned hana hosts: [@unassignedHanaHostsA]";
  if (scalar(@unassignedCluNodesA) > 0) {
    my %Hh2Cn = mapHDBhost2CLUnode ($SID,  @unassignedCluNodesA);
    # loop through map and add in reverse to Cn2Hh
    # also mark the related HANA host and cluster node as assigned
    for my $Hh (keys %Hh2Cn) {
      my $Cn = $Hh2Cn{$Hh};
      $Cn2Hh{$Cn} = $Hh;
      $unassignedclunodes{$Cn} = 0;
      $unassignedhanahosts{$Hh} = 0;
    }
    @unassignedCluNodesA = grep { $unassignedclunodes{$_} == 1 } keys(%unassignedclunodes);
    @unassignedHanaHostsA = grep { $unassignedhanahosts{$_} == 1 } keys(%unassignedhanahosts);
    LogVerbose "Unassigned cluster nodes /2 : [@unassignedCluNodesA]";
    LogVerbose "Unassigned hana hosts /2 : [@unassignedHanaHostsA]";
  }

  return (%Cn2Hh);
}

#############################################################################
#
# easydeploy - collect values for easy deployment
#
# Arguments:
#  - comma seperated list of name=value pairs (as a string)
#
# So far the following names are supported:
# sap_system: SAP SID (0..1)
# sap_instance: SAP instance name... (0..n)
# db_instance: DB SID (0..1)
# ...more...
#
# Example: sap_system=VM7,sap_instance=SCS15
#
#############################################################################
sub easydeploy {
  LogEnterSub( @_ );

  my $sapsystem;
  my $dbinstance;
  my $hdbinstance;
  my $repdbinstance;
  my $lcinstance;
  my %sapinstances=();
  my $nfsspec;
  %vhosts=(); # collect all (unique) vhosts
  %volgrps=(); # collect all (unique) VGs
  %filesyss=(); # collect all (unique) filesystems
  # global vars above TBD
  # look at "Creating Persistent Private Variables "
  #    http://docstore.mik.ua/orelly/perl/cookbook/ch10_03.htm
  # to limit the scope of the variables above to the output functions
  # using "BEGIN { .... }"
  # (also the indexes below)

  my $svcindex=0; # used internally be FormatInstance and FormatDB
  $vhostindex=0; # dto, but gobal (TBD)
  $vgindex=0; #dto
  $fsindex=0; # dto
  local $vmfsindex=0;  # uniq id for each vmfs

  my $edarglist = shift;
  $edarglist =~ s/[\s]//g;
  LogVerbose ("easy deployment argument list: $edarglist");

  my @nvpairs=split(/,/,$edarglist);
  # LogVerbose ("nvpairs: @nvpairs");

  foreach my $nvpair (@nvpairs)
  {
    my ($nvname,$nvvalue)=split(/=/,$nvpair);
    #LogVerbose("nvname: $nvname - nvvalue: $nvvalue -");
    if ($nvvalue eq "")  # this one doesn't work Why??
    {
      LogError("No value for $nvname specified. Ignored"),
      next;
    }

    if ($nvname eq "sapsid") {
      if (defined $sapsystem) {
        LogError("sapsid already set ($sapsystem). Ignored")
      }
      else {
        $sapsystem=$nvvalue;
        $pkgdescr = $pkgdescr." $sapsystem";
        LogVerbose "\$sapsystem set to $sapsystem\n";
      }
    }
    elsif ($nvname eq "sapinstance") {
      $sapinstances{$nvvalue} = 1;
      # $pkgdescr = $pkgdescr." $nvvalue";
      LogVerbose "$nvvalue added to list of sapinstances\n";
    }
    elsif ($nvname eq "dbinstance") {
      if (defined $dbinstance) {
        LogError("dbinstance already set ($dbinstance). Ignored");
      }
      else {
        $dbinstance=$nvvalue;
        $pkgdescr = $pkgdescr.",DB $dbinstance";
        LogVerbose "\$dbinstance set to $dbinstance\n";
      }
    }
    elsif ($nvname eq "hdbinstance") {
      if (defined $hdbinstance) {
        LogError("hdbinstance already set ($hdbinstance). Ignored");
      }
      else {
        $hdbinstance=$nvvalue;
        # check for "forced mode"
        $hdbforcedmode = "";
        if ($hdbinstance =~ /^([A-Z][A-Z,0-9][A-Z,0-9])(s|p)/) {
        	$hdbinstance = $1;
        	$hdbforcedmode = $2;
        	if ($hdbforcedmode eq "p") {
        		$hdbforcedmode = "primary";
        	}
        	else {
        		$hdbforcedmode = "sync"; # kind of sync
        	}
        }
        $pkgdescr = $pkgdescr.",HDB $hdbinstance";
        LogVerbose "\$hdbinstance set to $hdbinstance\n";
      }
    }
	elsif ($nvname eq "repdbinstance") {
	  if (defined $repdbinstance) {
        LogError("repdbinstance already set ($repdbinstance). Ignored");
      }
      else {
        $repdbinstance=$nvvalue;
        $pkgdescr = $pkgdescr.",DB RS $repdbinstance";
        LogVerbose "\$repdbinstance set to $repdbinstance\n";
      }
	}
    elsif ($nvname eq "lcinstance") {
      if (defined $lcinstance) {
        LogError("lcinstance already set ($lcinstance). Ignored")
      }
      else {
        $lcinstance=$nvvalue;
        $pkgdescr = $pkgdescr.",LC $lcinstance";
        LogVerbose "\$lcinstance set to $lcinstance\n";
      }
    }
    elsif ($nvname eq "nfs") {
      if (defined $nfsspec) {
        LogError("\$nfsspec already set ($nfsspec). Ignored")
      }
      else {
        $nfsspec=$nvvalue;
        $pkgdescr = $pkgdescr.",NFS";
        LogVerbose "\$nfsspec set to $nfsspec\n";
      }
    }
    else {
      LogError("unknown name $nvname - Igored");
    }
  }

  # check if sap_instance's are defined, but no sap_system
  if ((scalar keys %sapinstances > 0) and ($sapsystem eq "")) {
    LogError("sapinstance needs sapsid be set. Exiting");
    exit 2;
  }

  #  get into collecting values...

  if (defined $sapsystem) {
    #getsapinstance ($sapsystem, keys %sapinstances);
    getsapinstance ($sapsystem);
    FormatInstancesForPosix(keys %sapinstances);
  }

  if (defined $dbinstance) {
    @dbvalues=getdbinstance ($dbinstance);
    FormatDBForPosix(@dbvalues);
  }

  if (defined $hdbinstance) {
    @hdbvalues=gethdbinstance ($hdbinstance,$hdbforcedmode);
    # rewrite repl mode if "forced mode" (WVW: could be done in get routine already)
    if (($hdbforcedmode ne "") && ($hdbvalues[2] !~ $hdbforcedmode )) {
    	# forced mode doesn't match with actual (primary <-> sync),
    	# rewrite by replacing first entry in hdbvalues[2] (assuming 3 entries here)
      # $hdbvalues[2] = "$hdbforcedmode,@replarray[1,-1]"; # WVE: use splice on hdbvalues?
    	my @replarray = split(",",$hdbvalues[2]);
      shift @replarray;
    	$hdbvalues[2] = join(",",$hdbforcedmode,@replarray);
    	LogVerbose "rewriting replication mode: $hdbvalues[2]";
    }
    # get information on safesync generic resource usage on (forced) primary
    if ($hdbvalues[2] =~ /^primary/) {
      $ENV{"SAPGETCONF_USE_SAFESYNC"} = gethdb_haprovider($hdbinstance) unless (defined $ENV{"SAPGETCONF_USE_SAFESYNC"});
    }
    else {
      # do not use/set SAPGET_USE_SAFESYNC in other than primary
      $ENV{"SAPGETCONF_USE_SAFESYNC"} = 0 if (defined $ENV{"SAPGETCONF_USE_SAFESYNC"});
    }

    my %Hh2Cn = mapHDBhost2CLUnode($hdbinstance);
    FormatHDBForPosix(@hdbvalues,\%Hh2Cn);
  }

  if (defined $repdbinstance) {
    @repdbvalues=getrepdbinstance ($repdbinstance);
	  if (($repdbvalues[1] eq "SYB") && ($repdbvalues[2] ne "none")) {
		  FormatREPDBForPosix (@repdbvalues);
	  }
  }

  if (defined $lcinstance) {
    @lcvalues = getlcinstance ($lcinstance);
    FormatLivecacheForPosix(@lcvalues);
  }

  if (defined $nfsspec) {
    if ($nfsspec eq "1") {
      getnfsconfig("(/export/*)");
    }
    elsif ($nfsspec ne "0") {
    #  getnfsconfig ("(/export/sapmnt.*$nfsspec.*|/export/.*sap.*trans.*)");
       getnfsconfig ("$nfsspec");
    }
    # The code above assumens that exports for SAP SID start with "/export/sapmnt" and have a SID somewhere in them
    # For /usr/sap/trans the exported directory is also in /export and has "sap" and "trans" in it.
    #
    # To be implemented:
    # Special meaning for value "1", which means "print all nfs entries related to the config discovered so far".
    # Applies only if another option has been specified prior to "nfs=".
    # "nfs=0" surpresses NFS output.
    FormatNfsForPosix;
  }

  FormatLineForPosix(SG_PACKAGE_DESCRIPTION,"\"$pkgdescr\"");
  FormatModlistForPosix();

  # ToDo: only when VERBOSE is switched on
  my @printarr=();
  @printarr=%sapinstances;
#  LogError("sapinstances: @printarr ");

  LogReturnSub("\$sapsystem=<$sapsystem>,\$dbinstance=<$dbinstance>,\%sapinstances=<@printarr>");
}

#############################################################################
#
# sg_log - write to SG package logfile
#
# Arguments:
#  - package name
# - string to write
#
# Doesn't check loglevel
#
# uses
#
#############################################################################
sub sg_log
{
  LogEnterSub( @_ );

  my $pkgname=shift;
  my $outstr=shift;
  my $datefmt="+%b %e %H:%M:%S";  # should in sync with SGUTILS
  my $pkglog=$ENV{SGRUN}."/log/".$pkgname.".log";

  #LogVerbose "pkgname: $pkgname";
  #LogVerbose "outstr: $outstr";
  #LogVerbose "pkglog: $pkglog";

   `echo \$(date "$datefmt") \$(whoami)@\$(hostname) sapgetconf.pl[$$]: $outstr >> $pkglog`;
  # Alternative: use perl variables and functions to get date,user,hostname instead of shell commands
}

#############################################################################
#
# checkNFSdiraccess - check if a filespec is a directory and accessible on NFS
#
# Arguments:
#  - filespec
#
# Creates a subprocess for the check avoid the mainprocess to hang on NFS.
# Sets up a signal handler for ALRM signal and uses the function alarm() for timeout purposes.
# On timeout the subprocess is killed with signal INT.
#
# NFS caching effects may lead to wrong results if executed shortly after NFS has gone
# away or is available again
#
# Returns:
# 0 - is a directory (and accessible)
# 2 - not a directory (but accessible)
# 130 - check timed out (subprocess INTerrupted)
#
#############################################################################
sub checkNFSdiraccess {
  LogEnterSub( @_ );
  my $dirtocheck = shift;
  my $childpid = undef;
  my $rc=0;

  $childpid = fork();

  if ($childpid) {
    # in parent
    local $SIG{ALRM} = sub {
          LogError("Signal handler in checkNFSaccessdir triggered - terminate child with: kill INT $childpid");
          kill 'INT', $childpid;
    };
    alarm 5;
    waitpid ($childpid,0);
    my $childrc = $?;
    alarm 0;
    if ($childrc>>8) {
      $rc=2
    }
    elsif ((($childrc>>8) == 0) && ($childrc&127)) {
      LogError("Access failed - timed out on $dirtocheck");
      $rc=130;  # should be 128+$rc&127
    }
    LogReturnSub ("\$rc=$rc");
    return $rc;
  }
  else {
    # child
    exit 0 if (-d $dirtocheck);  # is dir (and accessible)?
    exit 2;  # not a dir
  }
}



sub catchSignal {
    my $signame = shift;
    die "Somebody sent me a SIG$signame";
}


#############################################################################
#
# This function is responsible for dispatching the requests based on the
# arguments to this script.
#
#  Arguments:
#       $ARGV[0]    $ARGV[1]
#       ----------- --------
#       sap_global
#       sapinstance MO7
#       dbinstance  [DO7]
#       lcinstance
#
#############################################################################
sub mainDispatcher {
    LogEnterSub( @_ );

    if (scalar @ARGV == 0) {
      LogError("Program parameters missing"),
      warn $USAGE;
      exit 2;
    }

    # get SGCONF and SGeSAP bin and SID dirs
    # override nwrksupp
    getSGdirs;
    $nwrksupp="$SGeSAPBINDIR/nwrksupp" if (defined($SGeSAPBINDIR));
    # set HOSTNAME ENV variable if not set (usually def'd on Linux, no on HPUX)
    $ENV{"HOSTNAME"}=(POSIX::uname())[1] unless (defined($ENV{"HOSTNAME"}));

    # check if output should be simulated by just printing out canned output
    my $simoutdir  = $ENV{"SAPGETCONF_SIMOUTDIR"};
    my $simoutfile = $ENV{"SAPGETCONF_SIMOUTFILE"};

    if (defined $simoutdir) {
      unless (defined $simoutfile) {
        $simoutfile=join("_",@ARGV);
        $simoutfile =~ s/[=,]/_/g;
      }
      LogVerbose "\$simoutdir=[$simoutdir],\$simoutfile=[$simoutfile]";
      open (my $fh,'<',"${simoutdir}/${simoutfile}") or die "Failed: open \$simoutfile=[$simoutfile] failed: \$?=[$?] \$!=[$!]";;
      while (<$fh>) {
        print $_ unless (/~#/);
      }
      close ($fh);
      exit 0;  # exit simulation
    }

    # no canned output. do real collection of values

    # write additional information if canned to file
    # - commented as it leads to unexpected results when using
    #   redirection in SGMgr (mainly apiversion?)
    #if (-f STDOUT) {
    #    print "## NODE\t$ENV{HOSTNAME}\n";
    #    print "## TIME\t".time()."\n";
    #    print "## DATE\t".localtime()." ".strftime("%Z", localtime())."\n";
    #    print "## CALL\t$0 @ARGV\n";
    #    my ($svnrev) = (q$Rev: 1723 $ =~ /(\d+)/);
    #    print "## REV\t$svnrev\n";
    #    print "##\n";
    #}

    while (@ARGV)
    {


      #######################################
      # sap_global
      #######################################
      if ( $ARGV[0] eq "sap_global" ) {
        shift @ARGV;
        # collect SIDs
        # LogVerbose("Scanning for SIDs in \$SAPSIDSInSapmnt=[$SAPSIDSInSapmnt]");
        LogVerbose("Scanning for SIDs in \$SAPSIDSInusrsap=[$SAPSIDSInusrsap]");
        my @files = <${SAPSIDSInusrsap}>;
        LogVerbose("\@files=[@files]");

        # get the group id of "sapsys" for comparing later
        my @sapsysgid = getgrnam("sapsys");
        LogVerbose("GID for sapsys: \$sapsysgid=[$sapsysgid[2]]");

        foreach $i (@files) {
            # SID will be the last item in the filename path -> ...[-1]
            # my $SID = ( split( /\//, $i ) )[-1];
            my $SID = ( split( /\//, $i ) )[3];
            LogVerbose("Found a \$SID=[$SID]");

            # Doublecheck if this really is a SAP SID by comparing the group the user belongs to.
            # Either "primary" GID is sapsys or <sid>adm belongs to sapsys group
            my $sidadm = lc($SID)."adm";
            if ( ((getpwnam($sidadm))[3] == $sapsysgid[2]) || (grep {$_ eq $sidadm} split(" ",$sapsysgid[3]))) {
              if (!defined $InstanceActiveFilter) {
                # no filter, but print only if not belonging to HDB or SMDA (based on directory check)
                FormatSIDForSGMgr($SID) unless (-d (</usr/sap/${SID}/{SMDA}[0-9][0-9]>)[0]) ;
              }
              else {
                @profiles=();
                CollectListOfSIDProfiles( $SID, "/sapmnt/$SID/profile", $InstanceActiveFilter); # need to check profiles for filtering, writes @profiles
                FormatSIDForSGMgr($SID) if (scalar @profiles > 0);  # some profiles "filtered in"
              }
            }
        }
      }
      #######################################
      # sapinstance
      #######################################
      elsif ( ( $ARGV[0] eq "sapinstance" ) and ( $ARGV[1] ne "" ) ) {
        LogVerbose("In sapinstance check [$ARGV[1]]");

        if ($testmode) {
          LogVerbose("Testmode on. Trying to find /tmp/test_".$ARGV[1].".sgc");
          print  `cat "/tmp/test_"$ARGV[1]".sgc"`;
          #print @cont;
          return 0;
        }

        $dirtocheck="/sapmnt/$ARGV[1]/profile";  # NFS - may hang
        if ( checkNFSdiraccess($dirtocheck) ) {
            # checkNFSdiraccess return >0 on failure
            LogError("No profile directory $dirtocheck");
            exit 2;
        }
        # just checking for profile directory on /sapmnt/<SID> might not be sufficient for some type
        # of installations. E.g diagnostic agent installations don't create the "SYS links" but keep SYS
        # on local FS.

        $dirtocheck="/usr/sap/$ARGV[1]";
        if ( !-d $dirtocheck ) {
            LogError("No local directory $dirtocheck");
            exit 2;
        }

        shift @ARGV; # get rid of "sapinstance"
        getsapinstance ($ARGV[0],$InstanceActiveFilter) ;
        shift @ARGV; # get rid of SID (may lead to empty list)
        FormatInstancesForSGMgr($ARGV[0]);
        shift @ARGV;
        #last; # "sapinstance" uses up all subsequent params as SID and (optional) instancenames
              # otherwise we'd need to search for the next keyword
      }
      #######################################
      # dbinstance
      #######################################
      elsif ( $ARGV[0] eq "dbinstance" ) {
        shift @ARGV;
        if ( $ARGV[0] ne "" ) {
            # get DB information for SAP SID, might be that <SID> != <DBSID>
            @dbvalues=getdbinstance $ARGV[0];
            if (scalar @dbvalues > 0) {
              FormatDBForSGMgr(@dbvalues);
              # special tweak for HANA scaleout. pretend to have discovered also
              # the primary or sync if local mode is sync or primary
              # (Alternative: have deploysappkgs "make up" this "other" instance)
              my @dbvaluesSO = createDBSOvalues (@dbvalues);
              FormatDBForSGMgr(@dbvaluesSO) if (scalar @dbvaluesSO > 0);
            }
            shift @ARGV;
        }
        else {
            LogVerbose("Scanning for SIDs in \$SAPSIDSInSapmnt=[$SAPSIDSInSapmnt]\n");
            my @files = <${SAPSIDSInSapmnt}>;
            LogVerbose("\@files=[@files]");

            # get the group id of "sapsys" for comparing later
            my $sapsysgid = getgrnam("sapsys");
            LogVerbose("GID for sapsys: \$sapsysgid=[$sapsysgid]");

            foreach $i (@files) {
                # SID will be the last item in the filename path -> ...[-1]
                my $SID = ( split( /\//, $i ) )[-1];
                LogVerbose("Found a \$SID=[$SID]");
                # verify SID
                if ( ( getpwnam( lc($SID) . "adm" ) )[3] == $sapsysgid ) {
                  @dbvalues=getdbinstance $SID;
                  FormatDBForSGMgr(@dbvalues) if (scalar @dbvalues > 0);
                }
                else {
                  LogVerbose ("... but no matching sidadm user exists");
                }
            }
        }
      }
	  #######################################
      # replicated dbinstance
      #######################################
      elsif (( $ARGV[0] eq "repdbinstance" ) and ( $ARGV[1] ne "" )) {
        # get DB information for a replicated DBSID
		shift @ARGV;
        @repdbvalues=getrepdbinstance $ARGV[0];
		shift @ARGV;
		# Format only for SYB and !=none
		if (($repdbvalues[1] eq "SYB") && ($repdbvalues[2] ne "none")) {
		  FormatREPDBForSGMgr @repdbvalues;
		}
	  }
      #######################################
      # hdbinstance
      #######################################
      elsif ( $ARGV[0] eq "hdbinstance" ) {
        shift @ARGV;
        if ( $ARGV[0] ne "" ) {
            # get HDB information for SID
            @dbvalues=gethdbinstance $ARGV[0];
            FormatHDBForSGMgr(@dbvalues) if (scalar @dbvalues > 0);
            shift @ARGV;
        }
        else {
            LogVerbose("Scanning for SIDs in \$HDBSIDInhshared=[$HDBSIDInhshared]\n");
            my @files = <${HDBSIDInhshared}>;
            LogVerbose("\@files=[@files]");

            # get the group id of "sapsys" for comparing later
            my $sapsysgid = getgrnam("sapsys");
            LogVerbose("GID for sapsys: \$sapsysgid=[$sapsysgid]");

            foreach $i (@files) {
                # SID will be the last item in the filename path -> ...[-1]
                my $SID = ( split( /\//, $i ) )[-1];
                LogVerbose("Found a \$SID=[$SID]");
                # verify SID
                if ( ( getpwnam( lc($SID) . "adm" ) )[3] == $sapsysgid ) {
                  @dbvalues=gethdbinstance $SID;
                  FormatHDBForSGMgr(@dbvalues) if (scalar @dbvalues > 0);
                }
                else {
                  LogVerbose ("... but no matching sidadm user exists");
                }
            }
        }
      }
      #######################################
      # lcinstance
      #######################################
      elsif ( $ARGV[0] eq "lcinstance" ) {    # and ($ARGV[1] ne "")) {
        shift @ARGV;
        if ($ARGV[0] eq "") {
          # get LiveCache information
          # list all LC information it can find, as they are not
          # directly related to a SAP SID
          getlcinstances;  # Formatting done within
        }
        else {
          # single LC name(s) provided as parameter
          @lcvalues = getlcinstance ($ARGV[0]);
          FormatLivecacheForSGMgr(@lcvalues);
          shift @ARGV;
        }
      }
      #######################################
      # mdminstance
      #######################################
      elsif ( $ARGV[0] eq "mdminstance" ) {
        shift @ARGV;
        if ( $ARGV[0] ne "" ) {

            LogVerbose("In mdminstance check [$ARGV[1]]");

            $dirtocheck="/sapmnt/$ARGV[0]/profile";  # NFS - may hang
            if ( !-d $dirtocheck ) {
                LogError("No profile directory $dirtocheck");
                exit 2;
            }

            $dirtocheck="/usr/sap/$ARGV[0]";
            if ( !-d $dirtocheck ) {
                LogError("No local directory $dirtocheck");
                exit 2;
            }
            getmdminstance $ARGV[0];
            # collect instance information even if MDSMinfo could not be found
            getsapinstance $ARGV[0];
            FormatInstancesForSGMgr();
            shift @ARGV;
        }
        else {
            # list all MDM and MDM database instances
            # found on the system
            %MDMSIDSmap      = ();
            getallmdminstances;
            FormatMDMSIDSForSGMgr();
        }
      }
      #######################################
      # get the nfs configuration for adding
      # to HA NFS
      #######################################
      elsif ( $ARGV[0] eq "nfs" ) {
            shift @ARGV;
            # optional filter
            if (defined $ARGV[0]) {
              getnfsconfig($ARGV[0]);
              shift @ARGV;
            }
            else {
              getnfsconfig;
            }
            FormatNfsForSGMgr();
      }
      # Sybase toolkit rs module "native" (unofficial, no easy output)
      elsif (( $ARGV[0] eq "sybtkrs" ) and ( $ARGV[1] ne "" )) {
        # get DB information for a replicated DBSID
		    shift @ARGV;
        @repdbvalues=getrepdbinstance $ARGV[0];
		    shift @ARGV;
		    # Format only for SYB and !=none
		    if (($repdbvalues[1] eq "SYB") && ($repdbvalues[2] ne "none")) {
		      FormatSYBTKRSForSGMgr @repdbvalues;
		    }
      }
      #######################################
      # get alerts from SGeSAP monitors
      #######################################
      elsif ( $ARGV[0] eq "alerts" ) {
        shift @ARGV;
        my $pkgpar = shift @ARGV if (defined $ARGV[0]);
        getalerts($alrtbasedir, $pkgpar);
        # As this is SGMgr specific anyway, there's no need to separate the output
        # ARGV[0] is a optional package name argument. Maybe undefined. As alerts
        # (like apiversion below) isn't combined with any SAP instance collection
        # it shouldn't be a problem to "consume" this argument.
      }
      #######################################
      # get SGeSAP module API version
      # and feature bits (future use)
      #######################################
      elsif ( $ARGV[0] eq "apiversion" ) {
        shift @ARGV;
        $apiversion=getapiversion() if $apiversion==0;
        printf ("%08X\n", $apiversion);
      }
      #######################################
      # generated Metrocluster site and nodes
      # for SAP HANA scaleout (undocumented)
      # WVE: generate from hdbnsutil -sr_state
      #  output when replication is fully
      # configured and up&running.
      # Output contains the "whole picture"
      #######################################
      elsif ( $ARGV[0] eq "hanametro" ) {
        shift @ARGV;
        # have getdbinstance report HDB properties
        $ENV{"SAPGETCONF_HDBASDBINSTANCE"}=1;
        my @hdbinfo = getdbinstance (shift @ARGV);
        # replication info is hdbinfo[2], hosts is hdbinfo[3]
        # [2] = mode, siteid, sitename
        # [3] = comma sep list of hana hosts
        my $hdbmode = (split(",",$hdbinfo[2]))[0];
        if ($hdbmode eq "none") {
        	warn "Replication mode not configured";
        	exit 2;
        }
        my $hdbsite = (split(",",$hdbinfo[2]))[2];
        print ("SITE_NAME $hdbsite\n");
        foreach my $hhost (split(",",$hdbinfo[3] )) {
        	print ("NODE_NAME $hhost\n");
        	print ("SITE $hdbsite\n");
        }
      }
      elsif ( ($ARGV[0] eq "Hh2Cn")  and ( $ARGV[1] ne "" ) ) {
          # HANA host to cluster node mapping
          shift @ARGV;
          my %Hh2Cn = mapHDBhost2CLUnode(@ARGV); shift @ARGV;
          foreach $Hh (keys %Hh2Cn) {
            print "$Hh;$Hh2Cn{$Hh}\n"
          }
          last;
      }
      elsif ( ($ARGV[0] eq "Cn2Hh")  and ( $ARGV[1] ne "" ) ) {
          # Cluster node to HANA host mapping
          shift @ARGV;
          my %Cn2Hh = mapCLUnode2HDBhost($ARGV[0]); shift @ARGV;
          my $mapit=0;
          $mapit=1 if (defined($ARGV[0]));
          $mapit=2 if (defined($ARGV[1]));
          print "$ARGV[0]\n" if ($mapit == 2);
          foreach $Cn (keys %Cn2Hh) {
            print "$Cn;$Cn2Hh{$Cn}\n" if ($mapit==0); # Cluster node to Hana host
            # the next to are Hana host map to cluser node (\n ??)
            print "$ARGV[0]$Cn2Hh{$Cn}=$Cn " if ($mapit==1);
            print "$ARGV[1]$Cn2Hh{$Cn}=$Cn\n" if ($mapit==2);
          }
          shift @ARGV if ($mapit>=1);
          shift @ARGV if ($mapit>1);
          last;
      }

      #######################################
      # easy deployment entry
      #######################################
      elsif (( $ARGV[0] eq "easy" ) and ( $ARGV[1] ne "" ))  {
        # parameter are a comma seperated list of name=value pairs
        # for a list of supported names, see documentation at easydeploy sub
        easydeploy("$ARGV[1]");
        shift @ARGV; shift @ARGV;
      }
      #
      # filter
      #
      elsif (( $ARGV[0] eq "-f" ) and ( $ARGV[1] ne "" ))  {
        shift @ARGV;
        $InstanceActiveFilter=shift @ARGV;
        # check against support symbolic filters
        unless (grep {$_ eq $InstanceActiveFilter} keys %InstanceFilters) {
          LogError("Unsupported instance filter: $InstanceActiveFilter");
          warn $USAGE;
          exit 2;
        }
      }
    #######################################
    # usage - problem with arguments
    #######################################
      else {
        LogError("Paramter(s) wrong or missing [ @ARGV ]");
        warn $USAGE;
        exit 2;
      }
    }
    LogReturnSub();
} # end mainDispatcher

#############################################################################
# Setup a timeout handler for this script. This is necessary as the under-
# lying NFS filesytems might cause the execution of this scriptto hang.
#
# The mainDispatcher() function is executed as a child. If it hangs the
# alarm timer will timeout and the parent wil kill the child
#
#############################################################################


LogVerbose"Script timeout set to \$timeout=[$timeout]";

# pid for this script
my $parentpid=$$;

$fork=$ENV{SAPGETCONF_FORK} if defined $ENV{SAPGETCONF_FORK};
$fork=1                        unless defined $fork;

if ( ! $fork ) {

  LogVerbose("INFO: No Child will be forked");

  sub CatchSignal {
      my $signame = shift;
      die "CatchSignal() was triggered - received SIG$signame";
  }

  $SIG{INT} = 'CatchSignal';
  $SIG{INT} = \&CatchSignal;
  $SIG{ALRM} = \&CatchSignal;

  # enable the alarm
  alarm($timeout);

  # Execute the sapgetconf mainDispatcher() function
  mainDispatcher();

  # disable the alarm
  alarm 0;

  # exit success
  exit 0;

} else {

  LogVerbose("INFO: Child will be forked");
  eval {
      # setup signal handler. If the alarm has expired
      # kill the child
      local $SIG{ALRM} = sub {
          LogError("Signal handler triggered - terminate child with: kill INT $childpid");
          kill 'INT', $childpid;
          die 'Alarm expired';
      };

      # set the alarm
      alarm($timeout);

      if ($childpid = fork) {
          LogVerbose("In Parent - child has \$childpid=[$childpid])");
          # wait for child to terminate
          waitpid ($childpid, 0);
          $save_child_rc=$?>>8;
          LogVerbose("In Parent - child returned with: \$save_child_rc=[$save_child_rc]");

      }
      else {
          LogVerbose("In Child - parent has \$parentpid=[$parentpid]");
          # Execute the mainDispatcher() function in the child
          mainDispatcher();
          LogVerbose("In Child - child has finished");
      }
      # reset the alarm
      alarm(0);

  };

  # check the return values of the eval statement
  if ($@ and $@ =~ /Alarm expired/) {
      carp "Timeout! Alarm went off after \$timeout=[$timeout] seconds! - Exiting";
      exit 2;
  }
  else {
      LogVerbose("Child returned with: \$save_child_rc=[$save_child_rc]");
      LogVerbose("Continue and exit!");
  }

  # DONE

  exit $save_child_rc;
}
